{******************************************************************************}
{**                                                                          **}
{**  Bitmap routines.                                                        **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**  Copyright  1995 Borland Inerternational                                **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntBmp;

{$Include AntUnits.Inc}

Interface

Uses
  Windows, SysUtils, Graphics;

{*******************************************************************************
  Function convert any Picture's Graphic to Bitmap
*******************************************************************************}
Function krpPictureToBitmap(Const APicture : TPicture;
                            Const ABitmap  : TBitmap) : Boolean;

{*******************************************************************************
  Function Copy part of Desktop (ADesktopRect) to Context (ADC) by StretchBlt,
so image will be stretched to AClientRect.
*******************************************************************************}
Function krpCopyDesktopRect(Const ADC          : HDC;
                            Const AClientRect  : TRect;
                            Const ADesktopRect : TRect) : Boolean;

{*******************************************************************************
  Load Bitmap from Resource or File
*******************************************************************************}
Function bmpLoadBitmap(Const AName : String) : HBitMap;
Function bmpLoadBitmapFromFile(Const AFileName : String) : HBitMap;

{*******************************************************************************
  Make Bliting With Transparent Effect
*******************************************************************************}
Function bmpTransparentStretchBlt(DstDC                  : HDC;
                                  DstX, DstY, DstW, DstH : integer;
                                  SrcDC                  : HDC;
                                  SrcX, SrcY, SrcW, SrcH : integer;
                                  MaskDC                 : HDC;
                                  MaskX, MaskY           : integer) : Boolean;


{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}
{$IfDef Delphi4andHigher}
Uses
  Jpeg;
{$EndIf Delphi4andHigher}


{*******************************************************************************
  Function convert any Picture's Graphic to Bitmap
*******************************************************************************}
Function krpPictureToBitmap(Const APicture : TPicture;
                            Const ABitmap  : TBitmap) : Boolean;
Begin
  Result := False;

  if (ABitmap = nil) or (APicture = nil) or (APicture.Graphic = nil)
   then Exit;

  with ABitmap do
   try
     Width  := APicture.Width;
     Height := APicture.Height;
{$IfDef Delphi3andHigher}
     HandleType  := bmDIB;

  {$IfDef Delphi4andHigher}
   {** Aproximation for JPEG **}
     if (APicture.Graphic is TJPEGImage)
      then PixelFormat := pf8bit;{}
  {$EndIf Delphi4andHigher}

{$EndIf Delphi3andHigher}

     Canvas.Draw(0, 0, APicture.Graphic);
   finally
   end;{with ABitmap do try..}

  Result := True;
End;{Function krpPictureToBitmap}


{*******************************************************************************
  Function Copy part of Desktop (ADesktopRect) to Context (ADC) by StretchBlt,
so image will be stretched to AClientRect.
*******************************************************************************}
Function krpCopyDesktopRect(Const ADC          : HDC;
                            Const AClientRect  : TRect;
                            Const ADesktopRect : TRect) : Boolean;
Var
  DesktopWND : HWND;
  DesktopDC  : HDC;
Begin
  DesktopWND := GetDesktopWindow;
  DesktopDC  := GetDC(DesktopWND);
  try
    Result := StretchBlt(ADC, AClientRect.Left , AClientRect.Top,
                              AClientRect.Right, AClientRect.Bottom,
                         DesktopDC, ADesktopRect.Left , ADesktopRect.Top,
                                   (ADesktopRect.Right  - ADesktopRect.Left),
                                   (ADesktopRect.Bottom - ADesktopRect.Top ),
                         SRCCOPY);
  finally
    ReleaseDC(DesktopWND, DesktopDC);
  end;{try..finally}
End;{Function krpCopyDesktopRect}



{*******************************************************************************
  Load Bitmap from Resource or File
*******************************************************************************}
Function bmpLoadBitmap(Const AName : String) : HBitMap;
Begin
  Result := LoadImage(GetModuleHandle(nil), PChar(AName),
                      Image_BitMap, 0, 0, LR_CreateDibSection) ;
  if Result = 0
   then  Result := LoadImage(0, PChar(AName), Image_BitMap,
                             0, 0, LR_CreateDibSection or LR_LoadFromFile);
  if Result = 0
   then Raise Exception.CreateFmt('Unable to load bitmap %s', [AName]);{}
End;{Function bmpLoadBitMap}

Function bmpLoadBitmapFromFile(Const AFileName : String) : HBitMap;
Begin
  Result := LoadImage(0, PChar(AFileName), Image_BitMap,
                      0, 0, LR_CreateDibSection or LR_LoadFromFile);
  if Result = 0
   then Raise Exception.CreateFmt('Unable to load bitmap from %s file',
                                  [AFileName]);{}
End;{Function bmpLoadBitmapFromFile}


{*******************************************************************************
  Make Bliting With Transparent Effect
*******************************************************************************}
Function bmpTransparentStretchBlt(DstDC                  : HDC;
                                  DstX, DstY, DstW, DstH : integer;
                                  SrcDC                  : HDC;
                                  SrcX, SrcY, SrcW, SrcH : integer;
                                  MaskDC                 : HDC;
                                  MaskX, MaskY           : integer) : Boolean;
Const
  ROP_DstCopy = $00AA0029;
Var
  MemDC  : HDC;
  MemBmp : HBITMAP;
  Save   : THandle;
Begin
  Result := False;

{** Work Arround MakeRop4 Makros For Win NT **}
  if (Win32Platform = VER_PLATFORM_WIN32_NT)
      and (SrcW = DstW) and (SrcH = DstH) then
   begin
     MemBmp := CreateCompatibleBitmap(SrcDC, 1, 1);
     if MemBmp = 0
      then Exit;
     MemBmp := SelectObject(MaskDC, MemBmp);
     try
       MaskBlt(DstDC , DstX , DstY, DstW, DstH,
               SrcDC , SrcX , SrcY,
               MemBmp, MaskX, MaskY,
               MakeRop4(ROP_DstCopy, SrcCopy));
     finally
       MemBmp := SelectObject(MaskDC, MemBmp);
       DeleteObject(MemBmp);
     end;{try..finally}
     Result := True;
     Exit;
   end;{if (Win32Platform = ....}

{** Create DC For Work Arround Mask **}
  MemDC := CreateCompatibleDC(0);
  if MemDC = 0
   then Exit;

  try
    MemBmp := CreateCompatibleBitmap(SrcDC, SrcW, SrcH);
    if MemBmp = 0
     then Exit;
    Save := SelectObject(MemDC, MemBmp);
  {** Blit To Memory **}
    StretchBlt(MemDC ,     0,     0, SrcW, SrcH,
               MaskDC, MaskX, MaskY, SrcW, SrcH,
               NOTSRCCOPY);
    StretchBlt(MemDC,    0,    0, SrcW, SrcH,
               SrcDC, SrcX, SrcY, SrcW, SrcH,
               SrcAnd);
  {** Blit To Dest **}
    StretchBlt(DstDC , DstX , DstY , DstW, DstH,
               MaskDC, MaskX, MaskY, SrcW, SrcH,
               SrcAnd); {}
    StretchBlt(DstDC, DstX, DstY, DstW, DstH,
               MemDC,    0,    0, SrcW, SrcH,
               SrcPaint {SrcInvert});
  {** Delete Saved Object **}
    if Save <> 0
     then SelectObject(MemDC, Save);
    DeleteObject(MemBmp);
    Result := True;
  finally
  end;{try..finally}
End;{Function bmpTransparentStretchBlt}


END{Unit AntBmp}.
