{******************************************************************************}
{**                                                                          **}
{**  Date and Time routines. Mainly based on 3rd part routines.              **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntDate;

{$Include AntUnits.Inc}

Interface

{** Info Routines **}
Function DaysPerMonth(AYear, AMonth : integer) : integer;
Function FirstDayOfNextMonth : TDateTime;
Function FirstDayOfPrevMonth : TDateTime;
Function LastDayOfPrevMonth : TDateTime;
Function CurrentYear : Word;

{** Extracting Date or its part **}
Function ExtractDay(ADate : TDateTime) : Word;
Function ExtractMonth(ADate : TDateTime) : Word;
Function ExtractYear(ADate : TDateTime) : Word;

{** Date Changing (Shifting, incresing...) **}
Function IncDate(ADate               : TDateTime;
                 Days, Months, Years : integer) : TDateTime;
Function IncDay(ADate : TDateTime;
                Delta : integer) : TDateTime;
Function IncMonth(ADate : TDateTime;
                  Delta : integer) : TDateTime;
Function IncYear(ADate : TDateTime;
                 Delta : integer) : TDateTime;

{** Time Changing (Shifting, incresing...) **}
Function IncTime(ATime                          : TDateTime;
                 Hours, Minutes, Seconds, MSecs : integer) : TDateTime;
Function IncHour(ATime : TDateTime;
                 Delta : integer) : TDateTime;
Function IncMinute(ATime : TDateTime;
                   Delta : integer) : TDateTime;
Function IncSecond(ATime : TDateTime;
                   Delta : integer) : TDateTime;
Function IncMSec(ATime : TDateTime;
                 Delta : integer) : TDateTime;

{** Date Deffirents and intervals **}
Procedure DateDiff(Date1, Date2            : TDateTime;
                   Var Days, Months, Years : Word);
Function DaysInPeriod(Date1, Date2 : TDateTime) : Longint;
Function DaysBetween(Date1, Date2 : TDateTime) : Longint;

{** Validation Routines **}
Function IsValidDate(Y, M, D : Word) : Boolean;
Function ValidDate(ADate : TDateTime) : Boolean;

{** Tools Routines **}
Function CutTime(ADate: TDateTime): TDateTime;

{$IfDef Delphi2andLower}
Function IsLeapYear(Year: Word): Boolean;
{$EndIf Delphi2andLower}

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}
Uses
  SysUtils, Windows;

{$IfDef Delphi2andLower}
Function IsLeapYear(Year: Word): Boolean;
Begin
  Result := (Year mod 4 = 0) and ((Year mod 100 <> 0) or (Year mod 400 = 0));
End;
{$EndIf Delphi2andLower}

{*******************}
{** Info Routines **}
{*******************}
Function DaysPerMonth(AYear, AMonth : integer) : integer;
Const
  DaysInMonth : Array[1..12] of integer = (
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
Begin
  Result := DaysInMonth[AMonth];
  if (AMonth = 2) and IsLeapYear(AYear)
   then Result := 29;
End;{Function DaysPerMonth}

Function FirstDayOfNextMonth : TDateTime;
Var
  Year, Month, Day : Word;
Begin
  DecodeDate(Date, Year, Month, Day);
  Day := 1;
  if Month >= 12 then
   begin
     Inc(Year);
     Month := 1;
   end
  else Inc(Month);
  Result := EncodeDate(Year, Month, Day);
End;{Function FirstDayOfNextMonth}

Function FirstDayOfPrevMonth : TDateTime;
Var
  Year, Month, Day: Word;
Begin
  DecodeDate(Date, Year, Month, Day);
  Day := 1;
  if Month <= 1 then
   begin
     Dec(Year);
     Month := 12;
   end
  else Dec(Month);
  Result := EncodeDate(Year, Month, Day);
End;{Function FirstDayOfPrevMonth}

Function LastDayOfPrevMonth : TDateTime;
Var
  Date             : TDateTime;
  Year, Month, Day : Word;
Begin
{** Get first day of prev month **}
  Date := FirstDayOfPrevMonth;
  DecodeDate(Date, Year, Month, Day);
{** Set Day to last day for that month **}
  Day := DaysPerMonth(Year, Month);
  Result := EncodeDate(Year, Month, Day);
End;{Function LastDayOfPrevMonth}

Function CurrentYear : Word;
Var
  SystemTime : TSystemTime;
Begin
  GetLocalTime(SystemTime);
  Result := SystemTime.wYear;
End;{Function CurrentYear}


{*********************************}
{** Extracting Date or its part **}
{*********************************}
Function ExtractDay(ADate : TDateTime) : Word;
Var
  A, B : Word;
Begin
  DecodeDate(ADate, A, B, Result);
End;

Function ExtractMonth(ADate : TDateTime) : Word;
Var
  A, B : Word;
Begin
  DecodeDate(ADate, A, Result, B);
End;

Function ExtractYear(ADate : TDateTime) : Word;
Var
  A, B : Word;
Begin
  DecodeDate(ADate, Result, A, B);
End;

{********************************************}
{** Date Changing (Shifting, incresing...) **}
{********************************************}
Function IncDate(ADate               : TDateTime;
                 Days, Months, Years : integer) : TDateTime;
Var
  D, M, Y: Word;
  Day, Month, Year: Longint;
Begin
  DecodeDate(D, Y, M, D);
  Year := Y; Month := M; Day := D;
{** Year increase **}
  Inc(Year, Years);
  Inc(Year, Months div 12);
{** Month increase **}
  Inc(Month, Months mod 12);
  if Month < 1 then
   begin
     Inc(Month, 12);
     Dec(Year);
   end else
  if Month > 12 then
   begin
    Dec(Month, 12);
    Inc(Year);
   end;{if Month > 12 then}
{** Day increase **}
  if Day > DaysPerMonth(Year, Month)
   then Day := DaysPerMonth(Year, Month); // Found at RX Library...
  Result := EncodeDate(Year, Month, Day) + Days + Frac(ADate);
End;{Function IncDate}

Function IncDay(ADate : TDateTime;
                Delta : integer) : TDateTime;
Begin
  Result := ADate + Delta;
End;

Function IncMonth(ADate : TDateTime;
                  Delta : integer) : TDateTime;
Begin
  Result := IncDate(ADate, 0, Delta, 0);
End;

Function IncYear(ADate : TDateTime;
                 Delta : integer) : TDateTime;
Begin
  Result := IncDate(ADate, 0, 0, Delta);
End;


{********************************************}
{** Time Changing (Shifting, incresing...) **}
{********************************************}
Function IncTime(ATime                          : TDateTime;
                 Hours, Minutes, Seconds, MSecs : integer) : TDateTime;
Begin
  Result := ATime + (Hours div 24) +
            (((Hours mod 24) * 3600000 +
              Minutes * 60000 +
              Seconds * 1000 +
              MSecs) / MSecsPerDay);
  if Result < 0
   then Result := Result + 1;
End;{Function IncTime}

Function IncHour(ATime : TDateTime;
                 Delta : integer) : TDateTime;
Begin
  Result := IncTime(ATime, Delta, 0, 0, 0);
End;

Function IncMinute(ATime : TDateTime;
                   Delta : integer) : TDateTime;
Begin
  Result := IncTime(ATime, 0, Delta, 0, 0);
End;

Function IncSecond(ATime : TDateTime;
                   Delta : integer) : TDateTime;
Begin
  Result := IncTime(ATime, 0, 0, Delta, 0);
End;

Function IncMSec(ATime : TDateTime;
                 Delta : integer) : TDateTime;
Begin
  Result := IncTime(ATime, 0, 0, 0, Delta);
End;


{***********************************}
{** Date Deffirents and intervals **}
{***********************************}
Procedure DateDiff(Date1, Date2            : TDateTime;
                   Var Days, Months, Years : Word);
Var
  RegDate : TDateTime;
  Day1, Day2, Month1, Month2, Year1, Year2: Word;
Begin
  if Date1 > Date2 then
   begin
     RegDate := Date1; Date1 := Date2; Date2 := RegDate;
   end;{if Date1 > Date2 then}
  DecodeDate(Date1, Year1, Month1, Day1);
  DecodeDate(Date2, Year2, Month2, Day2);
  Years  := Year2 - Year1;
  Months := 0;
  Days   := 0;
{** Month fixing **}
  if Month2 < Month1 then
   begin
     Months := 12;
     Dec(Years);
   end;{if Month2 < Month1 then}
  Inc(Months, Month2 - Month1);
{** Day fixing **}
  if Day2 < Day1 then
   begin
     Inc(Days, DaysPerMonth(Year1, Month1));
     if Months = 0 then
      begin
        Dec(Years);
        Months := 11;
      end
     else Dec(Months);
   end;{if Day2 < Day1 then}
  Inc(Days, Day2 - Day1);
End;{Procedure DateDiff}

Function DaysInPeriod(Date1, Date2 : TDateTime) : Longint;
Begin
  if ValidDate(Date1) and ValidDate(Date2)
   then Result := Abs(Trunc(Date2) - Trunc(Date1)) + 1
   else Result := 0;
End;{Function DaysInPeriod}

Function DaysBetween(Date1, Date2 : TDateTime) : Longint;
Begin
  Result := Trunc(Date2) - Trunc(Date1) + 1;
  if Result < 0
   then Result := 0;
End;{Function DaysBetween}


{*************************}
{** Validation Routines **}
{*************************}
Function IsValidDate(Y, M, D : Word) : Boolean;
Begin
  Result := (Y >= 1) and (Y <= 9999)
            and (M >= 1) and (M <= 12)
            and (D >= 1) and (D <= DaysPerMonth(Y, M));
End;{Function IsValidDate}

Function ValidDate(ADate : TDateTime) : Boolean;
Var
  Year, Month, Day : Word;
Begin
  try
    DecodeDate(ADate, Year, Month, Day);
    Result := IsValidDate(Year, Month, Day);
  except
    Result := False;
  end;
End;{Function ValidDate}

{********************}
{** Tools Routines **}
{********************}
Function CutTime(ADate: TDateTime): TDateTime;
Begin
  Result := Trunc(ADate);
End;


END{Unit AntDate}.
