{******************************************************************************}
{**                                                                          **}
{**  Some specific file routines.                                            **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**  Copyright  1999, UtilMind Solutions.                                   **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntFile;

{$Include AntUnits.Inc}

Interface

Uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics;

{** Create Directory Properly **}
Function CreateDirectoryCool(Const ADir : String) : Boolean;

{** Fix File Name By Checking All Charecters **}
Function FixFileName(AFileName : String) : String;

{** Select Graphics Type by Filename Extention **}
Function GetGraphicTypeByExtention(Const AExt     : String) : TGraphicClass;
Function GetGraphicTypeByFileName(Const AFileName : String) : TGraphicClass;

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}
Uses
{$IfDef Delphi4andHigher}
  Jpeg,
{$EndIf Delphi4andHigher}
  AntGif, AntPcx;

{*******************************}
{** Create Directory Properly **}
{*******************************}
Function CreateDirectoryCool(Const ADir : String) : Boolean;
Var
  i,Len : integer;
Begin
  Result := True;
  Len := Length(ADir);
  try
    for i:=4 to Len do
     begin
       if ADir[i] in ['/','\']
        then Result := CreateDirectory(PChar(Copy(ADir, 1, i)), nil);
     end;{for i:=4 to Len do}
  except
    Result := False;
  end;{try..except}
End;{Function CreateDirectoryCool}

{**********************************************}
{** Fix File Name By Checking All Charecters **}
{**********************************************}
Function FixFileName(AFileName : String) : String;
Var
  i : integer;
Begin
  for i:=Length(AFileName) downto 1 do
   begin
     if AFileName[i] in ['0'..'9','A'..'Z','a'..'z','.','_']
      then Continue;
     if AFileName[i] in ['/','|']
      then AFileName[i] := '\'
      else Delete(AFileName, i, 1);
   end;{for i:=Length(FileName) downto 1 do}
  Result := AFileName;
End;{Function FixFileName}

Function GetGraphicTypeByExtention(Const AExt : String) : TGraphicClass;
Begin
  Result := nil;
  if CompareText(AExt,'bmp') = 0
   then  Result := TBitmap
   else
  if CompareText(AExt,'ico') = 0
   then  Result := TIcon
   else
  if CompareText(AExt,'wmf') = 0
   then  Result := TMetafile
   else
  if CompareText(AExt,'emf') = 0
   then  Result := TMetafile
{$IfDef Delphi4andHigher}
   else
  if CompareText(AExt,'jpg') = 0
   then  Result := TJPEGImage
{$EndIf Delphi4andHigher}
   else
  if CompareText(AExt,'gif') = 0
   then  Result := TGif{}
   else
  if CompareText(AExt,'pcx') = 0
   then  Result := TPcx{}
End;{Function GetGraphicTypeByExtention}

Function GetGraphicTypeByFileName(Const AFileName : String) : TGraphicClass;
Var
  Ext : String;
Begin
  Ext := ExtractFileExt(AFileName);
  Delete(Ext, 1, 1);
  Result := GetGraphicTypeByExtention(Ext);
End;{Function GetGraphicTypeByFileName}

END{Unit AntFile}.
