{******************************************************************************}
{**                                                                          **}
{**  Some specific Math, Boolean and Geomrtric routines.                     **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**  Copyright  1999, UtilMind Solutions.                                   **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntMath;

{$Include AntUnits.Inc}

Interface

Uses
  Windows, SysUtils;

{******************************************************************************}
{** Binary Routines ***********************************************************}
{******************************************************************************}
Function ChangeByteOrderWord(AWord : Word) : Word;
Function ChangeByteOrderDWord(ADWord : DWord) : DWord;
Function BytesToWord(HiByte, LoByte : Byte) : Word;
Function Swap4Bits(AByte : Byte) : Byte;

{******************}
{** Bit Routines **}
{******************}
Function BitOfByte(AByte   : Byte;
                   ANumber : Byte) : Boolean;
Function BitOfWord(AWord   : Word;
                   ANumber : Byte) : Boolean;

{******************************************************************************}
{** Geometric Routines ********************************************************}
{******************************************************************************}
Function PointInRect(Const APoint : TPoint;
                     Const ARect  : TRect) : Boolean;
Function RectInRect (Const Rect1, Rect2 : TRect) : Boolean;
Function OverlapRect(Const Rect1, Rect2 : TRect) : Boolean;


{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}


{******************************************************************************}
{** Binary Routines ***********************************************************}
{******************************************************************************}
Function ChangeByteOrderWord(AWord : Word) : Word;
Begin
  Result := (AWord shl 8) + (AWord shr 8);
End;{Function ChangeByteOrderWord}

Function ChangeByteOrderDWord(ADWord : DWord) : DWord;
Begin
  Result := ( ADWord shl 24)                +
            ((ADWord shl 8 ) and $00FF0000) +
            ((ADWord shr 8 ) and $0000FF00) +
            ( ADWord shr 24);
End;{Function ChangeByteOrderDWord}

Function BytesToWord(HiByte, LoByte : Byte) : Word;
Begin
  Result := Word(HiByte) shl 8 + LoByte;
End;{Function BytesToWord}

{** Swap 4 bits in Byte **}
Function Swap4Bits(AByte : Byte) : Byte;
Begin
  Result := (AByte shl 4) + (AByte shr 4);
End;{Function Swap4Bits}


{******************}
{** Bit Routines **}
{******************}
Function BitOfByte(AByte   : Byte;
                   ANumber : Byte) : Boolean;
Begin
  case ANumber of
   0 :  Result := (AByte and $1 ) <> 0;
   1 :  Result := (AByte and $2 ) <> 0;
   2 :  Result := (AByte and $4 ) <> 0;
   3 :  Result := (AByte and $8 ) <> 0;
   4 :  Result := (AByte and $10) <> 0;
   5 :  Result := (AByte and $20) <> 0;
   6 :  Result := (AByte and $40) <> 0;
   else Result := (AByte and $80) <> 0;
  end;
End;{Function BitOfByte}

Function BitOfWord(AWord   : Word;
                   ANumber : Byte) : Boolean;
Begin
  if ANumber > 7
   then Result := BitOfByte(Hi(AWord), ANumber - 8)
   else Result := BitOfByte(Lo(AWord), ANumber);
End;{Function BitOfWord}


{******************************************************************************}
{** Geometric Routines ********************************************************}
{******************************************************************************}
Function PointInRect(Const APoint : TPoint;
                     Const ARect  : TRect) : Boolean;
Begin
  Result := (APoint.X >= ARect.Left)
            and (APoint.X <= ARect.Right )
            and (APoint.Y >= ARect.Top   )
            and (APoint.Y <= ARect.Bottom);
End;{Function PointInRect}

Function RectInRect(Const Rect1, Rect2 : TRect) : Boolean;
Begin
  Result := (Rect1.Left >= Rect2.Left)
            and (Rect1.Right  <= Rect2.Right )
            and (Rect1.Top    >= Rect2.Top   )
            and (Rect1.Bottom <= Rect2.Bottom);
End;{Function RectInRect}

Function OverlapRect(Const Rect1, Rect2 : TRect) : Boolean;
Begin
  Result := (Rect1.Left < Rect2.Right)
            and (Rect1.Right  > Rect2.Left  )
            and (Rect1.Top    < Rect2.Bottom)
            and (Rect1.Bottom > Rect2.Top   );
End;{Function OverlapRect}


END{Unit AntMath}.
