{******************************************************************************}
{**                                                                          **}
{**  Picture Editor like in Delphi IDE.                                      **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1995 Borland Inerternational                                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**  Copyright  1999, UtilMind Solutions.                                   **}
{**                                                                          **}
{******************************************************************************}
{**     BORLAND !!! ,          **}
{**   ?!    KARPOLAN.                                 **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntPEdit; {** Created from 'PicEdit' unit **}

Interface

Uses
  Windows, Classes, Graphics, Forms, Controls, Dialogs, Buttons, DsgnIntf,
  StdCtrls, ExtCtrls, ExtDlgs;

Type
  TantPictureEditorDlg = Class(TForm)
    OpenDialog : TOpenPictureDialog;
    SaveDialog : TSavePictureDialog;
    OKButton     : TButton;
    CancelButton : TButton;
    HelpButton   : TButton;
    GroupBox1  : TGroupBox;
    ImagePanel : TPanel;
    Load  : TButton;
    Save  : TButton;
    Clear : TButton;
    ImagePaintBox : TPaintBox;
    Label1: TLabel;
    Label2: TLabel;
    Procedure FormCreate(Sender: TObject);
    Procedure FormDestroy(Sender: TObject);
    Procedure LoadClick(Sender: TObject);
    Procedure SaveClick(Sender: TObject);
    Procedure ClearClick(Sender: TObject);
    Procedure HelpButtonClick(Sender: TObject);
    Procedure ImagePaintBoxPaint(Sender: TObject);
  Private
    Pic : TPicture;
  End;{TantPictureEditorDlg = Class(TForm)}

  TantPictureEditor = Class(TComponent)
  Private
    FGraphicClass : TGraphicClass;
    FPicture      : TPicture;
    FPicDlg       : TantPictureEditorDlg;
    Procedure SetPicture(Value: TPicture);
  Public
    Constructor Create(AOwner : TComponent);                           Override;
    Destructor  Destroy;                                               Override;
    Function Execute : Boolean;
    Property GraphicClass  : TGraphicClass
      Read   FGraphicClass
      Write  FGraphicClass;
    Property Picture       : TPicture
      Read   FPicture
      Write  SetPicture;
  End;{TantPictureEditor = Class(TComponent)}

{** TantPictureProperty
    Property editor the TPicture properties (e.g. the Picture Property).
    Brings up a file open dialog allowing loading a picture file.            **}
  TantPictureProperty = Class(TPropertyEditor)
  Public
    Procedure Edit;                                                    Override;
    Function  GetAttributes : TPropertyAttributes;                     Override;
    Function  GetValue : String;                                       Override;
    Procedure SetValue(Const Value : String);                          Override;
  End;

{** TantGraphicProperty *******************************************************}
  TantGraphicProperty = Class(TClassProperty)
  Public
    Procedure Edit;                                                    Override;
    Function  GetAttributes : TPropertyAttributes;                     Override;
    Function  GetValue : String;                                       Override;
    Procedure SetValue(Const Value : String);                          Override;
  End;

{** TantTGraphicEditor ********************************************************}
  TantGraphicEditor = Class(TDefaultEditor)
  Public
    Procedure EditProperty(PropertyEditor : TPropertyEditor;
                           Var Continue,
                           FreeEditor     : Boolean);                  Override;
  End;

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}
{$R *.DFM}

Uses
  TypInfo, SysUtils {, LibConst, LibHelp}; // And where that f****'n units ????

{******************************************************************************}
{** TPictureEditorDlg *********************************************************}
{******************************************************************************}
Procedure TantPictureEditorDlg.FormCreate(Sender: TObject);
Begin
  HelpContext := 0 {hcDPictureEditor};
  Pic := TPicture.Create;
  Save.Enabled := False;
End;

Procedure TantPictureEditorDlg.FormDestroy(Sender: TObject);
Begin
  Pic.Free;
End;

Procedure TantPictureEditorDlg.LoadClick(Sender: TObject);
Begin
  OpenDialog.Title := 'Open'{SLoadPictureTitle};
  if OpenDialog.Execute then
   begin
     Pic.LoadFromFile(OpenDialog.FileName);
     ImagePaintBox.Invalidate;
     Save.Enabled  := (Pic.Graphic <> nil) and not Pic.Graphic.Empty;
     Clear.Enabled := (Pic.Graphic <> nil) and not Pic.Graphic.Empty;
   end;{if OpenDialog.Execute then}
End;

Procedure TantPictureEditorDlg.SaveClick(Sender: TObject);
Begin
  if Pic.Graphic <> nil then
   begin
     SaveDialog.Title := 'Save'{SSavePictureTitle};
     with SaveDialog do
      begin
        DefaultExt := GraphicExtension(TGraphicClass(Pic.Graphic.ClassType));
        Filter := GraphicFilter(TGraphicClass(Pic.Graphic.ClassType));
        if Execute then Pic.SaveToFile(Filename);
      end;
   end;
End;

Procedure TantPictureEditorDlg.ImagePaintBoxPaint(Sender: TObject);
Var
  DrawRect : TRect;
  SNone    : String;
Begin
  with TPaintBox(Sender) do
   begin
     Canvas.Brush.Color := {Self.}Color;
     DrawRect := ClientRect; // Rect(Left, Top, Left + Width, Top + Height);
     if Pic.Width > 0 then
      begin
        with DrawRect do
          if (Pic.Width > Right - Left) or (Pic.Height > Bottom - Top) then
           begin
             if Pic.Width > Pic.Height then
               Bottom := Top + MulDiv(Pic.Height, Right - Left, Pic.Width)
             else
               Right := Left + MulDiv(Pic.Width, Bottom - Top, Pic.Height);
             Canvas.StretchDraw(DrawRect, Pic.Graphic);
           end{if (Pic.Width >.....}
          else
           with DrawRect do
             Canvas.Draw(Left + (Right - Left - Pic.Width) div 2, Top +
               (Bottom - Top - Pic.Height) div 2, Pic.Graphic);
      end{if Pic.Width > 0 then}
     else
       with DrawRect, Canvas do
        begin
          SNone := 'None'{srNone};
          TextOut(Left + (Right - Left - TextWidth(SNone)) div 2, Top + (Bottom -
            Top - TextHeight(SNone)) div 2, SNone);
        end;{with DrawRect, Canvas do}
   end;{with TPaintBox(Sender) do}
End;

Procedure TantPictureEditorDlg.ClearClick(Sender: TObject);
Begin
  Pic.Graphic := nil;
  ImagePaintBox.Invalidate;
  Save .Enabled := False;
  Clear.Enabled := False;
End;


{******************************************************************************}
{** TPictureEditor ************************************************************}
{******************************************************************************}
Constructor TantPictureEditor.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  FPicture := TPicture.Create;
  FPicDlg  := TantPictureEditorDlg.Create(Self);
  FGraphicClass := TGraphic;
End;

Destructor TantPictureEditor.Destroy;
Begin
  FPicture.Free;
  Inherited Destroy;
End;

Function TantPictureEditor.Execute: Boolean;
Begin
  FPicDlg.Pic.Assign(FPicture);
  with FPicDlg.OpenDialog do
   begin
     Options := [ofHideReadOnly, ofFileMustExist, ofShowHelp];

     DefaultExt := GraphicExtension(GraphicClass);
     if DefaultExt = ''
      then DefaultExt := GraphicExtension(TBitmap);{}
     Filter := GraphicFilter(GraphicClass);
     if Filter = ''
      then Filter := GraphicFilter(TGraphic);{}

     HelpContext := 0{hcDLoadPicture};
   end;

  with FPicDlg.SaveDialog do
   begin
     Options := [ofHideReadOnly, ofFileMustExist, ofShowHelp];

     DefaultExt := GraphicExtension(GraphicClass);
     if DefaultExt = ''
      then DefaultExt := GraphicExtension(TBitmap);{}
     Filter := GraphicFilter(GraphicClass);
     if Filter = ''
      then Filter := GraphicFilter(TGraphic);{}

     HelpContext := 0{hcDSavePicture};
   end;

  FPicDlg.Save .Enabled := (FPicture.Graphic <> nil) and not FPicture.Graphic.Empty;
  FPicDlg.Clear.Enabled := (FPicture.Graphic <> nil) and not FPicture.Graphic.Empty;

  Result := (FPicDlg.ShowModal = mrOK);
  if Result
   then FPicture.Assign(FPicDlg.Pic);
End;

Procedure TantPictureEditor.SetPicture(Value: TPicture);
Begin
  FPicture.Assign(Value);
End;

{******************************************************************************}
{** TPictureProperty **********************************************************}
{******************************************************************************}
Procedure TantPictureProperty.Edit;
Var
  PictureEditor: TantPictureEditor;
Begin
  PictureEditor := TantPictureEditor.Create(nil);
  try
    PictureEditor.Picture := TPicture(Pointer(GetOrdValue));
    if PictureEditor.Execute
     then SetOrdValue(Longint(PictureEditor.Picture));
  finally
    PictureEditor.Free;
  end;
End;

Function TantPictureProperty.GetAttributes : TPropertyAttributes;
Begin
  Result := [paDialog];
End;

Function TantPictureProperty.GetValue : String;
Var
  Picture: TPicture;
Begin
  Picture := TPicture(GetOrdValue);
  if Picture.Graphic = nil
   then Result := '(None)'{srNone}
   else Result := '(' + Picture.Graphic.ClassName + ')';
End;

Procedure TantPictureProperty.SetValue(Const Value: String);
Begin
  if Value = ''
   then SetOrdValue(0);
End;

{******************************************************************************}
{** TGraphicProperty **********************************************************}
{******************************************************************************}
Procedure TantGraphicProperty.Edit;
Var
  PictureEditor: TantPictureEditor;
Begin
  PictureEditor := TantPictureEditor.Create(nil);
  try
    PictureEditor.GraphicClass := TGraphicClass(GetTypeData(GetPropType)^.ClassType);
    PictureEditor.Picture.Graphic := TGraphic(Pointer(GetOrdValue));
    if PictureEditor.Execute then
      if (PictureEditor.Picture.Graphic = nil)
         or (PictureEditor.Picture.Graphic is PictureEditor.GraphicClass)
       then SetOrdValue(LongInt(PictureEditor.Picture.Graphic))
       else Raise Exception.Create('Invalid Format');
  finally
    PictureEditor.Free;
  end;
End;

Function TantGraphicProperty.GetAttributes: TPropertyAttributes;
Begin
  Result := [paDialog];
End;

Function TantGraphicProperty.GetValue: String;
Var
  Graphic: TGraphic;
Begin
  Graphic := TGraphic(GetOrdValue);
  if (Graphic = nil) or Graphic.Empty
   then Result := '(None)'{srNone}
   else Result := '(' + Graphic.ClassName + ')';
End;

Procedure TantGraphicProperty.SetValue(Const Value: String);
Begin
  if Value = ''
   then SetOrdValue(0);
End;

{******************************************************************************}
{** TPictureEditor ************************************************************}
{******************************************************************************}
Procedure TantGraphicEditor.EditProperty(PropertyEditor : TPropertyEditor;
                                         Var Continue,
                                             FreeEditor : Boolean);
Var
  PropName: String;
Begin
  PropName := PropertyEditor.GetName;
  if (CompareText(PropName, 'PICTURE') = 0)
     or (CompareText(PropName, 'IMAGE') = 0) then
   begin
     PropertyEditor.Edit;
     Continue := False;
   end;
End;

Procedure TantPictureEditorDlg.HelpButtonClick(Sender: TObject);
Begin
  Application.HelpContext(HelpContext);
End;

END{Unit AntPEdit}.
