{******************************************************************************}
{**                                                                          **}
{**  PCX, PCL file routines. Tested only on 8 bits PCX.                      **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**  Copyright  1999, UtilMind Solutions.                                   **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntPcx;

{$Include AntUnits.Inc}

{*******************************************************************************
   Install this file like component and standart TPicture properties will
   understand PCX format. You can use standart TImage component for PCXs
*******************************************************************************}

Interface

Uses
  Windows, SysUtils, Classes, Graphics;

Const
{** Deafault Signatura for PCX Resources **}
  pcxResTypeName   = 'PCX';
{** Deafault PCX File Signature (8 BPP) **}
  pcxFileSignature = $0A050108;

Type
{** PCX File Header, Size = 128 **}
  TPcxHeader = Packed Record
    Manufact    : Byte;                  { Manufactor     $OA       }
    Version     : Byte;                  { Version        $05       }
    EnCode      : Byte;                  { Group Coding   $01       }
    BitPerPixel : Byte;                  { Bit Per Pixel ($08)      }
    imgLeft     : Word;                  { Picture Coords...        }
    imgTop      : Word;                  {                          }
    imgRight    : Word;                  {                          }
    imgBottom   : Word;                  {                          }
    HRes        : Word;                  { Horisontal Resolution    }
    VRes        : Word;                  { Vertical Resolution      }
    Pal         : Array[0..47] of Byte;  { EGA Pallete             }
    VideoMode   : Byte;                  { Video Mode (Reserved)    }
    NumPlanes   : Byte;                  { Numder Of Color Planes   }
    BytePerLine : Word;                  { Byte Per Line            }
    PalInfo     : Word;                  { (1-Color,2-Gray)         }
    ScanHRes    : Word;                  { ScanersHor.  Resolution }
    ScanVRes    : Word;                  { ScanersVert. Resolution }
    Xtra        : Array[0..53] of Byte;  { Extra Data (Padding)     }
  End;{TPcxHeader = Packed Record}

{** PCL (PCX Library) File Header **}
  TPclHeader = Packed Record
    id{Software} : Array[0..9 ] of Byte;   { Lib Number               }
    Copyright    : Array[0..49] of Byte;   { Autor's Copyright        }
    Version      : Word;                   { Version PCL              }
    Note         : Array[0..39] of Char;   { Tome Label               }
    Xtra         : Array[0..19] of Byte;   { Extra Data (Filter)      }
  End;{TPclHeader = Packed Record}

{** PCX File in PCL Library Header **}
  TPcxInPclHeader = Packed Record
    SyncByte   : Byte;                   { Byte Of Syncro           }
    FileName   : Array[0..11] of Char;   { File Name                }
    FileSize   : DWord;                  { File Size               }
    FileDate   : Word;                   { File Date               }
    FileTime   : Word;                   { File Time               }
    FilePack   : Word;                   { Type Of Packing          }
    Note       : Array[0..39] of Char;   { File Label               }
    Xtra       : Array[0..20] of Byte;   { Extra Data (Filter)      }
  End;{TPcxInPclHeader = Packed Record}

{** PCX Color Data **}
  TPcxRGBColor = Packed Record
    Red   : Byte;
    Green : Byte;
    Blue  : Byte;
  End;{TPcxRGBColor = Packed Record}

{** PCX Palette, Size = 768 **}
  TPcxPalette = Array[0..255] of TPcxRGBColor;

{** PCX/PCL Memory Routine **}
  TPcxFileRecord = Packed Record
    pcxHeader      : TPcxHeader;
    pcxRleData     : Pointer;
    pcxRleDataSize : DWord;
    pcxPalette     : TPcxPalette;
  End;{TPcxFileRecord = Packed Record}
  TPcxInPclPictureRecord = TPcxFileRecord;


{##############################################################################}
{## New class for holding PCX Graphics ########################################}
{##############################################################################}
  EPcx = Class(Exception);

  TPcx = Class(TGraphic)
  Private
    fPcxStream : TMemoryStream;
    fBitmap    : TBitmap;
  {** Streams Routine **}
    Procedure WriteStream(Stream    : TStream;
                          WriteSize : Boolean);
    Procedure ReadStream(Size   : LongInt;
                         Stream : TStream);
  Protected
  {** Override This **}
    Procedure Draw(ACanvas    : TCanvas;
                   Const Rect : TRect);                                Override;
  {** Convert Routine **}
    Procedure Convert;                                                  Virtual;
  {** Properties Overrides **}
    Function  GetEmpty : Boolean;                                      Override;
    Function  GetWidth    : integer;                                   Override;
    Procedure SetWidth (A : integer);                                  Override;
    Function  GetHeight   : integer;                                   Override;
    Procedure SetHeight(A : integer);                                  Override;
  {** Streams Routine **}
    Procedure ReadData (Stream : TStream);                             Override;
    Procedure WriteData(Stream : TStream);                             Override;

  Public
    PcxHeader : TPcxHeader;
  {** Init/Done **}
		Constructor Create;                                                Override;
		Destructor  Destroy;                                               Override;
  {** Override This **}
    Procedure Assign(Source : TPersistent);                            Override;
  {** Files Routine **}
    Procedure LoadFromFile(Const FileName : String);                   Override;
    Procedure SaveToFile  (Const FileName : String);                   Override;
  {** Streams Routine **}
    Procedure LoadFromStream(Stream : TStream);                        Override;
    Procedure SaveToStream  (Stream : TStream);                        Override;
  {** ClipBoard Routine **}
    Procedure LoadFromClipBoardFormat(AFormat  : Word;
                                      AData    : THandle;
                                      APalette : HPalette);            Override;
    Procedure SaveToClipBoardFormat(Var AFormat  : Word;
                                    Var AData    : THandle;
                                    Var APalette : HPalette);          Override;
  {** Properties Routine **}
    Property PcxStream   : TMemoryStream
      Read   fPcxStream
      Write  fPcxStream;
    Property Bitmap      : TBitmap
      Read   fBitmap
      Write  fBitmap;
  {** Old Properties **}
    Property Empty;
    Property Height;
    Property Modified;
    Property Width;
    Property OnChange;
  End;{TPcx = Class(TGraphic)}


{##############################################################################}
{## VCL PCX Routines ##########################################################}
{##############################################################################}
{** Convert PCX file Styled Stream To BMP Styled One (Main Convertion Proc) **}
Procedure pcxStreamConvert(APcxStream : TStream;
                           ABmpStream : TMemoryStream);

{** Load PCX from hIstance or From File **}
Function pcxLoadAsBitmap(Const AName : String) : Graphics.TBitMap;

{** Load PCX from File **}
Function pcxLoadAsBitmapFromFile(Const AFileName : String) : Graphics.TBitMap;

{** Load PCX from Stream **}
Function pcxLoadAsBitmapFromStream(AStream : TStream) : Graphics.TBitMap;

{** Load PCX from Stream (From Position and By Size)**}
Function pcxLoadAsBitmapFromStreamDirectly(AStream   : TStream;
                                           APosition : LongInt;
                                           ASize     : LongInt) : Graphics.TBitMap;


{##############################################################################}
{## Windows API PCX Routines ##################################################}
{##############################################################################}
{** Load PCX file to memory **}
Function LoadPcxFile(Const AFileName    : String;
                     Var   APcxFileRec  : TPcxFileRecord) : Boolean;

{** Convert Loaded PCX file to Bitmap **}
Function ConvertPcxToBitmap(APcxFileRec  : TPcxFileRecord;
                            Var ABitmap  : HBitmap;
                            Var APalette : HPalette) : Boolean;

{** Load Pcx File and Convert Data to Bitmap and Palette **}
Function LoadPcxFileToBitmap(Const AFileName : String;
                             Var ABitmap     : HBitmap;
                             Var APalette    : HPalette) : Boolean;

Procedure Register;

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}

Procedure Register;
Begin
  RegisterClass(TPcx);
//  TPicture.RegisterFileFormat('PCX', 'PCX Files', TPcx);
End;{Procedure Register}

{##############################################################################}
{## TPcx = Class(TGraphic) ####################################################}
{##############################################################################}
Const
  bmpFileSignature  = $4d42;
  strNoFile         = 'PCX File Not Found';
  strWrongFile      = 'File isn''t a PCX File';

{******************************************************************************}
{** Public ********************************************************************}
{******************************************************************************}
{** Init/Done **}
{***************}
Constructor TPcx.Create;
Begin
  fPcxStream := TMemoryStream.Create;
  fBitmap    := TBitmap.Create;
  Inherited Create;
End;{Constructor TPcx.Create}

Destructor TPcx.Destroy;
Begin
  Inherited Destroy;
  fBitmap.Free;
  fPcxStream.Free;
End;{Destructor TPcx.Destroy}

{*******************}
{** Override This **}
{*******************}
Procedure TPcx.Assign(Source : TPersistent);
Begin
  if (Source is TPcx) then
   begin
     fPcxStream.LoadFromStream(TPcx(Source).PcxStream);
   {** Change yourself **}
     Convert;
     Changed(Self);
   end
  else Inherited Assign(Source);
End;{Procedure TPcx.Assign}

{*******************}
{** Files Routine **}
{*******************}
Procedure TPcx.LoadFromFile(Const FileName: String);
Begin
{** Check File Existing **}
  if not FileExists(FileName)
   then Raise EPcx.Create(strNoFile);
{** Check File Extention **}
  if (UpperCase(ExtractFileExt(FileName)) <> '.PCX')
   then Raise EPcx.Create(strWrongFile);{}
{** Load fPcxStream From File **}
  fPcxStream.LoadFromFile(FileName);
{** Change yourself **}
  Convert;
  Changed(Self);
End;{Procedure TPcx.LoadFromFile}

Procedure TPcx.SaveToFile(Const FileName: String);
Begin
  fPcxStream.SaveToFile(FileName);
End;{Procedure TPcx.SaveToFile}

{********************}
{** Stream Routine **}
{********************}
Procedure TPcx.LoadFromStream(Stream : TStream);
Begin
{** Check For Stream **}
  if not Assigned(Stream)
   then Exit;
{** Load fPcxStream From Stream **}
  fPcxStream.LoadFromStream(Stream);
{** Change yourself **}
  Convert;
  Changed(Self);
End;{Procedure TPcx.LoadFromStream}

Procedure TPcx.SaveToStream(Stream : TStream);
Begin
  WriteStream(Stream, False);
End;{Procedure TPcx.SaveToStream}

{***********************}
{** ClipBoard Routine **}
{***********************}
Procedure TPcx.LoadFromClipBoardFormat(AFormat  : Word;
                                       AData    : THandle;
                                       APalette : HPalette);
Begin
End;{Procedure TPcx.LoadFromClipBoardFormat}

Procedure TPcx.SaveToClipBoardFormat(Var AFormat  : Word;
                                     Var AData    : THandle;
                                     Var APalette : HPalette);
Begin
End;{Procedure TPcx.SaveToClipBoardFormat}

{******************************************************************************}
{** Protected *****************************************************************}
{******************************************************************************}
{** Override This **}
{*******************}
Procedure TPcx.Draw(ACanvas    : TCanvas;
                    Const Rect : TRect);
Var
  destDC                : HDC;
  destX, destY,
  destWidth, destHeight : SmallInt;
  srcDC                 : HDC;
  srcX, srcY,
  srcWidth, srcHeight   : SmallInt;
  ROpFlag               : LongInt;
Begin
{** Prepare Destanation **}
  destDC := ACanvas.Handle;
  with Rect do
   begin
     destX      := Left;
     destY      := Top;
     destWidth  := Right  - Left;
     destHeight := Bottom - Top;
   end;{with Rect do}
{** Prepare Source **}
  srcDC := fBitmap.Canvas.Handle;
  with fBitmap do
   begin
     srcX      := 0;
     srcY      := 0;
     srcWidth  := Width;
     srcHeight := Height;
   end;{with fBitmap do}
{** Set Raster Operations Flag **}
  ROpFlag := ACanvas.CopyMode;
{** StretchBlt **}
  StretchBlt(destDC, destX, destY, destWidth, destHeight,
             srcDC,  srcX,  srcY,  srcWidth,  srcHeight,
             ROpFlag);
End;{Procedure TPcx.Draw}


{*************************}
{** Convertaion Routine **}
{*************************}
Procedure TPcx.Convert;
Var
  msTemp : TMemoryStream;
Begin
  fPcxStream.Seek(0, soFromBeginning);
  fPcxStream.Read(PcxHeader, SizeOf(PcxHeader));

  msTemp := TMemoryStream.Create;
  try
    pcxStreamConvert(fPcxStream, msTemp);
    fBitmap.LoadFromStream(msTemp)
  finally
    msTemp.Free;
  end;{with msTemp do try..}
End;{Procedure TPcx.Convert}


{**************************}
{** Properties Overrides **}
{**************************}
Function TPcx.GetEmpty : Boolean;
Begin
  Result := fBitmap.Empty;
End;{Function TPcx.GetEmpty}

Function TPcx.GetWidth : integer;
Begin
 with PcxHeader do
  Result := ((imgRight - imgLeft) + 1);{}

// Result := fBitmap.Width;
End;{Function TPcx.GetWidth}
Procedure TPcx.SetWidth(A : integer);
Begin
End;{Procedure TPcx.SetWidth}

Function TPcx.GetHeight : integer;
Begin
 with PcxHeader do
  Result := ((imgBottom - imgTop ) + 1);{}

// Result := fBitmap.Height;
End;{Function TPcx.GetHeight}
Procedure TPcx.SetHeight(A : integer);
Begin
End;{Procedure TPcx.SetHeight}

{*********************}
{** Streams Routine **}
{*********************}
Procedure TPcx.ReadData(Stream : TStream);
Var
  Size: LongInt;
Begin
  Stream.Read(Size, SizeOf(Size));
  ReadStream(Size, Stream);
{** Change yourself **}
  Convert;
  Changed(Self);
End;{Procedure TPcx.ReadData}

Procedure TPcx.WriteData(Stream : TStream);
Begin
  WriteStream(Stream, True);
End;{Procedure TPcx.WriteData}

{******************************************************************************}
{** Private *******************************************************************}
{******************************************************************************}
{** Streams Routine **}
{*********************}
Procedure TPcx.ReadStream(Size   : LongInt;
                          Stream : TStream);
Begin
  fPcxStream.SetSize(Size);
  Stream.ReadBuffer(fPcxStream.Memory^, Size);
End;{Procedure TPcx.ReadStream}

Procedure TPcx.WriteStream(Stream    : TStream;
                           WriteSize : Boolean);
Var
  Size : LongInt;
Begin
  Size := fPcxStream.Size;
  if WriteSize
   then Stream.WriteBuffer(Size, SizeOf(Size));
  if Size > 0
   then Stream.WriteBuffer(fPcxStream.Memory^, Size);
End;{Procedure TPcx.WriteStream}


{##############################################################################}
{## VCL PCX Routines ##########################################################}
{##############################################################################}
Const
  strErrFromResource  = 'Unable load "%s" PCX resource';
  strErrFromFile      = 'Unable load PCX from "%s" file';
  strErrFromStream    = 'Unable load PCX from stream';

{**************************************************}
{** Convert PCX file Styled Stream To BMP Styled **}
{**************************************************}
Procedure pcxStreamConvert(APcxStream : TStream;
                           ABmpStream : TMemoryStream);
Const
  Bounder : DWord     = $00000000;
Var
{** PCX Routine **}
  APcxFileRec       : TPcxFileRecord;
  RleDataPointer    : LongInt;
  pcxWidth          : LongInt;
  pcxHeight         : LongInt;
{** Memory Image Routine **}
  ByteOfData        : Byte;
  BytesCount        : Byte;
  ptrImage          : Pointer;
  ImageLen          : LongInt;
  ImagePointer      : LongInt;
{** Bmp Routine **}
  ABitmapInfoHeader : Windows.TBitmapInfoHeader;
  ABitmapFileHeader : Windows.TBitmapFileHeader;
  BoundarySize      : LongInt;

  Procedure ConvertRleToBmp;
  Var
    Count : integer;
  Begin
    ImagePointer   := Integer(ptrImage);
    RleDataPointer := Integer(APcxFileRec.pcxRleData);

    Count := 0;
    while Count < ImageLen do
     begin
     {** Read One Byte of Data **}
       ByteOfData := Byte(Pointer(RleDataPointer)^);

     {** Check This Byte **}
       if ByteOfData >= 192 then
        begin {** Byte is technickal = (Length of bytes row + 192) **}
          BytesCount := ByteOfData - 192;
        {** Read next byte - Real value of color **}
          Inc(RleDataPointer);
          ByteOfData := Byte(Pointer(RleDataPointer)^);
        {** Fill row of bytes **}
          FillChar(Pointer(ImagePointer + Count)^, BytesCount, ByteOfData);
          Inc(Count, BytesCount);
        end else
        begin {** Byte is color data **}
        {** Simple copy the byte **}
          Byte(Pointer(ImagePointer + Count)^) := ByteOfData ;
          Inc(Count);
        end;{if else}

     {** Prepare for next step (byte) **}
       Inc(RleDataPointer);
     end;{while Count < ImageLen do}
  End;{INTERNAL Procedure ConvertRleToBmp}

  Procedure PrepareHeaders;
  Begin
  {** Calculate Boundary Size **}
    BoundarySize := pcxWidth mod 4;
    if BoundarySize <> 0
     then BoundarySize := 4 - BoundarySize;

  {** Prepare Bitmap Info Header **}
    with ABitmapInfoHeader, APcxFileRec.pcxHeader do
     begin
       biSize          := SizeOf(ABitmapInfoHeader);
       biWidth         := pcxWidth;
       biHeight        := -pcxHeight;     {** Top-Down Bitmap **}
       biPlanes        := NumPlanes;
       biBitCount      := BitPerPixel;
       biCompression   := BI_RGB;         {** Not Compressed Bitmap **}
       biSizeImage     := 0;
       biXPelsPerMeter := ScanHRes;
       biYPelsPerMeter := ScanVRes;
       biClrUsed       := 0;
       biClrImportant  := 0;              {** All colors are important **}
     end;{with ABitmapInfoHeader do}

   {** Prepare Bitmap File Header **}
    with ABitmapFileHeader, APcxFileRec.pcxHeader, ABitmapInfoHeader do
     begin
       bfType      := bmpFileSignature;
       bfOffBits   := 1024 + SizeOf(ABitmapFileHeader) + SizeOf(ABitmapInfoHeader);
       bfSize      := bfOffBits + (Abs(biHeight) * (biWidth + BoundarySize));
       bfReserved1 := 0;
       bfReserved2 := 0;
     end;{with ABitmapFileHeader, APcxFileRec.pcxHeader, ABitmapInfoHeader do}
  End;{INTERNAL Procedure PrepareHeaders}

  Procedure WriteToStream;
  Var
    i : integer;
  Begin
  {** Write To ABmpStream **}
    with ABmpStream, APcxFileRec do
     begin
     {** Resize & Reset Stream **}
       SetSize(ABitmapFileHeader.bfOffBits + ABitmapFileHeader.bfSize);
       Seek(0, soFromBeginning);

     {** Write Headers **}
       Write(ABitmapFileHeader, SizeOf(ABitmapFileHeader));
       Write(ABitmapInfoHeader, SizeOf(ABitmapInfoHeader));

     {** Write RGB Palette **}
       for i := 0 to 255 do
        with pcxPalette[i] do
         begin
           Write(Blue   , 1);
           Write(Green  , 1);
           Write(Red    , 1);
           Write(Bounder, 1); {** Filler **}
         end;{with pcxPalette[i] do}

     {** Write All Lines to Stream **}
       ImagePointer := Integer(ptrImage);
       with pcxHeader do
        for i := imgTop to imgBottom do
         begin
         {** Write pixels data (Full Line) **}
           Write(Pointer(ImagePointer)^, BytePerLine);
         {** Aditional Zeros For 32 bit Alignment **}
           Write(Bounder, BoundarySize);
         {** Next Line Offset **}
           Inc(ImagePointer, BytePerLine);
         end;{for i := ...}

     {** Reset Stream **}
       Seek(0, soFromBeginning);
     end;{with ABmpStream, APcxFileRec do}
  End;{INTERNAL Procedure WriteToStream}

Begin
{** Load PCX file Header **}
  APcxStream.Seek(0, soFromBeginning);
  APcxStream.Read(APcxFileRec.pcxHeader, SizeOf(APcxFileRec.pcxHeader));
  with APcxFileRec.pcxHeader do
   begin
     pcxWidth  := ((imgRight  - imgLeft) + 1);
     pcxHeight := ((imgBottom - imgTop ) + 1);
   end;{with APcxFileRec.pcxHeader do}

  with APcxFileRec do
   begin
   {** Prepare Memory for RLE **}
     pcxRleDataSize := APcxStream.Size - SizeOf(pcxHeader) - SizeOf(pcxPalette);
     GetMem(pcxRleData, pcxRleDataSize);
     try
     {** Load RLE Data **}
       APcxStream.Read(pcxRleData^ , pcxRleDataSize);
     {** Load PCX file Palette **}
       APcxStream.Read(pcxPalette, SizeOf(pcxPalette));;

     {** Prepare Memory for BMP Data **}
       ImageLen := pcxWidth * pcxHeight;
       GetMem(ptrImage, ImageLen);
       try
         ConvertRleToBmp;
         PrepareHeaders;
         WriteToStream;
       finally
         FreeMem(ptrImage, ImageLen);
       end;{try..finally}

     finally
       FreeMem(pcxRleData, pcxRleDataSize);
     end;{try..finally}
   end;{with APcxFileRec do}

End;{Procedure pcxStreamConvert}

{*****************************************}
{** Load PCX from hIstance or From File **}
{*****************************************}
Function pcxLoadAsBitmap(Const AName : String) : Graphics.TBitMap;
Var
  ResHandle : THandle;
  ResPtr    : PByte;
  ResSize   : LongInt;
  MemHandle : THandle;
{** Streams For Convertion **}
  msPcxImage : TMemoryStream;
  msBmpImage : TMemoryStream;
Begin
//  Result := nil;
  ResHandle := FindResource(hInstance, PChar(AName), pcxResTypeName);
  try
  {** if No Resource - Try Load From File ... **}
    if ResHandle = 0 then
     begin
       Result := pcxLoadAsBitmapFromFile(AName);
       Exit;
     end;{if ResHandle = 0 then}

  {** Resource Routine **}
    MemHandle := LoadResource(hInstance, ResHandle);
    ResPtr    := LockResource(MemHandle);
    ResSize   := SizeOfResource(hInstance, ResHandle);
  {** Load PCX Resource to Memory Stream **}
    msPcxImage := TMemoryStream.Create;
    msBmpImage := TMemoryStream.Create;
    try
      with msPcxImage do
       begin
       {** Load Resource to Stream **}
         SetSize(ResSize);
         Write(ResPtr^, ResSize);
       {** Free Resource **}
         FreeResource(MemHandle);
       {** Convert Stream Routine **}
         pcxStreamConvert(msPcxImage, msBmpImage);
         Result := Graphics.TBitmap.Create;
         Result.LoadFromStream(msBmpImage);
       end;{with PcxStream do}
    finally
      msBmpImage.Free;
      msPcxImage.Free;
    end;{try..finally}
  except
    Raise Exception.CreateFmt(strErrFromResource, [AName]);
  end;{try..except}
End;{Function pcxLoadAsBitmap}

{************************}
{** Load PCX from File **}
{************************}
Function pcxLoadAsBitmapFromFile(Const AFileName : String) : Graphics.TBitMap;
Var
  fsPcxFile  : TFileStream;
  msBmpImage : TMemoryStream;
Begin
//  Result := nil;
  try
    fsPcxFile  := TFileStream  .Create(AFileName, fmOpenRead or fmShareDenyNone);
    msBmpImage := TMemoryStream.Create;
    try
    {** Convert Stream Routine **}
      pcxStreamConvert(fsPcxFile, msBmpImage);
      Result := Graphics.TBitmap.Create;
      Result.LoadFromStream(msBmpImage);
    finally
      msBmpImage.Free;
      fsPcxFile .Free;
    end;{try..finally}
  except
    Raise Exception.CreateFmt(strErrFromFile, [AFileName]);
  end;{try .. except}
End;{Function pcxLoadAsBitmapFromFile}

{**************************}
{** Load PCX from Stream **}
{**************************}
Function pcxLoadAsBitmapFromStream(AStream : TStream) : Graphics.TBitMap;
Var
  msBmpImage : TMemoryStream;
Begin
//  Result := nil;
  try
    msBmpImage := TMemoryStream.Create;
    try
   {** Convert Stream Routine **}
      pcxStreamConvert(AStream, msBmpImage);
      Result := Graphics.TBitmap.Create;
      Result.LoadFromStream(msBmpImage);
    finally
      msBmpImage.Free;
    end;{try..finally}
  except
    Raise Exception.Create(strErrFromStream);
  end;{try .. except}
End;{Function pcxLoadAsBitmapFromStream}

{*****************************************************}
{** Load PCX from Stream (From Position and By Size)**}
{*****************************************************}
Function pcxLoadAsBitmapFromStreamDirectly(AStream   : TStream;
                                           APosition : LongInt;
                                           ASize     : LongInt) : Graphics.TBitMap;
Var
  msPcxImage  : TMemoryStream;
  OldPosition : LongInt;
Begin
//  Result := nil;
  msPcxImage := TMemoryStream.Create;
  try
  {** Prepare Stream for Loading **}
    OldPosition := AStream.Position;
    AStream.Position := APosition;
    msPcxImage.SetSize(ASize);
    msPcxImage.CopyFrom(AStream, ASize);
    AStream.Position := OldPosition;
  {** Load From Stream **}
    Result := pcxLoadAsBitmapFromStream(msPcxImage);
  finally
    msPcxImage.Free;
  end;{try..finally}
End;{Function pcxLoadAsBitmapFromStreamDirectly}


{##############################################################################}
{## Windows API PCX Routines ##################################################}
{##############################################################################}
{*****************************}
{** Load PCX file to memory **}
{*****************************}
Function LoadPcxFile(Const AFileName    : String;
                     Var   APcxFileRec  : TPcxFileRecord) : Boolean;
Var
  fsPcxFile : TFileStream;
Begin
  fsPcxFile := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyNone);
  try
    with fsPcxFile, APcxFileRec do
     begin
     {** Load PCX file Header **}
       Read(pcxHeader, SizeOf(pcxHeader));
     {** Get mem and load RLE data **}
       pcxRleDataSize := Size - SizeOf(pcxHeader) - SizeOf(pcxPalette);
       GetMem(pcxRleData, pcxRleDataSize);
       Read(pcxRleData^, pcxRleDataSize);
     {** Load PCX file Palette **}
       Read(pcxPalette, SizeOf(PcxPalette));

       Result := True;
     end;{with fsPcxFile, APcxFile do}
  finally
    fsPcxFile.Free;
  end;{try..finally}
End;{Function LoadPcxFile}

{***************************************}
{** Convert Loaded PCX file to Bitmap **}
{***************************************}
Function ConvertPcxToBitmap(APcxFileRec  : TPcxFileRecord;
                            Var ABitmap  : HBitmap;
                            Var APalette : HPalette) : Boolean;

Var
  RleDataPointer,
  ImagePointer                : LongInt; {** Used like pointers **}
  ImageLen,
  Count,
  BytesCount                  : LongInt;
  Image                       : Pointer;
  ByteOfData                  : Byte;
  i                           : integer;

  BitmapRecord                : Windows.TBITMAP;
  PaletteRecord               : Windows.TLOGPALETTE;
  PalEntries                  : Array[0..255] of Windows.TPALETTEENTRY;

Begin

  with APcxFileRec, APcxFileRec.PcxHeader do
   begin
   {** Get memory for Image = Width * Height by BPP **}
    { ImageLen  := (HRes * VRes * BitPerPixel + 7) div 8; {}
     ImageLen  := (HRes * VRes); {}
     GetMem(Image, ImageLen);

   {** Convert RLE Data To Image Data (Bitmap format) (Word Aligned Width ?) **}
     ImagePointer    := Integer(Image);
     RleDataPointer  := Integer(PcxRleData);
     Count := 0;
     while Count <= ImageLen do
      begin
      {** Read One Byte of Data **}
        ByteOfData := Byte(Pointer(RleDataPointer)^);
      {** Check This Byte **}
        if ByteOfData >= 192 then
         begin {** Byte is technickal = (Length of bytes row + 192) **}
           BytesCount := ByteOfData - 192;
         {** Read next byte - Real value of color **}
           Inc(RleDataPointer);
           ByteOfData := Byte(Pointer(RleDataPointer)^);
         {** Fill row of bytes **}
           FillChar(Pointer(ImagePointer + Count)^, BytesCount, ByteOfData);
           Inc(Count, BytesCount);
         end else
         begin {** Byte is color data **}
         {** Simple copy the byte **}
           Byte(Pointer(ImagePointer + Count)^) := ByteOfData ;
           Inc(Count);
         end;{if else}
      {** Prepare for next step (byte) **}
        Inc(RleDataPointer);
      end;{while Count < ImageLen do}

   {** Prepare Bitmap info Record and Create Bitmap **}
     with BitmapRecord do
      begin
        bmType        := 0;
        bmWidth       := HRes;
        bmHeight      := VRes;
        bmWidthBytes  := ((HRes + 1) div 2) * 2; {** Word Aligned Width **}
        bmPlanes      := NumPlanes;
        bmBitsPixel   := BitPerPixel;            {** 256 colors **}
        bmBits        := Image;
      end;{with BitmapRecord do}
    ABitmap := CreateBitmapIndirect(BitmapRecord);

   {** Prepare Palette info Record and Create Palette **}
    for i := 0 to 255 do
     with pcxPalette[i], PalEntries[i] do
      begin
        peRed    := Red;
        peGreen  := Green;
        peBlue   := Blue;
        peFlags  := 0;
      end;{with pcxPalette[i], PalEntries[i] do}
    with PaletteRecord do
     begin
       palVersion     := $300;
       palNumEntries  := 256;
       palPalEntry[0] := PalEntries[0];
     end;{with PaletteRecord do}
    APalette := CreatePalette(PaletteRecord);

   end;{with AImage, AImage.PcxHeader do}

  Result := True;
End;{Function ConvertPcxToBitmap}

{**********************************************************}
{** Load Pcx File and Convert Data to Bitmap and Palette **}
{**********************************************************}
Function LoadPcxFileToBitmap(Const AFileName : String;
                             Var ABitmap     : HBitmap;
                             Var APalette    : HPalette) : Boolean;
Var
  APcxFileRecord : TPcxFileRecord;
Begin
  Result := False;
{** Load PCX File Data **}
  if not LoadPcxFile(AFileName, APcxFileRecord)
   then Exit;
{** Convert Data to Bitmap and Palette **}
  try
    Result := ConvertPcxToBitmap(APcxFileRecord, ABitmap, APalette);
  finally
    FreeMem(APcxFileRecord.PcxRleData, APcxFileRecord.PcxRleDataSize);
  end;{try..finally}
End;{Function LoadPcxFileToBitmap}


{##############################################################################}
{******************************************************************************}
{##############################################################################}
Initialization
  RegisterClass(TPcx);
{$IfDef AntUnits_UsePcx}
  TPicture.RegisterFileFormat('Pcx', 'Pcx Image File', TPcx);
{$EndIf AntUnits_UsePcx}

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Finalization
{$IfDef Delphi3andHigher}
  {$IfDef AntUnits_UsePcx}
  TPicture.UnregisterGraphicClass(TPcx);
  {$EndIf AntUnits_UsePcx}
{$EndIf Delphi3andHigher}

End{Unit AntPcx}.
