{******************************************************************************}
{**                                                                          **}
{**  Windows Regions routines.                                               **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**  Copyright  1995 Borland Inerternational                                **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntRgn;

{$Include AntUnits.Inc}

Interface

Uses
   Windows, Controls, Graphics, Forms;

{*******************************************************************************
   Function return a Region (Result) from Mask bitmap (ABitmap). Key is a color
value (AColor). Region contain all pixels with color = AColor if AInverted True
or all other Pixels in other case.
   WARNING !!! There are a some problems with AColor = clWhite, becouse function
use bitmap masking.
*******************************************************************************}
Function krpBitmapToRegion(Const ABitmap : TBitmap;
                           AColor        : TColor;
                           AInverted     : Boolean) : HRgn;

{*******************************************************************************
  Function is like krpBitmapToRegion but works with any type of Graphics
*******************************************************************************}
Function krpGraphicToRegion(Const AGraphic : TGraphic;
                            AColor         : TColor;
                            AInverted      : Boolean) : HRgn;

{*******************************************************************************
  Function is like krpGraphicToRegion but use default Transparent Color for Key
*******************************************************************************}
Function krpGraphicToRegionColorDefault(Const AGraphic : TGraphic;
                                        AInverted      : Boolean) : HRgn;

{*******************************************************************************
  Function return a Rect of Form's Caption depend Form's BorderStyle
*******************************************************************************}
Function krpGetCaptionRect(Const AForm : TForm) : TRect;

{*******************************************************************************
  Function return a Region of Form Caption's Rect
*******************************************************************************}
Function krpGetCaptionRegion(Const AForm : TForm) : HRgn;

{*******************************************************************************
  Function set Control's bounds Region's bounds
*******************************************************************************}
Function krpBoundsToRegion(Const AControl : TControl;
                           Const ARgn     : HRgn) : Boolean;

{*******************************************************************************
  Function apply Region to WinControl and set bounds Region's bounds
*******************************************************************************}
Function krpApplyRgnToControl(Const AControl : TWinControl;
                              Const ARgn     : HRgn) : Boolean;


{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}
Uses
{$IfDef DEBUG}
  Dialogs,
{$EndIf DEBUG}
{$IfDef Delphi4andHigher}
  Jpeg,  
{$EndIf Delphi4andHigher}
  Classes, SysUtils;


Function krpBitmapToRegion(Const ABitmap : TBitmap;
                           AColor        : TColor;
                           AInverted     : Boolean) : HRgn;
Var
  ABMP          : TBitmap;
  ARgnData      : PRgnData;
  AGlobal       : HGlobal;
  NumberOfRects : Cardinal;

  Procedure _RectsToRegion;
  Var
    ARgn : HRgn;
  Begin
    ARgn := ExtCreateRegion(nil, SizeOf(TRgnDataHeader) + (SizeOf(TRect) * ARgnData^.rdh.nCount), ARgnData^);
    if Result <> 0 then
     begin  {** Region contains more then 2000 rects, so add new part **}
{$IfDef DEBUG}
       ShowMessage('DEBUG at krpBitmapToRegion :  Region contains more then 2000 rects. It''s can be a problem under Windows95');
{$EndIf DEBUG}
       CombineRgn(Result, Result, ARgn, rgn_OR);
       DeleteObject(ARgn);
     end
    else Result := ARgn;
  End;{Internal Procedure _RectsToRegion}

  Procedure _ResetRgnData;
  Begin
    with ARgnData^.rdh do
     begin
       dwSize   := SizeOf(TRgnDataHeader);
       iType    := RDH_Rectangles;
       nCount   := 0;
       nRgnSize := 0;
       rcBound  := Rect(MaxInt, MaxInt, 0, 0);
     end;{with ARgnData.rdh do}
  End;{Internal Procedure _ResetRgnData}

  Procedure _AddRectToRegion(ARect : TRect);
  Begin
  {** Check for big number of Rects **}
    if ARgnData^.rdh.nCount >= 2000 then
     begin
       _RectsToRegion;
       _ResetRgnData;
     end;{if ARgnData^.rdh.nCount >= 2000 then}

  {** Check for memory **}
    if ARgnData^.rdh.nCount >= NumberOfRects then
     begin
       GlobalUnlock(AGlobal);
       Inc(NumberOfRects, 200);
       AGlobal := GlobalReAlloc(AGlobal,
                                SizeOf(TRgnDataHeader) + SizeOf(TRect) * NumberOfRects,
                                GMem_Moveable);
       ARgnData := GlobalLock(AGlobal)
     end;{if ARgnData^.rdh.nCount >= NumberOfRects then}

  {** Fill RgnData data and fix rcBound **}
    with ARgnData^.rdh, ARgnData^.rdh.rcBound do
     begin
       PRect(@ARgnData^.Buffer[nCount * SizeOf(TRect)])^ := ARect;

       if ARect.Left   < Left
        then Left   := ARect.Left;
       if ARect.Top    < Top
        then Top    := ARect.Top;
       if ARect.Right  > Right
        then Right  := ARect.Right;
       if ARect.Bottom > Bottom
        then Bottom := ARect.Bottom;

       Inc(nCount);
     end;{with ARgnData^.rdh, ARgnData^.rdh.rcBound do}
  End;{Internal Procedure _AddRectToRegion}

  Procedure _BitmapToRects;
  Var
    X, Y          : Integer;
    Pixel         : PByte;
    ARect         : TRect;
    PixelInRegion : Boolean;
  Begin
    with ABMP, ARect do
     for Y:=0 to Height-1 do
      begin
        Pixel  := ScanLine[Y];
        Left   := -1;
        Top    := Y;
        Bottom := Y + 1;

      {** Run thru line and make rects **}
        for X:=0 to Width-1 do
         begin
           PixelInRegion := (Pixel^ <> 0);
           if AInverted
            then PixelInRegion := not PixelInRegion;

           if PixelInRegion then {** Pixel in region **}
            begin
              if Left < 0 then      {** New rect **}
               begin
                 Left  := X;
                 Right := X + 1;
               end
              else Right := X + 1;  {** Continue prev rect **}

            end else             {** Pixel outside region **}
            begin
              if Left >= 0 then     {** End of rect, add it to region **}
               begin
                 _AddRectToRegion(ARect);
                 Left := -1;
               end;
            end;{if Pixel^ = 0 else}

           Inc(Integer(Pixel), 1);
         end;{for X:=0 to Width-1 do}

      {** Check for ending rect **}
        if Left >= 0
         then _AddRectToRegion(ARect);

      end;{for Y:=0 to Height-1 do}
  End;{Internal Procedure _BitmapToRects}

Begin
  Result := 0;
  NumberOfRects := 200;

  if not Assigned(ABitmap)
   then Exit;

  ABMP := Graphics.TBitmap.Create;
  with ABMP do
   try
     Assign(ABitmap);
     Mask(AColor);
     HandleType  := bmDIB;
   {** Better use pf1bit - it eats low memory, but needs more coding for bits.
       I will change it to 1bit bitmap latter **}
     PixelFormat := pf8bit;

   {** Create Region **}
     AGlobal := GlobalAlloc(GMem_Moveable,
                            SizeOf(TRgnDataHeader) + SizeOf(TRect) * NumberOfRects);
     try
       ARgnData := GlobalLock(AGlobal);
       _ResetRgnData;
       _BitmapToRects;
       _RectsToRegion;
     finally
       GlobalFree(AGlobal);
     end;{try..finally}

   finally
     Free;
{$IfDef DEBUG}
     if Result = 0
      then ShowMessage('DEBUG at krpBitmapToRegion :  Result = 0, Creating of region failed');
{$EndIf DEBUG}
   end;{with ABMP do try..}

End;{Function krpBitmapToRegion}


Function krpGraphicToRegion(Const AGraphic : TGraphic;
                            AColor         : TColor;
                            AInverted      : Boolean) : HRgn;
Var
  ABMP : TBitmap;
Begin
  Result := 0;

  if not Assigned(AGraphic)
   then Exit;

  if (AGraphic is TBitmap) then
   begin
     Result := krpBitmapToRegion(TBitmap(AGraphic), AColor, AInverted);
     Exit;
   end;{if (AGraphic is TBitmap) then}

  ABMP := Graphics.TBitmap.Create;
  with ABMP do
   try
     Width  := AGraphic.Width;
     Height := AGraphic.Height;
     HandleType  := bmDIB;
{$IfDef Delphi4andHigher}
   {** Aproximation for JPEG **}
     if (AGraphic is TJPEGImage)
      then PixelFormat := pf8bit;
{$EndIf Delphi4andHigher}

     ABMP.Canvas.Draw(0, 0, AGraphic);

     Result := krpBitmapToRegion(ABMP, AColor, AInverted);
   finally
     Free;
   end;{with ABMP do try..}
End;{Function krpGraphicToRegion}


Function krpGraphicToRegionColorDefault(Const AGraphic : TGraphic;
                                        AInverted      : Boolean) : HRgn;
Var
  ABMP : TBitmap;
Begin
  Result := 0;

  if not Assigned(AGraphic)
   then Exit;

  if (AGraphic is TBitmap) then
   begin
     Result := krpBitmapToRegion(TBitmap(AGraphic),
                                 TBitmap(AGraphic).TransparentColor, AInverted);
     Exit;
   end;{if (AGraphic is TBitmap) then}

  ABMP := Graphics.TBitmap.Create;
  with ABMP do
   try
     Width  := AGraphic.Width;
     Height := AGraphic.Height;
     HandleType  := bmDIB;
{$IfDef Delphi4andHigher}
   {** Aproximation for JPEG **}
     if (AGraphic is TJPEGImage)
      then PixelFormat := pf8bit;
{$EndIf Delphi4andHigher}

     Canvas.Draw(0, 0, AGraphic);
     TransparentMode := tmAuto;

     Result := krpBitmapToRegion(ABMP, TransparentColor, AInverted);
   finally
     Free;
   end;{with ABMP do try..}
End;{Function krpGraphicToRegion}


Function krpGetCaptionRect(Const AForm : TForm) : TRect;
Var
  DeltaX, DeltaY, CaptionHeight : Integer;
Begin
  FillChar(Result, SizeOf(Result), 0);
  DeltaX        := 0;
  DeltaY        := 0;
  CaptionHeight := 0;

  if not Assigned(AForm)
   then Exit;

  with AForm do
   begin
     if (csDesigning in ComponentState) then
      begin
      {** At design time caption depend no of BorderStyle **}
        DeltaX        := GetSystemMetrics(SM_CXSizeFrame);
        DeltaY        := GetSystemMetrics(SM_CYSizeFrame);
        CaptionHeight := GetSystemMetrics(SM_CYCaption);
      end else
      begin
      {** Calculate coordinates depend of BorderStyle **}
        case BorderStyle of
         bsSingle : begin
           DeltaX        := GetSystemMetrics(SM_CXFixedFrame);
           DeltaY        := GetSystemMetrics(SM_CYFixedFrame);
           CaptionHeight := GetSystemMetrics(SM_CYCaption);
         end;{bsSingle : begin}

         bsSizeable : begin
           DeltaX        := GetSystemMetrics(SM_CXSizeFrame);
           DeltaY        := GetSystemMetrics(SM_CYSizeFrame);
           CaptionHeight := GetSystemMetrics(SM_CYCaption);
         end;{bsSizeable : begin}

         bsDialog : begin
           DeltaX        := GetSystemMetrics(SM_CXDlgFrame);
           DeltaY        := GetSystemMetrics(SM_CYDlgFrame);
           CaptionHeight := GetSystemMetrics(SM_CYCaption);
         end;{bsDialog : begin}

         bsToolWindow : begin
           DeltaX        := GetSystemMetrics(SM_CXFixedFrame);
           DeltaY        := GetSystemMetrics(SM_CYFixedFrame);
           CaptionHeight := GetSystemMetrics(SM_CYSmCaption);
         end;{bsToolWindow : begin}

         bsSizeToolWin : begin
           DeltaX        := GetSystemMetrics(SM_CXSizeFrame);
           DeltaY        := GetSystemMetrics(SM_CYSizeFrame);
           CaptionHeight := GetSystemMetrics(SM_CYSmCaption);
         end;{bsSizeToolWin : begin}
        end;{case BorderStyle of}

      end;{if (csDesigning in ComponentState) else}

     Result := Rect(0, 0, Width, Height);
     InflateRect(Result, -DeltaX, -DeltaY);
     Result.Bottom := Result.Top + CaptionHeight - 1;

   end;{with AForm do}
End;{Function krpGetCaptionRect}


Function krpGetCaptionRegion(Const AForm : TForm) : HRgn;
Begin
  Result := CreateRectRgnIndirect(krpGetCaptionRect(AForm));
End;{Function krpGetCaptionRegion}


Function krpBoundsToRegion(Const AControl : TControl;
                           Const ARgn     : HRgn) : Boolean;
Var
  ARect : TRect;
Begin
  Result := False;

{** Check params **}
  if (not Assigned(AControl)) or (ARgn = 0)
   then Exit;

{** Get BoundsRect of region, if region is Empty - Exit **}
  if GetRgnBox(ARgn, ARect) = NULLREGION
   then Exit;

{** Apply BoundsRect of region to Control's Bounds **}
  with ARect do
   AControl.SetBounds(Left, Top, Right - Left, Bottom - Top);

  Result := True;
End;{Function krpBoundsToRegion}


Function krpApplyRgnToControl(Const AControl : TWinControl;
                              Const ARgn     : HRgn) : Boolean;
Var
  ARect   : TRect;
  ARegion : HRgn;
Begin
  Result := False;

{** Check params **}
  if (not Assigned(AControl)) or (ARgn = 0)
   then Exit;

{** Create empty Region **}
  ARegion := CreateRectRgn(0, 0, 0, 0);
  try
  {** Copy incoming region to  Region **}
    CombineRgn(ARegion, ARgn, ARegion, rgn_Copy);

    if GetRgnBox(ARegion, ARect) <> NULLREGION then
     begin
     {** Set control's bounds **}
       with ARect do
        AControl.SetBounds(Left, Top, Right - Left, Bottom - Top);

     {** Move Region to control's coords and set it to control **}
       with AControl do
        begin
          OffsetRgn(ARegion, -Left, -Top);
   //       SetWindowRgn(Handle, 0      , False);
          SetWindowRgn(Handle, ARegion, Visible);
        end;{with AControl do}

       Result := True;
     end;{if GetRgnBox(ARegion, ARect) <> NULLREGION then}

  finally
    DeleteObject(ARegion);
  end;{try..finally}
End;{Function krpApplyRgnToControl}


END{Unit AntRgn}.
