{******************************************************************************}
{**                                                                          **}
{**  Bios, Hardware and System routines.                                     **}
{**                                                                          **}
{**  Author:    KARPOLAN                                                     **}
{**  E-Mail:    karpolan@yahoo.com , karpolan@utilmind.com                   **}
{**  Home Page: http://karpolan.i.am, http://www.utilmind.com                **}
{**  Copyright  1996-99 by KARPOLAN.                                        **}
{**                                                                          **}
{******************************************************************************}
{**  History:                                                                **}
{**                                                                          **}
{**  31 aug 1999 - Last modified... Prepared for public release.             **}
{******************************************************************************}
Unit AntSys;

Interface

Uses
  SysUtils, Windows, Messages, Classes, Controls;

Const
  adr_BiosName      = $FE061;
  adr_BiosCopyright = $FE091;
  adr_BiosInfo      = $FEC71;
  adr_BiosDate      = $FFFF5;

{*******************}
{** Bios Routines **}
{*******************}
Function GetBiosName : String;
Function GetBiosCopyright : String;
Function GetBiosInfo : String;
Function GetBiosDate : TDateTime;

{*************************}
{** SystemInfo Routines **}
{*************************}
Function GetCurrentComputerName : String;
Function GetCurrentUserName : String;

Function GetDiskSerialNumber(Latter : Char) : DWord;
Function GetDiskSerialNumberStr(Latter : Char) : String;

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}

{*******************}
{** Bios Routines **}
{*******************}
Function GetBiosName : String;
Begin
  try
    Result := String(PChar(Ptr(adr_BiosName)));
  except
    Result := 'NoName';
  end;{try..except}
End;{Function GetBiosName}

Function GetBiosCopyright : String;
Begin
  try
    Result := String(PChar(Ptr(adr_BiosCopyright)));
  except
    Result := 'NoCopyright';
  end;{try..except}
End;{Function GetBiosCopyright}

Function GetBiosInfo : String;
Begin
  try
    Result := String(PChar(Ptr(adr_BiosInfo)));
  except
    Result := '';
  end;{try..except}
End;{Function GetBiosInfo}

Function GetBiosDate : TDateTime;
Var
  RegStr, RegFormat : String;
  RegSeparator      : Char;
Begin
  Result := 0;
  try
    RegStr := String(PChar(Ptr(adr_BiosDate)));
  except
    Exit;
  end;{try..except}

  RegFormat := ShortDateFormat;
  RegSeparator := DateSeparator;
  try
    DateSeparator   := '/';
    try
      ShortDateFormat := 'm/d/y';
      Result := StrToDate(RegStr);
    except
      try
        ShortDateFormat := 'y/m/d';
        Result := StrToDate(RegStr);
      except
      end;
    end;
  finally
    ShortDateFormat := RegFormat;
    DateSeparator   := RegSeparator;
  end;{try..finally}
End;{Function GetBiosDate}


{*************************}
{** SystemInfo Routines **}
{*************************}
Function GetCurrentComputerName : String;
Var
  Name : PChar;
  Size : DWord;
Begin
  Size := MAX_COMPUTERNAME_LENGTH + 1;
  GetMem(Name, Size);
  try
    GetComputerName(Name, Size);
    Result := Trim(StrPas(Name));
  finally
    FreeMem(Name, Size);
  end;{try..finally}
End;{Function GetCurrentComputerName}

Function GetCurrentUserName : String;
Var
  Name : PChar;
  Size : DWord;
Begin
  Size := SizeOf(ShortString);
  GetMem(Name, Size);
  try
    GetUserName(Name, Size);
    Result := Trim(StrPas(Name));
  finally
    FreeMem(Name, Size);
  end;{try..finally}
End;{Function GetCurrentUserName}

Function GetDiskSerialNumber(Latter : Char) : DWord;
Var
  MaxFilenameLength,
  FileSystemFlags    : DWord;
Begin
  GetVolumeInformation(PChar(Latter + ':\'), nil, 0,
                       @Result,
                       MaxFilenameLength,
                       FileSystemFlags, nil, 0);
End;{Function GetDiskSerialNumber}

Function GetDiskSerialNumberStr(Latter : Char) : String;
Var
  RegDWord : DWord;
Begin
  RegDWord := GetDiskSerialNumber(Latter);
  Result := IntToHex(HiWord(RegDWord), 4) + '-' +
            IntToHex(LoWord(RegDWord), 4);
End;{Function GetDiskSerialNumber}


END{Unit AntSys}.
