// -----------------------------------------------------------------------------
// DogsEar.Pas                Copyright (c) 1996-99 DithoSoft Software Solutions
// Version 1.5                                           http://www.dithosoft.de
// -----------------------------------------------------------------------------
// The TDogsEar component will help you create applications looking like books.
// The component displays a little folded paper edge that fires events when
// klicked.
// -----------------------------------------------------------------------------
unit DogsEar;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TDogsEarAlign      = (adLeft, adRight);       
  TDogsEarFrameStyle = (dfFrame, dfNoFrame);    
  TDogsEarNumPages   = 1..MaxInt;               
  TDogsEarNotify     = procedure (Sender : TObject; NumPages : TDogsEarNumPages) of Object;
  TDogsEarPoly = Array[0..3] of TPoint;



  TDogsEar = class(TGraphicControl)
  private
    FAlignDogsEar     : TDogsEarAlign;
    FColorBackground,
    FColorDogsEar,
    FColorFrame       : TColor;
    FFrameStyle       : TDogsEarFrameStyle;
    FMultiPage        : Boolean;
    FNumPages         : TDogsEarNumPages;
    FClickEvent       : TDogsEarNotify;
  private
    procedure SetAlign(Align : TDogsEarAlign);
    procedure SetColorBackground(Color : TColor);
    procedure SetColorDogsEar(Color : TColor);
    procedure SetColorFrame(Color : TColor);
    procedure SetFrameStyle(Style : TDogsEarFrameStyle);
    procedure SetMultiPage(MP : Boolean);
    procedure SetNumPages(Pages : TDogsEarNumPages);

  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    function  CalcSizedPoly(Source : TDogsEarPoly) : TDogsEarPoly;

  public
    constructor Create(AOwner : TComponent); override;

  published
    property AlignDogsEar : TDogsEarAlign read FAlignDogsEar write SetAlign;
    property ColorBackground : TColor read FColorBackground write SetColorBackground;
    property ColorDogsEar : TColor read FColorDogsEar write SetColorDogsEar;
    property ColorFrame : TColor read FColorFrame write SetColorFrame;
    property Enabled;
    property FrameStyle : TDogsEarFrameStyle read FFrameStyle write SetFrameStyle;
    property MultiPage : Boolean read FMultiPage write SetMultiPage;
    property NumPages : TDogsEarNumPages read FNumPages write SetNumPages;
    property OnDogsEarClick : TDogsEarNotify read FClickEvent write FClickEvent;
    property Visible;
  end;

procedure Register;

implementation


const
 LargeDogsEarLeft  : TDogsEarPoly = ((x:0;y:0),
                                     (x:4;y:0),
                                     (x:4;y:4),
                                     (x:0;y:0));
 SmallDogsEarLeft  : TDogsEarPoly = ((x:0;y:0),
                                     (x:3;y:1),
                                     (x:4;y:4),
                                     (x:0;y:0));

 LargeDogsEarRight : TDogsEarPoly = ((x:0;y:0),
                                     (x:4;y:0),
                                     (x:0;y:4),
                                     (x:0;y:0));
 SmallDogsEarRight : TDogsEarPoly = ((x:1;y:1),
                                     (x:4;y:0),
                                     (x:0;y:4),
                                     (x:1;y:1));


procedure Register;
begin
  RegisterComponents('Andere', [TDogsEar]);
end;

constructor TDogsEar.Create(AOwner : TComponent);
begin
 Inherited Create(AOwner);

 FAlignDogsEar     := adLeft;
 FColorBackground  := clBtnFace;
 FColorDogsEar     := clWindow;
 FColorFrame       := clWindowFrame;
 FFrameStyle       := dfFrame;
 FMultiPage        := False;
 FNumPages         := 1;

 Width             := 30;
 Height            := 30;
end;

procedure TDogsEar.SetAlign(Align : TDogsEarAlign);
begin
 FAlignDogsEar := Align;
 Invalidate;
end;

procedure TDogsEar.SetColorBackground(Color : TColor);
begin
 FColorBackground := Color;
 Invalidate;
end;

procedure TDogsEar.SetColorDogsEar(Color : TColor);
begin
 FColorDogsEar := Color;
 Invalidate;
end;

procedure TDogsEar.SetColorFrame(Color : TColor);
begin
 FColorFrame := Color;
 Invalidate;
end;

procedure TDogsEar.SetFrameStyle(Style : TDogsEarFrameStyle);
begin
 FFrameStyle := Style;
 Invalidate;
end;

procedure TDogsEar.SetMultiPage(MP : Boolean);
begin
 FMultiPage := MP;
 Invalidate;
end;

procedure TDogsEar.SetNumPages(Pages : TDogsEarNumPages);
begin
 FNumPages := Pages;
 Invalidate;
end;

procedure TDogsEar.Paint;
var Poly,
    LargeDogsEar,
    SmallDogsEar  : TDogsEarPoly;
begin
 Canvas.Brush.Color := FColorBackground;

 if FFrameStyle = dfNoFrame then
  Canvas.Pen.Color := FColorBackground
 else
  Canvas.Pen.Color := FColorFrame;

 Canvas.Rectangle(0,0,ClientWidth,ClientHeight);

 if FAlignDogsEar = adLeft then begin
  LargeDogsEar := LargeDogsEarLeft;
  SmallDogsEar := SmallDogsEarLeft;
 end else begin
  LargeDogsEar := LargeDogsEarRight;
  SmallDogsEar := SmallDogsEarRight;
 end;

 Poly := CalcSizedPoly(LargeDogsEar);

 Canvas.Brush.Color := FColorDogsEar;
 Canvas.Pen.Color   := FColorFrame;
 Canvas.Polygon(Poly);

 if FMultiPage then begin
  Poly := CalcSizedPoly(SmallDogsEar);
  Canvas.Polygon(Poly);
 end;
end;

function TDogsEar.CalcSizedPoly(Source : TDogsEarPoly) : TDogsEarPoly;
begin
 Result[0].x := (Source[0].x*Pred(ClientWidth)) div 4;
 Result[0].y := (Source[0].y*Pred(ClientHeight)) div 4;
 Result[1].x := (Source[1].x*Pred(ClientWidth)) div 4;
 Result[1].y := (Source[1].y*Pred(ClientHeight)) div 4;
 Result[2].x := (Source[2].x*Pred(ClientWidth)) div 4;
 Result[2].y := (Source[2].y*Pred(ClientHeight)) div 4;
 Result[3].x := (Source[3].x*Pred(ClientWidth)) div 4;
 Result[3].y := (Source[3].y*Pred(ClientHeight)) div 4;
end;

procedure TDogsEar.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var Large,
    Small    : TDogsEarPoly;
    LargeRGN,
    SmallRGN : HRGN;
begin
 if not Assigned(FClickEvent) then Exit;

 if AlignDogsEar = adLeft then begin
  Large := CalcSizedPoly(LargeDogsEarLeft);
  Small := CalcSizedPoly(SmallDogsEarLeft);
 end else begin
  Large := CalcSizedPoly(LargeDogsEarRight);
  Small := CalcSizedPoly(SmallDogsEarRight);
 end;

 LargeRGN := CreatePolygonRgn(Large,4,WINDING);
 SmallRGN := CreatePolygonRgn(Small,4,WINDING);

 if (LargeRGN <> 0) and (SmallRGN <> 0) then begin
  if MultiPage then begin
   if (PtInRegion(SmallRGN,x,y)) then
    FClickEvent(Self,NumPages)
   else if (PtInRegion(LargeRGN,x,y)) then
    FClickEvent(Self,1);
  end else begin
   if (PtInRegion(LargeRGN,x,y)) then
    FClickEvent(Self,1);
  end;
 end;

 if LargeRGN <> 0 then DeleteObject(LargeRGN);
 if SmallRGN <> 0 then DeleteObject(SmallRGN);
end;


end.
