////////////////////////////////////////////////////////////////////////////////
// Computation Threads
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit HITComputationThreads;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes;

type
  THITComputationThread = class(TThread)
  private
    fExceptMsg: string;
    fExceptName: string;
    fExceptRaised: Boolean;
  protected
    procedure Computation; virtual; abstract;
    procedure Execute; override;
  public
    property ExceptMsg: string read fExceptMsg;
    property ExceptName: string read fExceptName;
    property ExceptRaised: Boolean read fExceptRaised;
  end;

implementation
//==============================================================================
{$C-,G-,O+,Q-,R-}
//==============================================================================
uses SysUtils, Integers, GF2;

////////////////////////////////////////////////////////////////////////////////
// THITComputationThread class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
procedure THITComputationThread.Execute;
begin
  try
    try
      Computation;
    finally

      // Each thread has its own local variable banks in INTEGERS and in GF2.
      // These banks must be freed before the thread is destroyed, otherwise
      // they will stay in memory until the application terminates.

    //GF2UnitVarsFree; //To activate if GF2 routines are used
      IUnitVarsFree;
    end;
  except

    // Need to catch exceptions (otherwise DELPHI crashes).

    if not (ExceptObject is EAbort) then
    begin
      fExceptName := ExceptObject.ClassName;
      if ExceptObject is Exception then
        fExceptMsg := Exception(ExceptObject).Message;
      fExceptRaised := true;
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

