{*******************************************************}
{                                                       }
{  TDCCalendar Component                                }
{                                                       }
{  Copyright (c) 1999 Dream Company                     }
{  e-mail: contact@dream-com.com                        }
{     WWW: http://www.dream-com.com                     }
{                                                       }
{*******************************************************}

unit dcCalend;

interface
{$IFDEF WIN32}
{$I dc.inc}
{$ENDIF}
uses
  {$IFDEF WIN32} Windows, {$ELSE} WinTypes, WinProcs, {$ENDIF}
  Classes, Messages, SysUtils, Controls, Graphics, StdCtrls, ExtCtrls, Forms;

const elemYear = 1; elemMonth = 2; elemDay = 3;

type
  TCaSelectType = (stOneDay, stWeek, stWorkWeek);
  TCaOption = (caDrawFocus, caShowPrevMonth, caShowNextMonth, caShowToday,
    caShowWeekNum, caUseEllipticSelection, caUseShortTitle);
  TCaOptions = set of TCaOption;
  TCaSelArea = (saWhole, saChanges, saOldSelect, saFirstCell, saFirstLine,
    saLastCell, saLastLine);
  TChangeDateEvent = procedure(Sender: TObject; newDate: TDateTime) of object;
  TCloseUpEvent = procedure (Sender: TObject; Accept: Boolean) of object;

  TDCMSelScrollBar = class(TScrollBar)
  private
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
  protected
    procedure Scroll(ScrollCode: TScrollCode; var ScrollPos: integer); override;
  public
    constructor Create(aOwner: TComponent); override;
  end;

  TPopupSelector = class(TCustomControl)
  private
    FScrollBar: TDCMSelScrollBar;
    FTimer: TTimer;
    FDate: TDateTime;
    FCloseUp: TCloseUpEvent;
    FMonthWidth, FMonthHeight: integer;
    FTrackDelta: integer;

    function GetMonthArea: TRect;
    procedure SetMonth(Value: integer);
    procedure DoScrollMonthes(Sender: TObject); {raised on timer}

    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMMouseActivate(var Message: TMessage); message WM_MOUSEACTIVATE;
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  protected
    procedure SetDate(Value: TDateTime);
    function GetDateElement(Index: integer): integer;

    procedure CreateParams(var Params: TCreateParams); override;
    {$IFNDEF WIN32}
    procedure CreateWnd; override;
    {$ENDIF}
    procedure Paint; override;
    procedure KeyDown(var Key: word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;

    procedure UpdateDate; virtual;
    procedure EventFromScroller(sCode: TScrollCode);
    procedure PopupMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: integer);
    procedure CloseUp(Accept: Boolean); virtual;

    property Year: integer index elemYear read GetDateElement stored False;
    property Month: integer index elemMonth read GetDateElement stored False;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(aLeft, aTop, aWidth, aHeight: integer); override;

    procedure Show(AlignTo: TRect);
    procedure Hide;

    property OnCloseUp: TCloseUpEvent read FCloseUp write FCloseUp;
    property Date: TDateTime read FDate write SetDate stored False;
    property Color;
  published
  end;

  TDCCalendar = class(TCustomControl)
  private
    FPopup: TPopupSelector;
    PopupForm: TForm; 
    FDate, FOldDate, FAnchor: TDateTime;
    FColors: array [1..6] of TColor;
  { 1: text of selected cells
    2: background of selected cells
    3: text of title
    4: background of title
    5: text of other months day cells
    6: frame of today cell
    also Font.Color - text of main win &
         Color - background of main win }
    FOptions: TCaOptions;
    FSelectType: TCaSelectType;
    FAutoSize: Boolean;
    FMultiSelect: Boolean;

    FOnChange: TChangeDateEvent;
    EdgeDepth,
    FUnderMouseDay, {number of day under mouse pointer when left button is pressed}
    cellW, cellH,   {Width & Height of one cell}
    Offset1stDay,   {= number of prev. month days to be drawed}
    CurMonth1stWeekNum: integer; {number of 1st week of current month}
    FPopupVisible, FFocused, IsMonthChg, IsMouseDown, InSelectMode,
      ContinueSelect, InDialogMode: Boolean;

    function GetTitle: string;
    function DaysThisMonth: integer;
    function GetUnderMouseDay(X, Y: integer): integer;
    procedure SetUnderMouseDay(umDay: integer);
    procedure CalcAutoCellSize(var aWidth, aHeight: integer);
    function ShiftDate(Elem, Delta: integer): TDateTime;
    procedure MoveCursorToDateCell;

    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
  protected
    procedure SetDate(Value: TDateTime);
    function GetRangeDate(Index: integer): TDateTime;
    function GetDateElement(Index: integer): integer;
    procedure SetDateElement(Index: integer; Value: integer);
    function GetColors(Index: integer): TColor;
    procedure SetColors(Index: integer; Value: TColor);
    procedure SetSelectType(Value: TCaSelectType);
    procedure SetAutoSize(Value: Boolean);
    procedure SetMultiSelect(Value: Boolean);
    procedure SetOptions(Value: TCaOptions);

    function GetTitleRect: TRect;
    function GetHeaderRect: TRect;
    function GetPrevMonthRect: TRect;
    function GetNextMonthRect: TRect;
    function GetWeekNumRect: TRect;
    function GetDaysArea: TRect;
    function GetCellRect(aCol, aRow: integer): TRect;
    function GetDayRect(aDay: integer): TRect;
    function GetSelectedArea(aType: TCaSelArea): TRect; {selected cells (line)}
    function GetRightMarginRect: TRect;
    function GetBottomMarginRect: TRect;
    function GetToday: integer;
    procedure UpdateDate;
    procedure ChangeDate; virtual; {raise OnChange event}
    procedure PopupDropDown;
    procedure PopupCloseUp(Sender: TObject; Accept: Boolean);

    procedure CreateParams(var Params: TCreateParams); override;
    procedure Paint; override;
    procedure KeyDown(var Key: word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(aLeft, aTop, aWidth, aHeight: integer); override;

    function Execute(X, Y: integer): Boolean;

    property StartDate: TDateTime index 1 read GetRangeDate;
    property FinishDate: TDateTime index 2 read GetRangeDate;
  published
    property Date: TDateTime read FDate write SetDate;
    property Year: integer index elemYear read GetDateElement
      write SetDateElement stored False;
    property Month: integer index elemMonth read GetDateElement
      write SetDateElement stored False;
    property Day: integer index elemDay read GetDateElement
      write SetDateElement stored False;
    property SelFontColor: TColor index 1 read GetColors write SetColors
      default clHighlightText;
    property SelBackColor: TColor index 2 read GetColors write SetColors
      default clHighlight;
    property TitleFontColor: TColor index 3 read GetColors write SetColors
      default clWindowText;
    property TitleBackColor: TColor index 4 read GetColors write SetColors
      default clBtnFace;
    property OtherMonthColor: TColor index 5 read GetColors write SetColors
      default clGrayText;
    property TodayColor: TColor index 6 read GetColors write SetColors
      default clMaroon;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property Options: TCaOptions read FOptions write SetOptions
      default [caShowToday, caDrawFocus];
    property SelectType: TCaSelectType read FSelectType write SetSelectType
      default stOneDay;
    property MultiSelect: Boolean read FMultiSelect write SetMultiSelect
      default False;
    property OnChange: TChangeDateEvent read FOnChange write FOnChange;

    property Align;
    {$IFDEF D4}
    property Anchors;
    {$ENDIF}
    property Color default clWindow;
    property Cursor;
    property DragCursor;
    {$IFDEF D4}
    property DragKind;
    {$ENDIF}
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor default False;
    property ParentFont default True;
    property ParentShowHint default True;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;

    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    {$IFDEF D4}
    property OnEndDock;
    {$ENDIF}
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    {$IFDEF D4}
    property OnStartDock;
    property OnStartDrag;
    {$ENDIF}
  end;

procedure Register;

implementation

{$IFDEF WIN32}
{$R *.D32}
{$ELSE}
{$R *.D16}
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Dream Company', [TDCCalendar]);
end;

{*** Utilities ******************************************************}

const DaysPerWeek = 7; MaxWeeksPerMonth = 6 {1+4*7+1=30}; MonthesPerYear = 12;
      PopupDelta = 3 {3 prev., current & 3 next monthes in popup = 7 monthes};
      WrongUnderMouseDay = -100; {mouse is not over days area}
      LinePadding = 5; DragDelta = 6;

function Min(A, B: Integer): Integer;
begin
  if A < B then
    Result := A
  else
    Result := B;
end;

function MonthName(Short: Boolean; Index: integer): string;
begin
  if Short then
    Result := ShortMonthNames[Index]
  else
    Result := LongMonthNames[Index]
end;

function DaysPerMonth(aYear, aMonth: integer): integer;
const DaysInMonth: array [1..MonthesPerYear] of integer =
      (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
  {$IFNDEF D3}
  function IsLeapYear(aYear: Integer): Boolean;
  begin
    Result := (aYear mod 4 = 0) and ((aYear mod 100 <> 0) or (aYear mod 400 = 0));
  end;
  {$ENDIF}
begin
  Result := DaysInMonth[aMonth];
  if (aMonth = 2) and IsLeapYear(aYear) then
    Inc(Result); {leap-year Feb is special}
end;

procedure TextOutCentered(aCanvas: TCanvas; aText: string; aRect: TRect);
var tLeft, tTop: integer;
    {$IFDEF WIN32} tx: TSize; {$ELSE} tex: longint; {$ENDIF}
begin {suppose that brush & font colors are correct}
  with aRect do
  begin
    {$IFDEF WIN32}
    GetTextExtentPoint32(aCanvas.Handle, PChar(aText), Length(aText), tx);
    tLeft := (Right + Left - tx.cx) div 2;
    tTop := (Bottom + Top - tx.cy) div 2;
    {$ELSE}
    tex := GetTextExtent(aCanvas.Handle, @aText[1], Length(aText));
    tLeft := (Right + Left - LoWord(tex)) div 2;
    tTop := (Bottom + Top - HiWord(tex)) div 2;
    {$ENDIF}
    ExtTextOut(aCanvas.Handle, tLeft, tTop, ETO_OPAQUE, @aRect, @aText[1], 
      Length(aText), nil);
  end;
end;

procedure ShiftMonth(var aYear, aMonth: word; Delta: integer);
begin
  if aMonth + Delta < 1 then
  begin
    Dec(aYear);
    aMonth := MonthesPerYear + aMonth + Delta;
  end
  else if aMonth + Delta > MonthesPerYear then
  begin
    Inc(aYear);
    aMonth := aMonth + Delta - MonthesPerYear;
  end
  else
    aMonth := aMonth + Delta;
end;

{$IFDEF WIN32}
var WHook: HHook;
    dcCa: TDCCalendar;

function CallWndProcHook(nCode: integer; wParam: Longint;
  var mhs: TMouseHookStruct): longint; stdcall;
var rc, rp: TRect;
begin
  with dcCa do
    if (FPopup <> nil) and FPopupVisible then
    begin
      GetWindowRect(Handle, rc);
      GetWindowRect(FPopup.Handle, rp);
      if (((wParam = WM_LBUTTONDOWN) or (wParam = WM_LBUTTONUP)
         or (wParam = WM_NCLBUTTONDOWN) or (wParam = WM_NCLBUTTONUP))) and
         not (PtInRect(rc, mhs.pt) or PtInRect(rp, mhs.pt)) then
        PopupCloseUp(FPopup, False)
    end;
  Result := CallNextHookEx(WHook, nCode, wParam, Longint(@mhs));
end;
{$ENDIF}

{*** TDCMSelScrollBar ***********************************************}

constructor TDCMSelScrollBar.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  TabStop := False;
  Kind := sbVertical;
  Align := alRight;
  ControlStyle := ControlStyle - [csFramed];
  Min := 0;
  Max := 2 * PopupDelta;
  Position := PopupDelta;
  SmallChange := 1;
  LargeChange := 2 * SmallChange;
end;

procedure TDCMSelScrollBar.Scroll(ScrollCode: TScrollCode;
  var ScrollPos: integer);
begin
  ScrollPos := PopupDelta; {always middle position}
  TPopupSelector(Parent).EventFromScroller(ScrollCode);
end;

procedure TDCMSelScrollBar.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin {control is opaque}
end;

{ *** TPopupSelector ************************************************}

constructor TPopupSelector.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  Parent := TWinControl(aOwner);
  Left := 1000;
  Top := 1000;
  FScrollBar := TDCMSelScrollBar.Create(Self);
  FScrollBar.Parent := Self;
  FTimer := TTimer.Create(Self);
  with FTimer do
  begin
    FTimer.Enabled := False;
    FTimer.OnTimer := DoScrollMonthes;
  end;
  ControlStyle := ControlStyle +
    [csOpaque {$IFDEF WIN32}, csNoDesignVisible{$ENDIF}];
  Visible := False;
  OnMouseUp := PopupMouseUp;
  ParentFont := True;
  SetBounds(Parent.Left, Parent.Top, 0, 0); {width & height set automatically}
end;

destructor TPopupSelector.Destroy;
begin
  FTimer.Free;
  FScrollBar.Free;
  inherited Destroy;
end;

procedure TPopupSelector.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := WS_POPUP or WS_BORDER or WS_CLIPCHILDREN;
    {$IFDEF WIN32}
    ExStyle := WS_EX_TOOLWINDOW;
    {$ENDIF}
    WindowClass.Style := WindowClass.Style or CS_SAVEBITS;
  end;
end;

{$IFNDEF WIN32}
procedure TPopupSelector.CreateWnd;
begin
  inherited CreateWnd;
  if (csDesigning in ComponentState) then SetParent(nil);
end;
{$ENDIF}

procedure TPopupSelector.SetBounds(aLeft, aTop, aWidth, aHeight: integer);
var tx: TSize; s: string; DC: HDC;
begin
  Font := Self.Font;
  s := Format('%s %d', [LongMonthNames[9], 2000]); {'september' is widest string}
  DC := GetDC(0);
  with Canvas do try
    Handle := DC;
    Font := Self.Font;
    {$IFDEF D3}
    tx := TextExtent(s);
    {$ELSE}
    tx.cx := TextWidth(s);
    tx.cy := TextHeight(s);
    {$ENDIF}
  finally
    Canvas.Handle := 0;
    ReleaseDC(0, DC);
  end;
  FMonthWidth := tx.cx + 16;
  FMonthHeight := tx.cy + 4;
  aWidth := FMonthWidth + GetSystemMetrics(SM_CXVSCROLL);
  aHeight := (2 * PopupDelta + 1) * FMonthHeight;
  inherited SetBounds(aLeft, aTop, aWidth, aHeight);
end;

procedure TPopupSelector.Paint;
var i: integer; r, c: TRect;

  function GetMonthRect(Shift: integer): TRect;
  begin
    Result := Rect(0, (Shift + PopupDelta) * FMonthHeight, FMonthWidth,
      (Shift + PopupDelta + 1) * FMonthHeight);
  end;

  function GetMonthString(Shift: integer): string;
  var aYear, aMonth: word;
  begin
    aYear := Year;
    aMonth := Month;
    ShiftMonth(aYear, aMonth, Shift);
    Result := Format('%s %d', [MonthName(False, aMonth), aYear]);
  end;

  procedure DrawMonthRect(Shift: integer; aRect: TRect);
  begin
    with Canvas do
      if Shift = 0 then
      begin
        Brush.Color := TDCCalendar(Self.Parent).SelBackColor;
        Font.Color := TDCCalendar(Self.Parent).SelFontColor;
      end
      else
      begin
        Brush.Color := TDCCalendar(Self.Parent).Color;
        Font.Color := TDCCalendar(Self.Parent).Font.Color;
      end;
    TextOutCentered(Canvas, GetMonthString(Shift), aRect);
  end;

begin
  with Canvas do
  begin
    for i := -PopupDelta to PopupDelta do
    begin
      r := GetMonthRect(i);
      if IntersectRect(c, ClipRect, r) {$IFNDEF WIN32}<> 0{$ENDIF} then
        DrawMonthRect(i, r);
    end;
  end;
end;

procedure TPopupSelector.KeyDown(var Key: word; Shift: TShiftState);
begin
  if Shift = [] then
    case Key of {Key:=0 is set in parent control}
      VK_UP:    EventFromScroller(scLineUp);
      VK_DOWN:  EventFromScroller(scLineDown);
      VK_PRIOR: EventFromScroller(scPageUp);
      VK_NEXT:  EventFromScroller(scPageDown);
    end;
  inherited KeyDown(Key, Shift);
end;

procedure TPopupSelector.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: integer);
var Sh: integer; pb: TPoint;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) then
    if PtInRect(GetMonthArea, Point(X, Y)) then
    begin
      Sh := (Y - 1) div FMonthHeight - PopupDelta;
      GetCursorPos(pb);
      SetCursorPos(pb.x, pb.y - Sh * FMonthHeight);
      SetMonth(Month + Sh);
    end
    else
    begin
      pb := FScrollBar.ScreenToClient(ClientToScreen(Point(X, Y)));
      FScrollBar.MouseDown(Button, Shift, pb.X, pb.Y);
    end;
end;

procedure TPopupSelector.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) then
    CloseUp(PtInRect(GetMonthArea, Point(X, Y)));
end;

procedure TPopupSelector.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  Msg.Result := DLGC_WANTARROWS; {allow change date with arrows}
end;

procedure TPopupSelector.WMMouseActivate(var Message: TMessage);
begin
  Message.Result := MA_NOACTIVATE;
end;

procedure TPopupSelector.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin {control is opaque}
end;

procedure TPopupSelector.CMFontChanged(var Message: TMessage);
begin
  inherited;
  SetBounds(Left, Top, Width, Height);
end;

procedure TPopupSelector.SetDate(Value: TDateTime);
begin
  if Value <> FDate then
  begin
    FDate := Value;
    UpdateDate;
  end;
end;

function TPopupSelector.GetDateElement(Index: integer): integer;
var aYear, aMonth, aDay: word;
begin
  DecodeDate(FDate, aYear, aMonth, aDay);
  case Index of
    elemYear:  Result := aYear;
    elemMonth: Result := aMonth;
    else       Result := 0;
  end;
end;

function TPopupSelector.GetMonthArea: TRect;
begin
  Result := Rect(0, 0, FMonthWidth, (2 * PopupDelta + 1) * FMonthHeight); 
end;

procedure TPopupSelector.SetMonth(Value: integer);
var aYear, aMonth, aDay: word;
begin
  DecodeDate(FDate, aYear, aMonth, aDay);
  if Value <> aMonth then
  begin
    ShiftMonth(aYear, aMonth, Value - aMonth);
    aDay := Min(aDay, DaysPerMonth(aYear, aMonth));
    FDate := EncodeDate(aYear, aMonth, aDay);
    UpdateDate;
  end;
end;

procedure TPopupSelector.DoScrollMonthes(Sender: TObject);
begin
  SetMonth(Month + FTrackDelta);
end;

procedure TPopupSelector.UpdateDate;
var r: TRect;
begin
  {redraw (2*PopupDelta+1) month cells} 
  r := GetMonthArea;
  InvalidateRect(Handle, @r, False);
end;

procedure TPopupSelector.EventFromScroller(sCode: TScrollCode);
{$IFDEF WIN32}var  si: TScrollInfo;{$ENDIF}
begin
  case sCode of
    scLineUp:   SetMonth(Pred(Month));
    scLineDown: SetMonth(Succ(Month));
    scPageUp:   SetMonth(Month - PopupDelta - 1);
    scPageDown: SetMonth(Month + PopupDelta + 1);
    scPosition: FTimer.Enabled := False;
    scTrack:    {$IFDEF WIN32}
                with si do
                begin
                  cbSize := SizeOf(si);
                  fMask := SIF_ALL;
                  GetScrollInfo(FSCrollBar.Handle, SB_CTL, si);
                  case nTrackPos of
                    0: begin
                         FTrackDelta := -MonthesPerYear;
                         FTimer.Interval := 100;
                       end;
                    1: begin
                         FTrackDelta := -PopupDelta - 1;
                         FTimer.Interval := 150;
                       end;
                    2: begin
                         FTrackDelta := -1;
                         FTimer.Interval := 200;
                       end;
                    2 * PopupDelta - 2: begin
                         FTrackDelta := 1;
                         FTimer.Interval := 200;
                       end;
                    2 * PopupDelta - 1: begin
                         FTrackDelta := PopupDelta + 1;
                         FTimer.Interval := 150;
                       end;
                    2 * PopupDelta: begin
                         FTrackDelta := MonthesPerYear;
                         FTimer.Interval := 100;
                       end;
                    else FTrackDelta := 0;
                  end;
                   FTimer.Enabled := FTrackDelta <> 0;
                end;
                {$ELSE}
                ; {Win16 havn't functions to retrieve TrackPos}
                {$ENDIF}
  end;
end;

procedure TPopupSelector.PopupMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: integer);
begin
  if Button = mbLeft then
    CloseUp(PtInRect(Self.ClientRect, Point(X, Y)));
end;

procedure TPopupSelector.CloseUp(Accept: Boolean);
begin
  if Assigned(FCloseUp) then
    FCloseUp(Self, Accept);
end;

procedure TPopupSelector.Show(AlignTo: TRect);
var pr: TRect; pp: TPoint;
begin
  SetBounds(Left, Top, Width, Height);
  pr := ClientRect;
  pp := Parent.ClientToScreen(Point(0,0));
  pp.x := pp.x +
    ((AlignTo.Right + AlignTo.Left) - (pr.Right - pr.Left)) div 2;
  pp.y := pp.y +
    ((AlignTo.Bottom + AlignTo.Top) - (pr.Bottom - pr.Top)) div 2;
  if Left < 0 then Left := 0;
  if Top < 0 then Top := 0;
  SetWindowPos(Handle, HWND_TOP, pp.x, pp.y, 0, 0, SWP_NOSIZE or
    SWP_NOZORDER or SWP_NOACTIVATE or SWP_SHOWWINDOW);
  Visible := True;
end;

procedure TPopupSelector.Hide;
begin
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE or
    SWP_NOZORDER or SWP_NOACTIVATE or SWP_HIDEWINDOW);
  Visible := False;
end;

{*** TDCCalendar ****************************************************}

constructor TDCCalendar.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  ControlState := ControlState + [csCreating];
  try
    ControlStyle := ControlStyle + [csOpaque, csCaptureMouse];
    Color := clWindow;
    FColors[1] := clHighlightText;
    FColors[2] := clHighlight;
    FColors[3] := clWindowText;
    FColors[4] := clBtnFace;
    FColors[5] := clGrayText;
    FColors[6] := clMaroon;
    FOptions := [caShowToday, caDrawFocus];
    FDate := SysUtils.Date; {initial date is today}
    FAnchor := FDate;
    Offset1stDay := DayOfWeek(EncodeDate(Year, Month, 1)) - 1;
    FSelectType := stOneDay;
    FAutoSize := True;
    FMultiSelect := False;
    IsMonthChg := True;
    IsMouseDown := False;
    InSelectMode := False;
    ContinueSelect := False;
    InDialogMode := False;
    TabStop := True;
    ParentColor := False;
    ParentFont := True;
    ParentShowHint := True;
    FPopup := TPopupSelector.Create(Self);
    FPopup.OnCloseUp := PopupCloseUp;
    UpdateDate;
    SetBounds(Left, Top, 100, 100);
  finally
    ControlState := ControlState - [csCreating];
  end;
end;

destructor TDCCalendar.Destroy;
begin
  if FPopup <> nil then
    FPopup.OnCloseUp := nil;
  FPopup.Free;
  FPopup := nil;
  inherited Destroy;
end;

procedure TDCCalendar.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if InDialogMode then
  begin
    Params.Style := Params.Style or WS_BORDER;
    EdgeDepth := 0; {used BS_BORDER style}
  end
  else
    EdgeDepth := 1;
end;

procedure TDCCalendar.CalcAutoCellSize(var aWidth, aHeight: integer);
var tx: TSize; DC: HDC;
begin
  if FAutoSize then with Canvas do
  begin
    DC := GetDC(0);
    try
      Handle := DC;
      Font := Self.Font;
      {$IFDEF D3}
      tx := TextExtent(ShortDayNames[4]); {Wednesday}
      {$ELSE}
      tx.cx := TextWidth(ShortDayNames[4]);
      tx.cy := TextHeight(ShortDayNames[4]);
      {$ENDIF}
    finally
      Canvas.Handle := 0;
      ReleaseDC(0, DC);
    end;
    cellW := tx.cx + 8;
    cellH := tx.cy + 4;
    aWidth := (DaysPerWeek + Ord(caShowWeekNum in FOptions)) * cellW +
      2 * EdgeDepth;
    aHeight := (MaxWeeksPerMonth + 2) * cellH + 2 * EdgeDepth;
  end
  else
  begin
    cellW := (aWidth - 2 * EdgeDepth) div
      (DaysPerWeek + Ord(caShowWeekNum in FOptions));
    cellH := (aHeight - 2 * EdgeDepth) div (MaxWeeksPerMonth + 2);
  end;
end;

procedure TDCCalendar.SetBounds(aLeft, aTop, aWidth, aHeight: integer);
begin
  CalcAutoCellSize(aWidth, aHeight);
  inherited SetBounds(aLeft, aTop, aWidth, aHeight);
end;

procedure TDCCalendar.Paint;
var curRect: TRect;

  procedure DrawMyFocusRect(aRect: TRect);
  var i, sx, sy: integer; bc: TColor;
  begin
    with Canvas do
    begin
      bc := Brush.Color;
      Brush.Color := Self.Font.Color;
      FrameRect(aRect);
      Brush.Color := bc;
      sx := Ord(not Odd(aRect.Bottom - aRect.Top));
      sy := Ord(not Odd(aRect.Right - aRect.Left));
      for i := 0 to (aRect.Right - aRect.Left + 1) div 2 do
      begin
        Pixels[aRect.Left + i shl 1, aRect.Top] := Self.Color;
        Pixels[aRect.Left + i shl 1 + sx, aRect.Bottom - 1] := Self.Color;
      end;
      for i := 0 to (aRect.Bottom - aRect.Top + 1) div 2 do
      begin
        Pixels[aRect.Left, aRect.Top + i shl 1] := Self.Color;
        Pixels[aRect.Right - 1, aRect.Top + i shl 1 + sy] := Self.Color;
      end;
    end;
  end;

  function IsSelectedDate(sDate: TDateTime): Boolean;
  begin
    if FMultiSelect then
      Result := (StartDate <= sDate) and (sDate <= FinishDate)
    else
      case FSelectType of
        stOneDay:   Result := sDate = FDate;
        stWorkWeek: Result := (FDate - DayOfWeek(FDate) + 2 <= sDate) and
                      (sDate <= FDate - DayOfWeek(FDate) + DaysPerWeek - 1);
        stWeek:     Result := (FDate - DayOfWeek(FDate) + 1 <= sDate) and
                      (sDate <= FDate - DayOfWeek(FDate) + DaysPerWeek);
        else        Result := False;
      end;
  end;

  procedure DrawSelectedCell(drawDate: TDateTime; Str: string; aRect: TRect;
    Color1, Color2: TColor; cmpDay: integer);
  begin
    with Canvas do
    begin
      {! Brush.Color := clRed; FillRect(aRect); Sleep(40); {for test only}
      if IsSelectedDate(drawDate) then
      begin
        Font.Color := Color1;
        if (caUseEllipticSelection in FOptions) and
          not (IsSelectedDate(drawDate - 1) and IsSelectedDate(drawDate + 1)) then
        begin { draw text on elliptic background }
          if IsSelectedDate(drawDate - 1) then
          begin
            Brush.Color := SelBackColor;
            Pen.Color := Brush.Color;
          end
          else
          begin
            Brush.Color := Color;
          end;
          with aRect do
            FillRect(Rect(Left, Top, (Left + Right) div 2, Bottom));
          if IsSelectedDate(drawDate + 1) then
          begin
            Brush.Color := SelBackColor;
            Pen.Color := Brush.Color;
          end
          else
          begin
            Brush.Color := Color;
          end;
          with aRect do
            FillRect(Rect((Left + Right) div 2, Top, Right, Bottom));
          Brush.Color := SelBackColor;
          Pen.Color := Brush.Color;
          Ellipse(aRect.Left, aRect.Top, aRect.Right, aRect.Bottom);
          Brush.Style := bsClear;
          DrawText(Handle, {$IFDEF WIN32}PChar(Str){$ELSE}@Str[1]{$ENDIF},
            Length(Str), aRect, DT_CENTER or DT_SINGLELINE or DT_VCENTER);
        end
        else
        begin {draw rectangle with SelColor}
          Brush.Color := SelBackColor;
          TextOutCentered(Canvas, Str, aRect);
        end;
      end
      else
      begin
        Brush.Color := Color;
        Font.Color := Color2;
        TextOutCentered(Canvas, Str, aRect);
      end;
      if Focused and (caDrawFocus in FOptions) and
        (FUnderMouseDay = cmpDay) then
        DrawMyFocusRect(aRect); {draw a frame around under mouse day}
    end;
  end;

  procedure DrawOuterEdge;
  var i: integer;
  begin
    with Canvas do
    begin
      for i := 0 to EdgeDepth - 1 do
      begin
        Pen.Color := clBtnShadow;
        MoveTo(i, Height - i);
        LineTo(i, i);
        LineTo(Width - i, i);
        Pen.Color := clBtnHighlight;
        MoveTo(i, Height - i - 1);
        LineTo(Width - i - 1, Height - i - 1);
        LineTo(Width - i - 1, i - 1);
      end;
    end;
  end;

  procedure DrawTitle;
  var r: TRect;
  begin
    r := GetTitleRect;
    with Canvas, r do
    begin
      Brush.Color := TitleBackColor;
      Font.Color := TitleFontColor;
      TextOutCentered(Canvas, GetTitle, r);
      Pen.Color := clBtnHighlight;
      MoveTo(Left, Bottom - 1);
      LineTo(Left, Top);
      LineTo(Right - 1, Top);
      Pen.Color := clBtnShadow;
      MoveTo(Left, Bottom - 1);
      LineTo(Right - 1, Bottom - 1);
      LineTo(Right - 1, Top);
    end;
  end;

  procedure DrawHeader;
  var r: TRect; i: integer;
  begin
    with Canvas, r do
    begin
      Brush.Color := Color;
      Font.Color := Self.Font.Color;
      if caShowWeekNum in FOptions then
        TextOutCentered(Canvas, '', GetCellRect(-1, 1));
      for i := 0 to DaysPerWeek - 1 do
        TextOutCentered(Canvas, ShortDayNames[i + 1], GetCellRect(i, 1));
      FillRect(Rect(EdgeDepth + (DaysPerWeek +
        Ord(caShowWeekNum in FOptions)) * cellW, EdgeDepth + cellH, 
        Width - EdgeDepth, EdgeDepth + 2 * cellH));
      r := GetHeaderRect;
      Pen.Color := clBtnShadow;
      MoveTo(Left + LinePadding + Ord(caShowWeekNum in FOptions) * cellW, 
        Bottom - 1);
      LineTo(Right - LinePadding, Bottom - 1);
    end;
  end;

  procedure DrawWeekNums;
  var i, wn: integer; r: TRect;
  begin
    with Canvas, r do
    begin
      Brush.Color := Color;
      for i := 0 to MaxWeeksPerMonth - 1 do
      begin
        wn := CurMonth1stWeekNum + i;
        if wn > 53 {max number of weeks in year} then
        begin
          if DayOfWeek(EncodeDate(Succ(Year), 1, 1)) = 1 then
            wn := 1
          else
            wn := 2;
          Font.Color := OtherMonthColor;
        end
        else
          Font.Color := Self.Font.Color;
        TextOutCentered(Canvas, IntToStr(wn), GetCellRect(-1, i + 2));
      end;
      r := GetWeekNumRect;
      Pen.Color := clBtnShadow;
      MoveTo(Right - 1, Top + LinePadding);
      LineTo(Right - 1, Bottom - LinePadding);
    end;
  end;

  procedure DrawPrevMonth;
  var i, D: integer; aRect: TRect; drawDate: TDateTime;
      pYear, pMonth, pDay: word;
  begin
    with Canvas do
    begin
      if Month = 1 then
      begin
        pYear := Pred(Year);
        pMonth := MonthesPerYear;
      end
      else
      begin
        pYear := Year;
        pMonth := Pred(Month);
      end;
      D := DaysPerMonth(pYear, pMonth);
      if caShowPrevMonth in FOptions then
        for i := 0 to Offset1stDay - 1 do
        begin
          aRect := GetCellRect(i, 2);
          pDay := D - Offset1stDay + i + 1;
          drawDate := EncodeDate(pYear, pMonth, pDay);
          DrawSelectedCell(drawDate, IntToStr(pDay), aRect, OtherMonthColor, 
            OtherMonthColor, i - Offset1stDay + 1);
        end
      else
      begin
        Brush.Color := Color;
        Brush.Style := bsSolid;
        Canvas.FillRect(GetPrevMonthRect);
      end;
    end;
  end;

  procedure DrawNextMonth;
  var i, D: integer; aRect: TRect; drawDate: TDateTime;
      pYear, pMonth: word;
  begin
    with Canvas do
    begin
      if Month = MonthesPerYear then
      begin
        pYear := Succ(Year);
        pMonth := 1;
      end
      else
      begin
        pYear := Year;
        pMonth := Succ(Month);
      end;
      D := MaxWeeksPerMonth * DaysPerWeek - DaysThisMonth - Offset1stDay;
      if caShowNextMonth in FOptions then
        for i := 1 to D do
        begin
          aRect := GetDayRect(DaysThisMonth + i);
          drawDate := EncodeDate(pYear, pMonth, i);
          DrawSelectedCell(drawDate, IntToStr(i), aRect, OtherMonthColor,
            OtherMonthColor, DaysThisMonth + i);
        end
      else
      begin
        Brush.Color := Color;
        Brush.Style := bsSolid;
        Canvas.FillRect(GetNextMonthRect);
      end;
    end;
  end;

  procedure DrawDayCell(aDay: integer; aRect: TRect);
  var s: string; drawDate: TDateTime;
  begin
    with Canvas do
    begin
      s := IntToStr(aDay);
      drawDate := EncodeDate(Year, Month, aDay);
      DrawSelectedCell(drawDate, s, aRect, SelFontColor, Self.Font.Color, aDay);
      if (caShowToday in FOptions) and
        (SysUtils.Date = EncodeDate(Year, Month, aDay)) then
      begin {draw a frame (1px inside) around today date}
        Brush.Color := TodayColor;
        InflateRect(aRect, -1, -1);
        FrameRect(aRect);
        InflateRect(aRect, -1, -1);
        FrameRect(aRect);
      end;
    end;
  end;

  procedure DrawDaysArea;
  var i: integer; r, c: TRect;
  begin
    with Canvas do
    begin
      for i := 1 to DaysThisMonth do
      begin
        r := GetDayRect(i);
        if IntersectRect(c, Canvas.ClipRect, r) {$IFNDEF WIN32}<> 0{$ENDIF} then
          DrawDayCell(i, r);
      end;
    end;
  end;

  procedure DrawMargin(aRect: TRect);
  begin
    with Canvas do
    begin
      Brush.Color := Color;
      FillRect(aRect);
    end;
  end;

begin
  Canvas.Font := Font;
  DrawOuterEdge;
  if IntersectRect(curRect, Canvas.ClipRect, GetTitleRect)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawTitle;
  if IntersectRect(curRect, Canvas.ClipRect, GetHeaderRect)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawHeader;
  if (IntersectRect(curRect, Canvas.ClipRect, GetWeekNumRect)
     {$IFNDEF WIN32}<> 0{$ENDIF}) and (caShowWeekNum in FOptions) then
    DrawWeekNums;
  if IntersectRect(curRect, Canvas.ClipRect, GetPrevMonthRect)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawPrevMonth;
  if IntersectRect(curRect, Canvas.ClipRect, GetNextMonthRect)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawNextMonth;
  if IntersectRect(curRect, Canvas.ClipRect, GetDaysArea)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawDaysArea;
  if IntersectRect(curRect, Canvas.ClipRect, GetRightMarginRect)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawMargin(GetRightMarginRect);
  if IntersectRect(curRect, Canvas.ClipRect, GetBottomMarginRect)
     {$IFNDEF WIN32}<> 0{$ENDIF} then
    DrawMargin(GetBottomMarginRect);
end;

procedure TDCCalendar.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  Msg.Result := DLGC_WANTARROWS; {allow change date with arrows}
end;

procedure TDCCalendar.WMSetFocus(var Msg: TWMSetFocus);
var r: TRect; aYear, aMonth, aDay: word;
begin
  inherited;
  FFocused := True;
  {redraw current cell}
  DecodeDate(FDate, aYear, aMonth, aDay);
  r := GetDayRect(aDay); {current cell}
  InvalidateRect(Handle, @r, False);
end;

procedure TDCCalendar.WMKillFocus(var Msg: TWMKillFocus);
var r: TRect; aYear, aMonth, aDay: word;
begin
  inherited;
  FFocused := False;
  PopupCloseUp(FPopup, False);
  {redraw current cell}
  DecodeDate(FDate, aYear, aMonth, aDay);
  r := GetDayRect(aDay); {current cell}
  InvalidateRect(Handle, @r, False);
end;

procedure TDCCalendar.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin {control is opaque}
end;

procedure TDCCalendar.CMFontChanged(var Message: TMessage);
begin
  inherited;
  SetBounds(Left, Top, Width, Height);
  Repaint; {redraw all}
end;

procedure TDCCalendar.CMCancelMode(var Message: TCMCancelMode);
begin
  if (Message.Sender <> Self) and (Message.Sender <> FPopup) and
    ((FPopup <> nil) and not FPopup.ContainsControl(Message.Sender)) then
    PopupCloseUp(FPopup, False);
end;

procedure TDCCalendar.KeyDown(var Key: word; Shift: TShiftState);
begin
  if IsMouseDown then
    exit;
  if FPopupVisible then
  begin
    if Shift = [] then
      case Key of
        VK_PRIOR, VK_NEXT, VK_UP, VK_DOWN:
          FPopup.KeyDown(Key, Shift);
        VK_RETURN, VK_ESCAPE, VK_F2:
          PopupCloseUp(FPopup, Key = VK_RETURN);
      end;
    Key := 0;
  end;
  InSelectMode := FMultiSelect and (ssShift in Shift);
  ContinueSelect := InSelectMode;
  Shift := Shift - [ssShift];
  if Shift = [] then
    case Key of
      VK_ESCAPE: if InDialogMode and (PopupForm <> nil) then
                   PopupForm.ModalResult := mrCancel;
      VK_RETURN: if InDialogMode and (PopupForm <> nil) then
                   PopupForm.ModalResult := mrOK;
      VK_F2:     if (FPopup <> nil) then
                   PopupDropDown;
      VK_LEFT:   begin
                   SetDate(ShiftDate(elemDay, -1));
                   Key := 0;
                 end;
      VK_RIGHT:  begin
                   SetDate(ShiftDate(elemDay, 1));
                   Key := 0;
                 end;
      VK_PRIOR:  begin
                   SetDate(ShiftDate(elemMonth, -1));
                   Key := 0;
                 end;
      VK_NEXT:   begin
                   SetDate(ShiftDate(elemMonth, 1));
                   Key := 0;
                 end;
      VK_UP:     begin
                   SetDate(ShiftDate(elemDay, -DaysPerWeek));
                   Key := 0;
                 end;
      VK_DOWN:   begin
                   SetDate(ShiftDate(elemDay, DaysPerWeek));
                   Key := 0;
                 end;
      VK_HOME:   begin
                   Day := 1; {there is no jumps between monthes/years}
                   Key := 0;
                 end;
      VK_END:    begin
                   Day := DaysThisMonth; {there is no jumps between monthes/years}
                   Key := 0;
                 end;
    end
  else if ssCtrl in Shift then
    case Key of
      VK_PRIOR: begin {Ctrl-PageUp}
                  SetDate(ShiftDate(elemYear, -1));
                  Key := 0;
                end;
      VK_NEXT:  begin {Ctrl-PageDown}
                  SetDate(ShiftDate(elemYear, 1));
                  Key := 0;
                end;
    end;
  inherited KeyDown(Key, Shift);
end;

procedure TDCCalendar.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: integer);
var r: TRect;
begin
  {show or close popup window}
  if (FPopup <> nil) and (Button = mbLeft) then
  begin
    if CanFocus then
      SetFocus;
    if not FFocused then
      exit;
    if FPopupVisible then
      PopupCloseUp(FPopup, False)
    else if PtInRect(GetTitleRect, Point(X, Y)) then
       PopupDropDown;
  end;
  if (Button = mbLeft) and PtInRect(GetDaysArea, Point(X, Y)) then
  begin
    IsMouseDown := True;
    SetUnderMouseDay(GetUnderMouseDay(X, Y));
    InSelectMode := FMultiSelect and ((caShowPrevMonth in Options)
      or (FUnderMouseDay >= 1)) and ((caShowNextMonth in Options)
      or (FUnderMouseDay <= DaysThisMonth));
    ContinueSelect := False;
    if InSelectMode then
    begin {Set anchor & date to UnderMouseDay. SetDate not used,
      'cause need to redraw old selection when FDate is new & FAnchor is old.
      Following lines emulate SetDate but calls InvalidateRect for saOldSelect}
      FOldDate := FDate;
      FDate := ShiftDate(elemDay, FUnderMouseDay - Day);
      r := GetSelectedArea(saOldSelect);
      InvalidateRect(Handle, @r, False);
      FAnchor := FDate;
      UpdateDate;
      MoveCursorToDateCell;
    end;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TDCCalendar.MouseMove(Shift: TShiftState; X, Y: Integer);
var aRect: TRect;
begin
  if IsMouseDown then
  begin
    aRect := GetDaysArea;
    if PtInRect(aRect, Point(X, Y)) then
      SetUnderMouseDay(GetUnderMouseDay(X, Y))
    else
      SetUnderMouseDay(WrongUnderMouseDay);
    if InSelectMode then
    begin
      ContinueSelect := True;
      if PtInRect(aRect, Point(X, Y)) and (FUnderMouseDay <> Day) then
      begin
        SetDate(ShiftDate(elemDay, FUnderMouseDay - Day));
        MoveCursorToDateCell;
      end
      else with aRect do
        if (Left <= X) and (X <= Right) then
          if Y < Top - DragDelta then
          begin
            SetDate(ShiftDate(elemDay, GetUnderMouseDay(X, Top + 1)
              - DaysPerWeek - Day));
            MoveCursorToDateCell;
          end
          else if Y > Bottom + DragDelta then
          begin
            SetDate(ShiftDate(elemDay, GetUnderMouseDay(X, Bottom - 1)
              + DaysPerWeek - Day));
            MoveCursorToDateCell;
          end;
    end;
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TDCCalendar.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if IsMouseDown then {for any mouse button}
  begin
    IsMouseDown := False;
    if PtInRect(GetDaysArea, Point(X, Y)) and ((caShowPrevMonth in Options)
      or (FUnderMouseDay >= 1)) and ((caShowNextMonth in Options)
      or (FUnderMouseDay <= DaysThisMonth)) then
    begin
      SetDate(ShiftDate(elemDay, FUnderMouseDay - Day));
      if InDialogMode and (PopupForm <> nil) then
        PopupForm.ModalResult := mrOK;
    end
    else
      SetUnderMouseDay(Day);
  end;
  inherited MouseDown(Button, Shift, X, Y);
  InSelectMode := False;
  ContinueSelect := False;
end;

procedure TDCCalendar.SetDate(Value: TDateTime);
var oldY, oldM, oldD, newY, newM, newD: word;
begin
  if Value <> FDate then
  begin
    DecodeDate(FDate, oldY, oldM, oldD);
    FOldDate := FDate;
    FDate := Value;
    DecodeDate(FDate, newY, newM, newD);
    IsMonthChg := (oldM <> newM) or (oldY <> newY);
    UpdateDate;
  end;
end;

function TDCCalendar.GetRangeDate(Index: integer): TDateTime;
begin
  if Index = 1 then
  begin
    if FDate < FAnchor then
    begin
      Result := FDate;
      if InSelectMode and (FUnderMouseDay < Day) then
        Result := Result - Day + FUnderMouseDay;
    end
    else
      Result := FAnchor;
    case FSelectType of
      {stOneDay - nothing to do}
      stWorkWeek: Result := Result - DayOfWeek(Result) + 2;
      stWeek:     Result := Result - DayOfWeek(Result) + 1;
    end;
  end
  else
  begin
    if FDate < FAnchor then
      Result := FAnchor
    else
    begin
      Result := FDate;
      if InSelectMode and (FUnderMouseDay > Day) then
        Result := Result - Day + FUnderMouseDay;
    end;
    case FSelectType of
      {stOneDay - nothing to do}
      stWorkWeek: Result := Result - DayOfWeek(Result) + DaysPerWeek - 1;
      stWeek:     Result := Result - DayOfWeek(Result) + DaysPerWeek;
    end;
  end;
end;

function TDCCalendar.GetDateElement(Index: integer): integer;
var aYear, aMonth, aDay: word;
begin
  DecodeDate(FDate, aYear, aMonth, aDay);
  case Index of
    elemYear:  Result := aYear;
    elemMonth: Result := aMonth;
    elemDay:   Result := aDay;
    else       Result := 0;
  end;
end;

procedure TDCCalendar.SetDateElement(Index: integer; Value: integer);
var aYear, aMonth, aDay: word; 
begin
  IsMonthChg := False;
  DecodeDate(FDate, aYear, aMonth, aDay);
  if Value < 1 then
    Value := 1; {correct low limit for all elements}
  case Index of
    elemYear:  begin
                 IsMonthChg := IsMonthChg or (Value <> aYear);
                 if IsMonthChg then
                   aYear := Value
                 else
                   exit;
               end;
    elemMonth: begin
                 Value := Min(Value, MonthesPerYear); {correct max month value}
                 IsMonthChg := IsMonthChg or (Value <> aMonth);
                 if IsMonthChg then
                   aMonth := Value
                 else
                   exit;
               end;
    elemDay:   begin
                 {setting max day value 1 line below can be omitted
                  as Day will be corrected before EncodeDate in any case}
                 Value := Min(Value, DaysThisMonth); {correct max day value}
                 if Value <> aDay then
                   aDay := Value
                 else
                   exit;
               end;
    else
               exit;
  end;
  {in any case correct aDay according to aMonth & aYear}
  aDay := Min(aDay, DaysPerMonth(aYear, aMonth));
  SetDate(EncodeDate(aYear, aMonth, aDay));
end;

function TDCCalendar.GetColors(Index: integer): TColor;
begin
  Result := FColors[Index];
end;

procedure TDCCalendar.SetColors(Index: integer; Value: TColor);
var r: TRect;
begin
  if Value <> FColors[Index] then
  begin
    FColors[Index] := Value;
    case Index of
      1, 2: begin {redraw selected cells}
           r := GetSelectedArea(saWhole);
           InvalidateRect(Handle, @r, False);
         end;
      3, 4: begin {redraw title}
           r := GetTitleRect;
           InvalidateRect(Handle, @r, False);
         end;
      5: begin {redraw prev. & next monthes cells only}
           r := GetPrevMonthRect;
           InvalidateRect(Handle, @r, False);
           r := GetNextMonthRect;
           InvalidateRect(Handle, @r, False);
         end;
      6: begin {redraw today cell only}
           r := GetDayRect(GetToday);
           InvalidateRect(Handle, @r, False);
         end;
    end;
  end;
end;

procedure TDCCalendar.SetMultiSelect(Value: Boolean);
var r: TRect;
begin
  if Value <> FMultiSelect then
  begin
    FMultiSelect := Value;
    {FMultiSelect has new value, but FAnchor still old. So, redraw old
     selected area with new style. In any case FDate is inside _old_ selected
     area and it is not necessary to redraw a _new_ selected area}
    r := GetSelectedArea(saWhole);
    InvalidateRect(Handle, @r, False);
    if not FMultiSelect then
      FAnchor := FDate;
  end;
end;

procedure TDCCalendar.SetOptions(Value: TCaOptions);
var r: TRect; oldOptions: TCaOptions;

  function IsOptionChanged(aOption: TCaOption): Boolean;
  begin
    Result := (aOption in oldOptions) xor (aOption in FOptions);
  end;
  
begin
  if Value <> FOptions then
  begin
    oldOptions := FOptions;
    FOptions := Value;
    if IsOptionChanged(caDrawFocus) then
    begin {redraw focused cell}
      r := GetDayRect(Day);
      InvalidateRect(Handle, @r, False);
    end;
    if IsOptionChanged(caShowPrevMonth) then
    begin {redraw prev. month area}
      r := GetPrevMonthRect;
      InvalidateRect(Handle, @r, False);
    end;
    if IsOptionChanged(caShowNextMonth) then
    begin {redraw next. month area}
      r := GetNextMonthRect;
      InvalidateRect(Handle, @r, False);
    end;
    if IsOptionChanged(caShowToday) then
    begin {redraw today cell}
      r := GetDayRect(GetToday);
      InvalidateRect(Handle, @r, False);
    end;
    if IsOptionChanged(caShowWeekNum) then
    begin
      SetBounds(Left, Top, Width, Height);
      Repaint; {redraw all}
    end;
    if IsOptionChanged(caUseEllipticSelection) then
    begin
      r := GetSelectedArea(saFirstCell);
      InvalidateRect(Handle, @r, False);
      r := GetSelectedArea(saLastCell);
      InvalidateRect(Handle, @r, False);
    end;
    if IsOptionChanged(caUseShortTitle) then
    begin
      r := GetTitleRect;
      InvalidateRect(Handle, @r, False);
    end;
  end;
end;

procedure TDCCalendar.SetAutoSize(Value: Boolean);
begin
  if Value <> FAutoSize then
  begin
    FAutoSize := Value;
    PostMessage(Parent.Handle, WM_SIZE, SIZE_RESTORED, 0);
    SetBounds(Left, Top, Width, Height);
  end;
end;

procedure TDCCalendar.SetSelectType(Value: TCaSelectType);
var r: TRect;
begin
  if Value <> FSelectType then
  begin
    FSelectType := Value;
    {redraw selected cells}
    r := GetSelectedArea(saFirstLine);
    InvalidateRect(Handle, @r, False);
    r := GetSelectedArea(saLastLine);
    InvalidateRect(Handle, @r, False);
  end;
end;

function TDCCalendar.GetTitle: string;
begin
  Result :=
    Format('%s %d', [MonthName(caUseShortTitle in FOptions, Month), Year]);
end;

function TDCCalendar.DaysThisMonth: integer;
begin
  Result := DaysPerMonth(Year, Month);
end;

function TDCCalendar.GetUnderMouseDay(X, Y: integer): integer;
var mx, my: integer;
begin
  mx := (X - EdgeDepth) div cellW - Ord(caShowWeekNum in FOptions);
  my := (Y - EdgeDepth) div cellH - 2;
  Result := mx + my * DaysPerWeek - Offset1stDay + 1;
end;

procedure TDCCalendar.SetUnderMouseDay(umDay: integer);
var r: TRect;
begin
  if umDay <> FUnderMouseDay then
  begin
    r := GetDayRect(FUnderMouseDay); {old cell under mouse}
    InvalidateRect(Handle, @r, False);
    FUnderMouseDay := umDay;
    r := GetDayRect(FUnderMouseDay); {new cell under mouse}
    InvalidateRect(Handle, @r, False);
  end;
end;

function TDCCalendar.ShiftDate(Elem, Delta: integer): TDateTime;
var aYear, aMonth, aDay: word;
begin
  Result := FDate;
  IsMonthChg := False;
  DecodeDate(FDate, aYear, aMonth, aDay);
  case Elem of
    elemYear:  begin
                 aYear := aYear + Delta;
                 IsMonthChg := True;
               end;
    elemMonth: begin
                 ShiftMonth(aYear, aMonth, Delta);
                 IsMonthChg := True;
               end;
    elemDay:   begin
                 if aDay + Delta < 1 then
                 begin
                   ShiftMonth(aYear, aMonth, -1);
                   aDay := DaysPerMonth(aYear, aMonth) + aDay + Delta;
                   IsMonthChg := True;
                 end
                 else if aDay + Delta > DaysPerMonth(aYear, aMonth) then
                 begin
                   aDay := aDay + Delta - DaysPerMonth(aYear, aMonth);
                   ShiftMonth(aYear, aMonth, 1);
                   IsMonthChg := True;
                 end
                 else
                   aDay := aDay + Delta;
               end;
    else
               exit;
  end;
  aDay := Min(aDay, DaysPerMonth(aYear, aMonth));
  Result := EncodeDate(aYear, aMonth, aDay);
end;

procedure TDCCalendar.MoveCursorToDateCell;
var aRect: TRect; aPoint: TPoint;
begin
  aRect := GetDayRect(Day);
  aPoint := ClientToScreen(aRect.TopLeft);
  with aRect do
    SetCursorPos(aPoint.x + (Right - Left) div 2, 
      aPoint.y + (Bottom - Top) div 2);
end;

function TDCCalendar.GetTitleRect: TRect;
begin
  Result := Rect(EdgeDepth, EdgeDepth, Width - EdgeDepth, 
    EdgeDepth + cellH);
end;

function TDCCalendar.GetHeaderRect: TRect;
begin
  Result := Rect(EdgeDepth, EdgeDepth + cellH, Width - EdgeDepth, 
    EdgeDepth + 2 * cellH);
end;

function TDCCalendar.GetPrevMonthRect: TRect;
begin
  Result := Rect(EdgeDepth + Ord(caShowWeekNum in FOptions) * cellW, 
    EdgeDepth + 2 * cellH, EdgeDepth + (Offset1stDay +
    Ord(caShowWeekNum in FOptions)) * cellW, EdgeDepth + 3 * cellH);
end;

function TDCCalendar.GetNextMonthRect: TRect;
var nmd: integer;
begin
  nmd := MaxWeeksPerMonth * DaysPerWeek - DaysThisMonth - Offset1stDay;
  if (nmd + DaysPerWeek - 1) div DaysPerWeek > 1 then
    Result := Rect(
      EdgeDepth + Ord(caShowWeekNum in FOptions) * cellW, 
      EdgeDepth + MaxWeeksPerMonth * cellH, 
      EdgeDepth + (DaysPerWeek + Ord(caShowWeekNum in FOptions)) * cellW,
      EdgeDepth + (MaxWeeksPerMonth + 2) * cellH)
  else
    Result := Rect(
      EdgeDepth + (DaysPerWeek - nmd + Ord(caShowWeekNum in FOptions)) * cellW, 
      EdgeDepth + (MaxWeeksPerMonth + 1) * cellH, 
      EdgeDepth + (DaysPerWeek + Ord(caShowWeekNum in FOptions)) * cellW, 
      EdgeDepth + (MaxWeeksPerMonth + 2) * cellH);
end;

function TDCCalendar.GetWeekNumRect: TRect;
begin
  Result := Rect(EdgeDepth, EdgeDepth + 2 * cellH,
    EdgeDepth + Ord(caShowWeekNum in FOptions) * cellW,
    EdgeDepth + (MaxWeeksPerMonth + 2) * cellH);
end;

function TDCCalendar.GetDaysArea: TRect;
begin
  Result := Rect(
    EdgeDepth + Ord(caShowWeekNum in FOptions) * cellW, 
    EdgeDepth + 2 * cellH, 
    EdgeDepth + (DaysPerWeek + Ord(caShowWeekNum in FOptions)) * cellW, 
    EdgeDepth + (MaxWeeksPerMonth + 2) * cellH);
end;

function TDCCalendar.GetCellRect(aCol, aRow: integer): TRect;
var l, t: integer;
begin
  l := EdgeDepth + (aCol + Ord(caShowWeekNum in FOptions)) * cellW;
  t := EdgeDepth + aRow * cellH;
  Result := Rect(l, t, l + cellW, t + cellH);
end;

function TDCCalendar.GetDayRect(aDay: integer): TRect;
begin
  Result := GetCellRect((Offset1stDay + aDay - 1) mod DaysPerWeek,
    (Offset1stDay + aDay - 1) div DaysPerWeek + 2);
end;

function TDCCalendar.GetSelectedArea(aType: TCaSelArea): TRect;
var aYear, aMonth, aDay: word; r1, r2: TRect; D: integer; S: TDateTime;

  function GetDateLine(aDate: TDateTime): TRect;
  begin
    S := EncodeDate(Year, Month, 1);
    if (S - Offset1stDay <= aDate) and (aDate <=
        S + MaxWeeksPerMonth * DaysPerWeek - Offset1stDay - 1) then
    begin
      DecodeDate(aDate, aYear, aMonth, aDay);
      if aDate < S then
        D := aDay - DaysPerMonth(aYear, aMonth) - DayOfWeek(aDate)
      else if aDate > S + DaysThisMonth - 1 then
        D := aDay + DaysThisMonth - DayOfWeek(aDate)
      else
        D := aDay - DayOfWeek(aDate);
      UnionRect(Result,GetDayRect(D + 1), GetDayRect(D + DaysPerWeek));
    end
    else
      Result := Rect(0,0,0,0);
  end;

begin
  case aType of
    saWhole: begin {between StartDate & FinishDate (FDate & FAnchor)}
      DecodeDate(StartDate,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
        r1 := GetDayRect(aDay - DayOfWeek(StartDate) + 1)
      else
        r1 := GetCellRect(0,2);
      DecodeDate(FinishDate,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
        r2 := GetDayRect(aDay - DayOfWeek(FinishDate) + DaysPerWeek)
      else
        r2 := GetCellRect(DaysPerWeek - 1, MaxWeeksPerMonth + 1);
      UnionRect(Result, r1, r2);
    end;
    saChanges: begin {between FOldDate & FDate}
      r1 := GetDayRect(Day - DayOfWeek(FDate) + 1);
      DecodeDate(FOldDate,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
        r2 := GetDayRect(aDay - DayOfWeek(FOldDate) + DaysPerWeek)
      else if (aYear < Year) or ((aYear = Year) and (aMonth < Month)) then
        r2 := GetCellRect(0,2)
      else
        r2 := GetCellRect(DaysPerWeek - 1, MaxWeeksPerMonth + 1);
      UnionRect(Result, r1, r2);
    end;
    saOldSelect: begin {between FOldDate & FAnchor}
      DecodeDate(FOldDate,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
        r1 := GetDayRect(aDay - DayOfWeek(FOldDate) + 1)
      else
        UnionRect(r1, GetCellRect(0, 2), GetCellRect(DaysPerWeek - 1, 2));
      DecodeDate(FAnchor,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
        r2 := GetDayRect(aDay - DayOfWeek(FAnchor) + DaysPerWeek)
      else if (aYear < Year) or ((aYear = Year) and (aMonth < Month)) then
        UnionRect(r2, GetCellRect(0, 2), GetCellRect(DaysPerWeek - 1, 2))
      else
        UnionRect(r2, GetCellRect(0, MaxWeeksPerMonth + 1),
          GetCellRect(DaysPerWeek - 1, MaxWeeksPerMonth + 1));
      UnionRect(Result, r1, r2);
    end;
    saFirstCell: begin
      DecodeDate(StartDate,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
        Result := GetDayRect(aDay)
      else
        Result := Rect(0,0,0,0);
    end;
    saFirstLine: Result := GetDateLine(StartDate);
    saLastCell: Result := GetDateLine(FinishDate);
    saLastLine: begin
      DecodeDate(FinishDate,aYear,aMonth,aDay);
      if (aYear = Year) and (aMonth = Month) then
      UnionRect(Result,GetDayRect(aDay - DayOfWeek(FinishDate) + 1),
        GetDayRect(aDay - DayOfWeek(FinishDate) + DaysPerWeek))
      else
        Result := Rect(0,0,0,0);
    end;
    else
      Result := GetDaysArea;
  end;
end;

function TDCCalendar.GetRightMarginRect: TRect;
begin
  Result := Rect((DaysPerWeek + Ord(caShowWeekNum in FOptions)) * cellW +
    EdgeDepth, EdgeDepth + 2 * cellH, Width - EdgeDepth, EdgeDepth +
    8 * cellH);
end;

function TDCCalendar.GetBottomMarginRect: TRect;
begin
  Result := Rect(EdgeDepth, EdgeDepth + 8 * cellH, Width - EdgeDepth, 
    Height - EdgeDepth);
end;

function TDCCalendar.GetToday: integer;
var aYear, aMonth, aDay: word;
begin
  DecodeDate(SysUtils.Date, aYear, aMonth, aDay);
  Result := aDay;
end;

procedure TDCCalendar.UpdateDate;
var r: TRect; aYear, aMonth, aDay: word;
begin
  if not IsMonthChg and FMultiSelect and not ContinueSelect then
  begin {clear old selection}
    r := GetSelectedArea(saOldSelect);
    InvalidateRect(Handle, @r, False);
  end;
  if not InSelectMode then
    FAnchor := FDate;
  FUnderMouseDay := Day;
  if IsMonthChg then
  begin
    {redraw whole component if year and/or month changed}
    Offset1stDay := DayOfWeek(EncodeDate(Year, Month, 1)) - 1;
    CurMonth1stWeekNum :=
      Round(EncodeDate(Year, Month, 1) - EncodeDate(Year, 1, 1) +
      DayOfWeek(EncodeDate(Year, 1, 1)) - 1) div DaysPerWeek + 1;
    Repaint; {redraw all}
  end
  else if FMultiSelect then
  begin
    if ContinueSelect then
      r := GetSelectedArea(saChanges) {draw only changes of selected area}
    else
      {draw new selection - it can't take more than 1 line}
      r := GetSelectedArea(saFirstLine);
    InvalidateRect(Handle, @r, False);
  end
  else
  begin
    {redraw only old & new lines}
    DecodeDate(FOldDate, aYear, aMonth, aDay);
    UnionRect(r, GetDayRect(aDay - DayOfWeek(FOldDate) + 1),
      GetDayRect(aDay - DayOfWeek(FOldDate) + DaysPerWeek));
    InvalidateRect(Handle, @r, False);
    UnionRect(r, GetDayRect(Day - DayOfWeek(FDate) + 1),
      GetDayRect(Day - DayOfWeek(FDate) + DaysPerWeek));
    InvalidateRect(Handle, @r, False);
  end;
  ChangeDate;
end;

procedure TDCCalendar.ChangeDate;
begin
  if Assigned(FOnChange) then
    FOnChange(Self, FDate);
end;

procedure TDCCalendar.PopupDropDown;
begin
  if (FPopup <> nil) and not FPopupVisible then
  begin
    if CanFocus then
      SetFocus;
    FPopup.Date := FDate;
    FPopup.Show(GetTitleRect);
    FPopupVisible := True;
    {$IFDEF WIN32}
    dcCa := Self;
    WHook := SetWindowsHookEx(WH_MOUSE, @CallWndProcHook, 0, GetCurrentThreadId);
    {$ENDIF}
  end;
end;

procedure TDCCalendar.PopupCloseUp(Sender: TObject; Accept: Boolean);
var cYear, cMonth, cDay, pYear, pMonth, pDay: word;
begin
  if (FPopup <> nil) and FPopupVisible then
  begin
    if GetCapture <> 0 then
      SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
    {$IFDEF WIN32}
    UnhookWindowsHookEx(WHook);
    {$ENDIF}
    FPopup.Hide;
    try
      if CanFocus then
        SetFocus;
      if Accept then
      begin
        DecodeDate(FPopup.Date, pYear, pMonth, pDay);
        DecodeDate(Self.FDate, cYear, cMonth, cDay);
        pDay := Min(cDay, DaysPerMonth(pYear, pMonth));
        Date := EncodeDate(pYear, pMonth, pDay); {not FDate}
      end;
    finally
      FPopupVisible := False;
    end;
  end;
end;

function TDCCalendar.Execute(X, Y: integer): Boolean;
var SaveDate: TDateTime; SaveParent: TWinControl;
begin
  InDialogMode := True;
  Result := False;
  SaveDate := FDate;
  SaveParent := Parent;
  PopupForm := TForm.Create(nil);
  try
    RecreateWnd;
    Left := 0;
    Top := 0;
    Parent := PopupForm;
    Visible := True;
    with PopupForm do
    begin
      AutoScroll := False;
      {$IFDEF D4}
      AutoSize := True;
      {$ELSE}
      Width := Self.Width;
      Height := Self.Height;
      {$ENDIF}
      BorderStyle := bsNone;
      KeyPreview := True;
      if X + Self.Width > Screen.Width then
        X := Screen.Width - Self.Width;
      if Y + Self.Height > Screen.Height then
        Y := Screen.Height - Self.Height;
      Left := X;
      Top := Y;
      Visible := False;
      Result := ShowModal = mrOK;
    end;
  finally
    Visible := False;
    Parent := SaveParent;
    if not Result then
      SetDate(SaveDate); {restore original date}
    PopupForm.Free;
    InDialogMode := False;
  end;
end;

end.
