unit WinG;  {WinG import unit for Borland Pascal}

interface
uses Windows, SysUtils, Classes, Graphics, Dialogs, Forms;

const
   DirDll = 'WING32.DLL';
   PC_RESERVED = 1;
   PC_EXPLICIT = 2;
   PC_NOCOLLAPSE = 4;

type
   EWinGError = class(Exception);

   { Declaracao do DLL WinG }
   TWinGDither=(winG4x4Dispersed,winG8x8Dispersed,winG4x4Clustered);

   TWinG = class(TPersistent)
   private
      FWidth: Word;
      FHeight: Word;
      FBitmap: HBITMAP;
      FHandle: HDC;
      FPixels: PChar;
      FCanvas: TCanvas;
      FOnChange: TNotifyEvent;

      BitmapInfo: record
         bmiHeader: TBitmapInfoHeader;
         bmiColors: array [0..255] of TRGBQUAD;
      end;
      ghBitmapMonochrome: HBITMAP;

      procedure SetWidth( Value: Word );
      procedure SetHeight( Value: Word );
      procedure SetPixelAt(X, Y: Integer; Value: Byte);
      function GetPixelAt(X, Y: Integer): Byte;
      function GetColorTable: PRGBQUAD;
      function GetWidthBytes: Word;
      function GetCanvas: TCanvas;
      procedure CreateWinG;
      procedure DeleteWinG;

   public
      property Bitmap: HBITMAP read FBitmap;
      property Canvas: TCanvas read GetCanvas;
      property Handle: HDC read FHandle;
      property PixelAt[X,Y: Integer]: Byte read GetPixelAt write SetPixelAt;
      property Pixels: PChar read FPixels;
      property ColorTable: PRGBQUAD read GetColorTable;
      property WidthBytes: Word read GetWidthBytes;

      constructor Create( wWidth, wHeight: Word );
      destructor Destroy; override;
      function BitBlt( hdcDst: HDC; nXOriginDst, nYOriginDst, nWidthDst,
         nHeightDst: Integer; nXOriginSrc, nYOriginSrc: Integer): Boolean;
      function StretchBlt( hdcDst: HDC; nXOriginDst, nYOriginDst, nWidthDst,
        nHeightDst: Integer; nXOriginSrc, nYOriginSrc, nWidthSrc,
		    nHeightSrc: Integer): Boolean;
      procedure Copy(Dest: TWinG; SrcX, SrcY, DestX, DestY, CopyWidth,
        CopyHeight: Integer);
      procedure TransCopy(Dest: TWinG; SrcX, SrcY, DestX, DestY,
        CopyWidth, CopyHeight: Integer; TransColor: Byte);

   published
      property Width: Word read FWidth write SetWidth;
      property Height: Word read FHeight write SetHeight;

      property OnChange: TNotifyEvent read FOnChange write FOnChange;
   end;


	function WinGCreateDC: hDC; stdcall; external DirDll;
	function WinGRecommendDIBFormat(pFormat:PBitmapInfo):boolean;
       stdcall; external DirDll;
	function WinGCreateBitmap(WinGDC:hDC; pHeader:pBitmapInfo;
		ppBits:pointer):hBitmap; stdcall; external DirDll;
	function WinGGetDIBPointer(WinGBitmap:hBitmap; pHeader:pBitmapInfo):pointer;
       stdcall; external DirDll;
	function WinGGetDIBColorTable(WinGDC:hDC; StartIndex, NumberOfEntries:word;
		pColors:pointer):word; stdcall; external DirDll;
	function WinGSetDIBColorTable(WinGDC:hDC; StartIndex, NumberOfEntries:word;
		pColors:pointer):word; stdcall; external DirDll;
	function WinGCreateHalftonePalette: hPalette; stdcall; external DirDll;
	function WinGCreateHalftoneBrush(context:hDC; crColor:tColorRef;
		ditherType:tWinGDither):hBrush; stdcall; external DirDll;
	function WinGBitBlt(hdcDst:hDC; nXOriginDst, nYOriginDst, nWidthDst,
		nHeightDst:integer; hdcSrc:hDC; nXOriginSrc, nYOriginSrc:integer):
       boolean; stdcall; external DirDll;
	function WinGStretchBlt(hdcDst:hDC; nXOriginDst, nYOriginDst, nWidthDst,
		nHeightDst:integer; hdcSrc:hDC; nXOriginSrc, nYOriginSrc, nWidthSrc,
		nHeightSrc:integer):boolean; stdcall; external DirDll;

  function ClipCopyParameters( SrcWidth, SrcHeight, DestWidth,
    DestHeight: Integer; var DestX, DestY, SrcX, SrcY, CopyWidth,
    CopyHeight: Integer): Boolean;



implementation

uses Dib256;



//***** TWinG ******

constructor TWinG.Create( wWidth, wHeight: Word );
begin
   inherited Create;

{REG}

   FWidth := wWidth;
   FHeight := wHeight;

   FCanvas := TCanvas.Create;

   CreateWinG;
end;


destructor TWinG.Destroy;
begin
  DeleteWinG;

  FCanvas.Free;

	inherited Destroy;
end;

function TWinG.BitBlt( hdcDst: HDC; nXOriginDst, nYOriginDst, nWidthDst, nHeightDst: Integer;
   nXOriginSrc, nYOriginSrc: Integer): Boolean;
begin
   result := WinGBitBlt( hdcDst, nXOriginDst, nYOriginDst, nWidthDst,
      nHeightDst, FHandle, nXOriginSrc, nYOriginSrc );
end;

function TWinG.StretchBlt( hdcDst: HDC; nXOriginDst, nYOriginDst, nWidthDst,
   nHeightDst: Integer; nXOriginSrc, nYOriginSrc, nWidthSrc,
   nHeightSrc: Integer): Boolean;
begin
   result := WinGStretchBlt( hdcDst, nXOriginDst, nYOriginDst, nWidthDst,
      nHeightDst, FHandle, nXOriginSrc, nYOriginSrc, nWidthSrc,
      nHeightSrc );
end;

procedure TWinG.SetWidth( Value: Word );
begin
   if (Value = FWidth) then
      exit;

   DeleteWinG;
   FWidth := Value;
   CreateWinG;

   if Assigned(FOnChange) then OnChange(Self);
end;

procedure TWinG.SetHeight( Value: Word );
begin
   if (Value = FHeight) then
      exit;

   DeleteWinG;
   FHeight := Value;
   CreateWinG;

   if Assigned(FOnChange) then OnChange(Self);
end;


function TWinG.GetPixelAt(X,Y: Integer): Byte;
begin
  if (X < 0) or (X >= Width) or (Y < 0) or (Y >= Height) then
    raise EWinGError.Create('Pixel coordinates out of canvas')
  else
    Result := Byte((PChar(Pixels) + X + Y * WidthBytes)^);
end;

procedure TWinG.SetPixelAt(X, Y: Integer; Value: Byte);
begin
  if (X < 0) or (X >= Width) or (Y < 0) or (Y >= Height) then
    raise EWinGError.Create('Pixel coordinates out of canvas')
  else
    (PChar(Pixels) + X + Y * WidthBytes)^ := Char(Value);
end;


function TWinG.GetColorTable: PRGBQUAD;
begin
   Result := @BitmapInfo.bmiColors;
end;


function TWinG.GetWidthBytes: Word;
begin
   Result := (FWidth + 3) and $FFFC;
end;


function TWinG.GetCanvas: TCanvas;
begin
  if FHandle = 0 then raise EWinGError.Create('Handle not valid');
  FCanvas.Handle := FHandle;
  Result := FCanvas;
end;


procedure TWinG.CreateWinG;
begin
   { init. bitmap structure with recommend format }
   if WinGRecommendDibFormat( @BitmapInfo ) = False then
      raise EWinGError.Create('Calling ''WinGRecommendDibFormat''');

   BitmapInfo.bmiHeader.biBitCount := 8;
   BitmapInfo.bmiHeader.biClrUsed := 256;
   BitmapInfo.bmiHeader.biCompression := BI_RGB;

   { init. dimensions, overlooking biHeight and setting allways the DIB in
   Top-Down format }
   BitmapInfo.bmiHeader.biWidth := Integer(FWidth);
   BitmapInfo.bmiHeader.biHeight := Abs(Integer(FHeight)) * -1;

   // Create WinG DC
   FHandle := WinGCreateDC;
   if (FHandle = 0) then
      raise EWinGError.Create('Can''t create WinG DC');
   FCanvas.Handle := FHandle;

   // Create WinG Bitmap
   FBitmap := WinGCreateBitmap( FHandle, @BitmapInfo, @FPixels );
   if (FBitmap = 0) then
      raise EWinGError.Create('Can''t create WinG Bitmap');

   ghBitmapMonochrome := SelectObject( FHandle, FBitmap );
end;

procedure TWinG.DeleteWinG;
begin
   FCanvas.Handle := 0;

   if FBitmap <> 0 then
   begin
      SelectObject( FHandle, ghBitmapMonochrome );
      DeleteObject( FBitmap );
   end;
   FBitmap := 0;

   if FHandle <> 0 then
      DeleteDC( FHandle );
   FHandle := 0;
end;


procedure TWinG.Copy(Dest: TWinG; SrcX, SrcY, DestX, DestY,
  CopyWidth, CopyHeight: Integer);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Width, Height, Dest.Width, Dest.Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Pixels) + SrcX + SrcY * WidthBytes;
    PDest := PChar(Dest.Pixels) + DestX + DestY * Dest.WidthBytes;
    CopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, Dest.WidthBytes,
      WidthBytes);
  end;
end;

procedure TWinG.TransCopy(Dest: TWinG; SrcX, SrcY, DestX, DestY,
  CopyWidth, CopyHeight: Integer; TransColor: Byte);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Width, Height, Dest.Width, Dest.Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Pixels) + SrcX + SrcY * WidthBytes;
    PDest := PChar(Dest.Pixels) + DestX + DestY * Dest.WidthBytes;
    TransCopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, Dest.WidthBytes,
      WidthBytes, TransColor);
  end;
end;



function ClipCopyParameters(
  SrcWidth, SrcHeight, DestWidth, DestHeight: Integer;
  var DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight: Integer): Boolean;
{ Clips source and destination parameters, returns False if nothing is left
 to
  be copied. }
begin
  Result := False;
  if DestX >= DestWidth then Exit;
  if DestY >= DestHeight then Exit;
  if DestX < 0
  then begin
    CopyWidth := CopyWidth + DestX;
    SrcX := SrcX - DestX;
    DestX := 0;
  end;
  if DestY < 0
  then begin
    CopyHeight := CopyHeight + DestY;
    SrcY := SrcY - DestY;
    DestY := 0;
  end;
  if SrcX >= SrcWidth then Exit;
  if SrcY >= SrcHeight then Exit;
  { corrections start here }
  if SrcX < 0
  then begin
    CopyWidth := CopyWidth + SrcX;
    DestX := DestX - SrcX;
    SrcX := 0;
  end;
  if SrcY < 0
  then begin
    CopyHeight := CopyHeight + SrcY;
    DestY := DestY - SrcY;
    SrcY := 0;
  end;
  { corrections end here }

  if DestX + CopyWidth >= DestWidth then CopyWidth := DestWidth - DestX;
  if DestY + CopyHeight >= DestHeight then CopyHeight := DestHeight -
DestY;
  if SrcX + CopyWidth >= SrcWidth then CopyWidth := SrcWidth - SrcX;
  if SrcY + CopyHeight >= SrcHeight then CopyHeight := SrcHeight - SrcY;
  if CopyWidth <= 0 then Exit;
  if CopyHeight <= 0 then Exit;

  Result := True;
end;


end.

