{$DEFINE DEBUG}
{$B-,H+,I+,X+}
{$IFDEF DEBUG}
{$D+,Q+,R+,Y+}
{$ELSE}
{$D-,Q-,R-,Y-}
{$ENDIF}
unit cStrings;

interface



{                                                                              }
{ Strings unit v0.07 (L0)                                                      }
{                                                                              }
{   A collection of string functions.                                          }
{                                                                              }
{                                                                              }
{ This unit is copyrighted  1999 by David Butler (david@e.co.za)              }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   1999/10/19  v0.01  Spawned from Maths unit.                                }
{   1999/10/26  v0.02  Documentation.                                          }
{   1999/10/30  v0.03  Added Count, Reverse.                                   }
{                      Implemented the Boyer-Moore-Horspool pattern searching  }
{                      algorithm in assembly.                                  }
{   1999/10/31  v0.04  Coded Match function in assembly.                       }
{                      Added Replace, Count, PadInside.                        }
{   1999/11/06  v0.05  261 lines interface, 772 lines implementation.          }
{                      Added Remove, TrimEllipse.                              }
{   1999/11/09  v0.06  Added Pack functions.                                   }
{   1999/11/17  v0.07  Added Cut to Pad )                                      }
{                      Added PosN, Before, After and Between.                  }
{                      Added CountWords.                                       }
{                                                                              }
{                                                                              }




{                                                                              }
{ Global types                                                                 }
{                                                                              }
type
  CharSet = Set of Char;
  CharSetArray = Array of CharSet;
  StringArray = Array of String;

Function AsCharSetArray (const S : Array of CharSet) : CharSetArray;
{ Returns an open array of CharSet as a dynamic array of CharSet             }



{                                                                              }
{ Character constants                                                          }
{                                                                              }
const
  // ASCII codes
  ASCII_NULL      = #0;
  ASCII_BELL      = #7;
  ASCII_BS        = #8;
  ASCII_HT        = #9;
  ASCII_LF        = #10;
  ASCII_CR        = #13;
  ASCII_EOF       = #26;
  ASCII_ESC       = #27;
  ASCII_SP        = #32;

  c_Tab           = ASCII_HT;
  c_Space         = ASCII_SP;
  c_EOL           = ASCII_CR + ASCII_LF;
  c_DecimalPoint  = '.';


{                                                                              }
{ Character set constants                                                      }
{                                                                              }
const
  cs_Everything       = [#0..#255];
  cs_WhiteSpace       = [#0..#32];
  cs_AlphaLow         = ['a'..'z'];
  cs_AlphaHigh        = ['A'..'Z'];
  cs_Numeric          = ['0'..'9'];
  cs_Alpha            = cs_AlphaLow + cs_AlphaHigh;
  cs_AlphaNumerical   = cs_Numeric + cs_Alpha;
  cs_Exponent         = ['E', 'e'];
  cs_HexDigit         = cs_Numeric + ['A'..'F', 'a'..'f'];
  cs_OctalDigit       = ['0'..'7'];
  cs_BinaryDigit      = ['0'..'1'];
  cs_Sign             = ['+', '-'];
  cs_Parentheses      = ['(', ')'];
  cs_CurlyBrackets    = ['{', '}'];
  cs_BlockBrackets    = ['[', ']'];
  cs_Punctuation      = ['.', ',', ':', '/', '?', '<', '>', ';', '"', '''',
                         '[', ']', '{', '}', '+', '=', '-', '\', '(', ')', '*',
                         '&', '^', '%', '$', '#', '@', '!', '`', '~'];
  cs_LowerA           = ['a', '', '', '', '', ''];
  cs_LowerE           = ['e', '', '', '', ''];
  cs_LowerI           = ['i', '', '', '', ''];
  cs_LowerN           = ['n', ''];
  cs_LowerO           = ['o', '', '', '', ''];
  cs_LowerU           = ['u', '', '', '', ''];
  cs_LowerY           = ['y', ''];
  cs_Lower            = cs_AlphaLow + cs_LowerA + cs_LowerE + cs_LowerI +
                        cs_LowerN + cs_LowerO + cs_LowerU + cs_LowerY;
  cs_UpperA           = ['A', '', ''];
  cs_UpperE           = ['E', ''];
  cs_UpperN           = ['N', ''];
  cs_UpperO           = ['O', ''];
  cs_UpperU           = ['U', ''];
  cs_Upper            = cs_AlphaHigh + cs_UpperA + cs_UpperE + cs_UpperN +
                        cs_UpperO + cs_UpperU;
  cs_Word             = cs_Lower + cs_Upper + ['-', '='];
  cs_Text             = cs_Word + cs_Punctuation + cs_WhiteSpace + cs_Numeric;
  cs_WordDelim        = cs_Text - cs_Word;




{                                                                              }
{ Copy                                                                         }
{   Variantions on Delphi's Copy. Just like Delphi's Copy, illegal values for  }
{   Start (<1,>len), Stop (<start,>len) and Count (<0,>end) are tolerated.     }
{                                                                              }
Function CopyRange (const S : String; const Start, Stop : Integer) : String;
Function CopyFrom (const S : String; const Start : Integer) : String;
Function CopyLeft (const S : String; const Count : Integer) : String;
Function CopyRight (const S : String; const Count : Integer = 1) : String;



{                                                                              }
{ Paste                                                                        }
{   Paste from Dest [DestPos] to Source [SourcePos]                            }
{                                                                              }
Procedure Paste (var Dest : String; const DestPos : Integer;
          const Source : String; const SourcePos, Count : Integer);



{                                                                              }
{ Trim                                                                         }
{   TrimQuotes removes quotes around and double quotes inside a string.        }
{   TrimEllipse trims the string and puts '...' at the end if it's longer      }
{     than Length.                                                             }
{                                                                              }
Function TrimLeft (const S : String; const TrimSet : CharSet = cs_WhiteSpace) : String;
Function TrimRight (const S : String; const TrimSet : CharSet = cs_WhiteSpace) : String;
Function Trim (const S : String; const TrimSet : CharSet = cs_WhiteSpace) : String;          // Faster than calling TrimLeft (TrimRight (S))
Function TrimQuotes (const S : String) : String;
Function TrimEllipse (const S : String; const Length : Integer) : String;


{                                                                              }
{ Dup                                                                          }
{   Illegal values for Count (<0) are tolerated.                               }
{                                                                              }
Function Dup (const S : String; const Count : Integer) : String; overload                    // Pretty fast
Function Dup (const Ch : Char; const Count : Integer) : String; overload;                    // Blazing



{                                                                              }
{ Reverse                                                                      }
{                                                                              }
Function Reverse (const S : String) : String;
Function IsPalindrome (const S : String) : Boolean;                                          // Reverse (S) = S



{                                                                              }
{ Match                                                                        }
{  True if Find matches S [Pos] (or S [Pos..Pos+Count-1])                      }
{  Returns False if Pos or Count is invalid                                    }
{                                                                              }
Function Match (const Find : CharSet; const S : String; const Pos : Integer = 1;
         const Count : Integer = 1) : Boolean; overload;
Function Match (const Find : CharSetArray; const S : String; const Pos : Integer = 1)
         : Boolean; overload;
Function Match (const Find, S : String; const Pos : Integer = 1) : Boolean; overload;        // Blazing



{                                                                              }
{ PosNext                                                                      }
{   Returns first Match of Find in S after LastPos.                            }
{   To find the first match, set LastPos to 0.                                 }
{   Returns 0 if not found or illegal value for LastPos (<0,>length(s))        }
{                                                                              }
Function PosNext (const Find : CharSet; const S : String;
         const LastPos : Integer = 0) : Integer; overload;
Function PosNext (const Find : CharSetArray; const S : String;
         const LastPos : Integer = 0) : Integer; overload;
Function PosNext (const Find : String; const S : String;
         const LastPos : Integer = 0) : Integer; overload;
Function PosPrev (const Find : String; const S : String;
         const LastPos : Integer = 0) : Integer;

type
  { Implements the Boyer-Moore-Horspool pattern searching algorithm.           }
  { Converted to a class and rewritten in assembly by David Butler             }
  {  (david@e.co.za) from a highly optimized Pascal unit BMH 1.11a written by  }
  {  Jody R Cairns (jodyc@cs.mun.ca) as she took it from the 'Handbook of      }
  {  Algorithms and Data Structures in Pascal and C', Second Edition,          }
  {  by G.H Gonnet and  R. Baeza-Yates.                                        }
  {                                                                            }
  { For longer values of Find (say length > 5) the function is faster than     }
  {   a brute force find (Delphi's Pos function), the time to Create the       }
  {   object excluded.                                                         }
  { When to use: If you are going to do a find a few times for the same        }
  {   string, that has length > 5.                                             }
  { See the implementation for speed comparisons                               }
  TBMHSearcher = class
    private
    FTable : Array [#0..#255] of Integer;
    FFind  : String;

    public
    Constructor Create (const Find : String);
    Function PosNext (const S : String; const LastPos : Integer = 0) : Integer;
  end;



{                                                                              }
{ Replace                                                                      }
{   QuoteText puts Quotes around S and replaces inside S with double Quotes.   }
{   Remove removes all characters in Ch from S.                                }
{   RemoveDup removes consequetive occurances of a character from Ch.          }
{                                                                              }
Function Replace (const Find, Replace, S : String) : String; overload;
Function Replace (const Find, Replace : Char; const S : String) : String; overload;
Function Replace (const Find : CharSet; const Replace : Char; const S : String) : String; overload;
Function QuoteText (const S : String; const Quotes : Char) : String;
Function Remove (const Ch : CharSet; const S : String) : String;
Function RemoveDup (const Ch : CharSet; const S : String) : String;



{                                                                              }
{ Count                                                                        }
{   Returns the number of occurances of Find in S.                             }
{                                                                              }
Function Count (const Find, S : String) : Integer; overload;
Function Count (const Find : CharSet; const S : String) : Integer; overload;
Function Count (const Find : CharSetArray; const S : String) : Integer; overload;
Function CountWords (const S : String) : Integer;



{                                                                              }
{ PosN                                                                         }
{   Finds the Nth occurance of Find in S from the left or the right.           }
{                                                                              }
Function PosN (const Find, S : String; const N : Integer = 1;
         const FromRight : Boolean = False) : Integer;



{                                                                              }
{ Before, After and Between                                                    }
{                                                                              }
{   Before and After returns the text before and after the Nth occurance of    }
{     Find in S, from the left and right, respectively, of S. It returns '' if }
{     N occurances of Find does not exist in S. It returns S if Find='' or if  }
{     N<1.                                                                     }
{   Between returns the text between the Nth occurance of Left from the left   }
{     of S and the Nth occurance of Right from the right of S. If Left='' or   }
{     LeftN=0 then returns from the left of S. If Right='' or RightN=0 then    }
{     returns from the right of S. If LeftN occurances of Left or RightN       }
{     occurances of Right in After (Left, S, LeftN) does not exist then it     }
{     returns ''.                                                              }
{     ie Between := Before (Right, After (Left, S, LeftN), RightN;)            }
{                                                                              }
Function Before (const Find, S : String; const N : Integer = 1) : String;
Function After (const Find, S : String; const N : Integer = 1) : String;
Function Between (const Left, Right, S : String; const LeftN : Integer = 1;
         const RightN : Integer = 1) : String;


{                                                                              }
{ Split                                                                        }
{   Splits S into pieces seperated by Delimiter. If Delimiter='' or S='' then  }
{   returns an empty list. If Token not found in S returns list with one       }
{   item, S.                                                                   }
{                                                                              }
Function Split (const S : String; const Delimiter : String = c_Space) : StringArray;



{                                                                              }
{ Pad                                                                          }
{   The default for Cut is False which won't shorten the string to Length      }
{   if Length < Length (S).                                                    }
{   PadLeft is equivalent to a right justify, PadRight a left justify,         }
{   Pad centering and PadInside a full justification.                          }
{                                                                              }
Function PadLeft (const S : String; const PadCh : Char; const Length : Integer;
         const Cut : Boolean = False) : String;
Function PadRight (const S : String; const PadCh : Char; const Length : Integer;
         const Cut : Boolean = False) : String;
Function Pad (const S : String; const PadCh : Char; const Length : Integer;
         const Cut : Boolean = False) : String;
Function PadInside (const S : String; const PadCh : Char; const Length : Integer) : String;



{                                                                              }
{ Type checking                                                                }
{                                                                              }
Function IsNumber (const S : String) : Boolean;
Function IsHexNumber (const S : String) : Boolean;
Function IsInteger (const S : String) : Boolean;
Function IsReal (const S : String) : Boolean;
Function IsScientificReal (const S : String) : Boolean;
Function IsQuotedString (const S : String) : Boolean;



{                                                                              }
{ Natural language                                                             }
{                                                                              }
Function Number (const Num : Int64; const USStyle : Boolean = False) : String; overload;
Function Number (const Num : Extended; const USStyle : Boolean = False) : String; overload;



{                                                                              }
{ Pack                                                                         }
{   Packs paramater (in its binary format) into a string                       }
{                                                                              }
Function Pack (const D : Int64) : String; overload;
Function Pack (const D : Integer) : String; overload;
Function Pack (const D : Byte) : String; overload;
Function Pack (const D : Word) : String; overload;
Function Pack (const D : String) : String; overload;
Function PackShortString (const D : ShortString) : String;
Function Pack (const D : Extended) : String; overload;
Function PackSingle (const D : Single) : String;
Function PackDouble (const D : Double) : String;
Function PackCurrency (const D : Currency) : String;
Function PackDateTime (const D : TDateTime) : String;
Function Pack (const D : Boolean) : String; overload;



implementation



uses
  // System units
  SysUtils,
  Math;



{                                                                              }
{                                                                              }
{                                                                              }
Function AsCharSetArray (const S : Array of CharSet) : CharSetArray;
var I : Integer;
  Begin
    SetLength (Result, High (S) + 1);
    For I := 0 to High (S) do
     Result [I] := S [I];
  End;



{                                                                              }
{ Copy                                                                         }
{                                                                              }
Function CopyRange (const S : String; const Start, Stop : Integer) : String;
  Begin
    Result := Copy (S, Start, Stop - Start + 1);
  End;

Function CopyFrom (const S : String; const Start : Integer) : String;
  Begin
    Result := Copy (S, Start, Length (S) - Start + 1);
  End;

Function CopyLeft (const S : String; const Count : Integer) : String;
  Begin
    Result := Copy (S, 1, Count);
  End;

Function CopyRight (const S : String; const Count : Integer) : String;
  Begin
    Result := Copy (S, Length (S) - Count + 1, Count);
  End;



{                                                                              }
{ Paste                                                                        }
{                                                                              }
Procedure Paste (var Dest : String; const DestPos : Integer;
          const Source : String; const SourcePos, Count : Integer);
  Begin
    Move (Source [SourcePos], Dest [DestPos], Min (Count, Length (Dest) - DestPos + 1));
  End;



{                                                                              }
{ Trim                                                                         }
{                                                                              }
Function TrimLeft (const S : String; const TrimSet : CharSet) : String;
var F : Integer;
  Begin
    F := 1;
    While (F <= Length (S)) and (S [F] in TrimSet) do
      Inc (F);
    Result := CopyFrom (S, F);
  End;

Function TrimRight (const S : String; const TrimSet : CharSet) : String;
var F : Integer;
  Begin
    F := Length (S);
    While (F >= 1) and (S [F] in TrimSet) do
      Dec (F);
    Result := CopyLeft (S, F);
  End;

Function Trim (const S : String; const TrimSet : CharSet) : String;
var F, G : Integer;
  Begin
    F := 1;
    While (F <= Length (S)) and (S [F] in TrimSet) do
      Inc (F);
    G := Length (S);
    While (G >= F) and (S [G] in TrimSet) do
      Dec (G);
    Result := CopyRange (S, F, G);
  End;

Function TrimQuotes (const S : String) : String;
  Begin
    if not IsQuotedString (S) then
      Result := S else
      Result := Replace (S [1] + S [1], S [1], Copy (S, 2, Length (S) - 2));
  End;

Function TrimEllipse (const S : String; const Length : Integer) : String;
  Begin
    if System.Length (S) <= Length then
      Result := S else
      if Length < 3 then
        Result := '' else
        Result := CopyLeft (S, Length - 3) + '...';
  End;



{                                                                              }
{ Dup                                                                          }
{                                                                              }
Function Dup (const S : String; const Count : Integer) : String;
var I, L : Integer;
  Begin
    L := Length (S);
    SetLength (Result, Count * L);
    For I := 0 to Count - 1 do
      Move (S [1], Result [I * L + 1], L);
  End;

Function Dup (const Ch : Char; const Count : Integer) : String;
  Begin
    SetLength (Result, Count);
    FillChar (Result [1], Count, Ord (Ch));
  End;



{                                                                              }
{ Reverse                                                                      }
{                                                                              }
Function Reverse (const S : String) : String;
var I, L : Integer;
  Begin
    L := Length (S);
    SetLength (Result, L);
    For I := 1 to L do
      Result [L - I + 1] := S [I];
  End;

Function IsPalindrome (const S : String) : Boolean;
var I, L : Integer;
  Begin
    L := Length (S);
    For I := 1 to L div 2 do
      if S [I] <> S [L - I + 1] then
        begin
          Result := False;
          exit;
        end;
    Result := True;
  End;



{                                                                              }
{ Match                                                                        }
{                                                                              }
Function Match (const Find : CharSet; const S : String; const Pos : Integer; const Count : Integer) : Boolean;
var I, PosEnd : Integer;
  Begin
    PosEnd := Pos + Count - 1;
    if (Find = []) or (Pos < 1) or (Count = 0) or (PosEnd > Length (S)) then
      begin
        Result := False;
        exit;
      end;

    For I := Pos to PosEnd do
      if not (S [I] in Find) then
        begin
          Result := False;
          exit;
        end;

    Result := True;
  End;

Function Match (const Find : CharSetArray; const S : String; const Pos : Integer) : Boolean;
var J, C : Integer;
  Begin
    C := Length (Find);
    if (C = 0) or (Pos < 1) or (Pos + C - 1 > Length (S)) then
      begin
        Result := False;
        exit;
      end;

    For J := 0 to C - 1 do
      if not (S [J + Pos] in Find [J]) then
        begin
          Result := False;
          exit;
        end;

    Result := True;
  End;

{ Highly optimized version of Match. Equivalent to, but much faster and more   }
{ memory efficient than: Find = Copy (S, Pos, Length (Find))                   }
{ Does compare in 32-bit chunks (CPU's native type)                            }
Function Match (const Find, S : String; const Pos : Integer) : Boolean;
  Asm
      push esi
      push edi
      push edx                    // save state

      push Pos
      push Find
      push S                      // push parameters
      pop edi                     // edi = S [1]
      pop esi                     // esi = Find [1]
      pop ecx                     // ecx = Pos
      cmp ecx, 1
      jb @NoMatch                 // if Pos < 1 then @NoMatch

      mov edx, [esi - 4]
      or edx, edx
      jz @NoMatch                 // if Length (Find) = 0 then @NoMatch
      add edx, ecx
      dec edx                     // edx = Pos + Length (Find) - 1

      cmp edx, [edi - 4]
      ja @NoMatch                 // if Pos + Length (Find) - 1 > Length (S) then @NoMatch

      add edi, ecx
      dec edi                     // edi = S [Pos]
      mov ecx, [esi - 4]          // ecx = Length (Find)

      // All the following code is an optimization of just two lines:         //
      //     rep cmsb                                                         //
      //     je @Match                                                        //
      mov dl, cl                                                              //
      and dl, $03                                                             //
      shr ecx, 2                                                              //
      jz @CheckMod                 { Length (Find) < 4 }                      //
                                                                              //
      { The following is faster than:  {}                                     //
      {     rep cmpsd                  {}                                     //
      {     jne @NoMatch               {}                                     //
      @c1:                             {}                                     //
        mov eax, [esi]                 {}                                     //
        cmp eax, [edi]                 {}                                     //
        jne @NoMatch                   {}                                     //
        add esi, 4                     {}                                     //
        add edi, 4                     {}                                     //
        dec ecx                        {}                                     //
        jnz @c1                        {}                                     //
                                                                              //
      or edx, edx                                                             //
      jz @Match                                                               //
                                                                              //
      { Check remaining dl (0-3) bytes   {}                                   //
    @CheckMod:                           {}                                   //
      mov eax, [esi]                     {}                                   //
      mov ecx, [edi]                     {}                                   //
      cmp al, cl                         {}                                   //
      jne @NoMatch                       {}                                   //
      dec dl                             {}                                   //
      jz @Match                          {}                                   //
      cmp ah, ch                         {}                                   //
      jne @NoMatch                       {}                                   //
      dec dl                             {}                                   //
      jz @Match                          {}                                   //
      and eax, $00ff0000                 {}                                   //
      and ecx, $00ff0000                 {}                                   //
      cmp eax, ecx                       {}                                   //
      je @Match                          {}                                   //

    @NoMatch:
      xor al, al                  // Result := False
      jmp @Fin

    @Match:
      mov al, 1                   // Result := True

    @Fin:
      pop edx                     // restore state
      pop edi
      pop esi
  End;



{                                                                              }
{ PosNext                                                                      }
{                                                                              }
Function PosNext (const Find : CharSet; const S : String; const LastPos : Integer) : Integer;
var I : Integer;
  Begin
    if Find = [] then
      begin
        Result := 0;
        exit;
      end;

    For I := Max (LastPos + 1, 1) to Length (S) do
      if S [I] in Find then
        begin
          Result := I;
          exit;
        end;

    Result := 0;
  End;

Function PosNext (const Find : CharSetArray; const S : String; const LastPos : Integer) : Integer;
var I, C : Integer;
  Begin
    C := Length (Find);
    if C = 0 then
      begin
        Result := 0;
        exit;
      end;

    For I := Max (LastPos + 1, 1) to Length (S) - C + 1 do
      if Match (Find, S, I) then
        begin
          Result := I;
          exit;
        end;

    Result := 0;
  End;

Function PosNext (const Find : String; const S : String; const LastPos : Integer = 0) : Integer;
var I : Integer;
  Begin
    if Find = '' then
      begin
        Result := 0;
        exit;
      end;

    For I := LastPos + 1 to Length (S) - Length (Find) + 1 do
      if Match (Find, S, I) then
        begin
          Result := I;
          exit;
        end;

    Result := 0;
  End;

Function PosPrev (const Find : String; const S : String; const LastPos : Integer = 0) : Integer;
var I, J : Integer;
  Begin
    if Find = '' then
      begin
        Result := 0;
        exit;
      end;

    if LastPos = 0 then
      J := Length (S) - Length (Find) + 1 else
      J := LastPos - 1;
    For I := J downto 1 do
      if Match (Find, S, I) then
        begin
          Result := I;
          exit;
        end;

    Result := 0;
  End;

{ Boyer-Moore-Horspool pattern searching                                       }
Constructor TBMHSearcher.Create (const Find : String);
  Begin
    inherited Create;
    FFind := Find;

    { Creates a Boyer-Moore-Horspool index table for the search string }
    asm
      push eax
      push ebx
      push edx
      push esi
      push edi                        // save state

      mov eax, self
      mov edx, FTable
      add edx, eax                    // edx = FTable [0]
      mov ebx, [eax + FFind]          // ebx = FFind [1]

      // FTable [0..255] := Length (FFind)                                    //
      mov eax, [ebx - 4]              { eax = Length (FFind) }                //
      mov edi, edx                    { edi = FTable [0] }                    //
      mov ecx, 256                                                            //
      rep stosd                                                               //

      // FTable [FFind [i = 1..Length (FFind)]] = Length (FFind) - i          //
      mov ecx, eax                    { ecx = Length (FFind) }                //
      dec ecx                                                                 //
      mov edi, edx                    { edi = FTable [0] }                    //
      xor esi, esi                    { esi = i - 1 = 0 }                     //
      xor edx, edx                                                            //
    @c1:                                                                      //
      mov dl, [ebx + esi]             { edx = FFind [i] }                     //
      mov [edi + edx * 4], ecx        { FTable [edx] := Length (FFind) - i }  //
      inc esi                                                                 //
      loop @c1                                                                //

      pop edi                         // restore state
      pop esi
      pop edx
      pop ebx
      pop eax
    end;
  End;

{ Results of a speed trial (you'll never be able to reproduce them :-) to      }
{ give you an idea of the speed relationship:                                  }
{                Len(Find)  Len(S)  System.Pos  TBMHSearcher.Pos  %Diff        }
{ Worst case        1        250    0.16 mil/s     0.04 mil/s      -300%       }
{                   1        60     0.6 mil/s      0.16 mil/s      -275%       }
{                   1        7      2.5 mil/s      1.2 mil/s       -108%       }
{                   5        250    0.16 mil/s     0.18 mil/s      +13%        }
{                   5        60     0.5 mil/s      0.6 mil/s       +20%        }
{                   5        7      3.5 mil/s      4.3 mil/s       +22%        }
{                   5        50000  713 /s         888 /s          +25%        }
{                   15       250    0.16 mil/s     0.36 mil/s      +157%       }
{                   15       60     0.7 mil/s      1.9 mil/s       +171%       }
{                   50       250    0.16 mil/s     1.35 mil/s      +744%       }
{ Best cast         250      50000  700 /s         30000 /s        +4286%      }
Function TBMHSearcher.PosNext (const S : String; const LastPos : Integer) : Integer;
  Asm
      push ebp
      push ebx
      push edx
      push esi
      push edi                                      // save state

      push LastPos
      push S
      push self                                     // push parameters
      pop ebp                                       // ebp = self
      pop eax                                       // eax = S
      pop ebx                                       // ebx = LastPos

      mov edx, [ebp + FFind]                        // edx = FFind
      or edx, edx
      jz @NoMatch                                   // if FFind = '' then NoMatch
      add ebx, [edx - 4]                            // ebx = counter, starting at LastPos + Length (FFind)

    // while ebx < Length (s)                                                                     //
    @WhileNotEnd:                                                                                 //
      cmp ebx, [eax - 4]                                                                          //
      ja @NoMatch                                                                                 //
                                                                                                  //
      mov ecx, [edx - 4]                            { loop count = Length (FFind) }               //
                                                                                                  //
      mov esi, eax                                                                                //
      add esi, ebx                                                                                //
      sub esi, ecx                                  { esi = S [1 + ebx - Length (FFind)] }        //
                                                                                                  //
      mov edi, edx                                  { edi = FFind [1] }                           //
                                                                                                  //
    { This is actually faster than REP CMPSB on a Pentium                     {}                  //
    @c1:                                                                      {}                  //
      cmpsb                                                                   {}                  //
      jne @NotEq                                                              {}                  //
      loop @c1                                                                {}                  //
                                                                                                  //
      jmp @Match                                    { Match found }                               //
                                                                                                  //
    @NotEq:                                                                                       //
      xor ecx, ecx                                                                                //
      mov cl, [eax + ebx - 1]                       { ecx = S [ebx] }                             //
                                                                                                  //
      add ebx, dword ptr [ebp + FTable + ecx * 4]   { Inc (ebx, FTable [ecx]) }                   //
                                                                                                  //
      jmp @WhileNotEnd                                                                            //

    @NoMatch:
      xor eax, eax                                  // Result = 0
      jmp @Fin

    @Match:
      mov eax, ebx
      inc eax
      sub eax, [edx - 4]                            // Result = ebx - Length (FFind) + 1

    @Fin:
      pop edi                                       // Restore state
      pop esi
      pop edx
      pop ebx
      pop ebp
  End;



{                                                                              }
{ PosN                                                                         }
{                                                                              }
Function PosN (const Find, S : String; const N : Integer = 1;
         const FromRight : Boolean = False) : Integer;
var F, I : Integer;
  Begin
    F := 0;
    For I := 1 to N do
      begin
        if FromRight then
          F := PosPrev (Find, S, F) else
          F := PosNext (Find, S, F);
        if F = 0 then
          break;
      end;
    Result := F;
  End;

Function Before (const Find, S : String; const N : Integer = 1) : String;
var I : Integer;
  Begin
    if (Find = '') or (N < 1) then
      begin
        Result := S;
        exit;
      end;

    I := PosN (Find, S, N, True);
    if I = 0 then
      Result := '' else
      Result := CopyLeft (S, I - 1);
  End;

Function After (const Find, S : String; const N : Integer = 1) : String;
var I : Integer;
  Begin
    if (Find = '') or (N < 1) then
      begin
        Result := S;
        exit;
      end;

    I := PosN (Find, S, N);
    if I = 0 then
      Result := '' else
      Result := CopyRight (S, I + 1);
  End;

Function Between (const Left, Right, S : String; const LeftN : Integer = 1;
         const RightN : Integer = 1) : String;
  Begin
    Result := Before (Right, After (Left, S, LeftN), RightN);
  End;



{                                                                              }
{ Split                                                                        }
{                                                                              }
Function Split (const S : String; const Delimiter : String = ' ') : StringArray;
var I, J, L : Integer;
  Begin
    SetLength (Result, 0);
    if (Delimiter = '') or (S = '') then
      exit;

    I := 0;
    L := 0;
    Repeat
      SetLength (Result, L + 1);
      J := PosNext (Delimiter, S, I);
      if J = 0 then
        Result [L] := CopyFrom (S, I + Length (Delimiter)) else
        begin
          Result [L] := CopyRange (S, I + Length (Delimiter), J - 1);
          I := J;
          Inc (L);
        end;
    Until J = 0;
  End;



{                                                                              }
{ Replace                                                                      }
{                                                                              }
Function Replace (const Find, Replace, S : String) : String;
var F, G, X, I,
    LR, LF      : Integer;
  Begin
    X := Count (Find, S);
    if X = 0 then
      begin
        Result := S;
        exit;
      end;

    LR := Length (Replace);
    LF := Length (Find);
    SetLength (Result, Length (S) + X * (LR - LF));
    F := PosNext (Find, S);
    Move (S [1], Result [1], F - 1);
    Move (Replace [1], Result [F], LR);
    I := F + LR;
    G := F + LF;
    Repeat
      F := PosNext (Find, S, G);
      if F > 0 then
        begin
          Move (S [G], Result [I], F - G);
          Inc (I, F - G);
          Move (Replace [1], Result [I], LR);
          Inc (I, LR);
          G := F + LF;
        end;
    Until F = 0;
    Move (S [G], Result [I], Length (S) - G + 1);
  End;

Function Replace (const Find, Replace : Char; const S : String) : String;
var I : Integer;
  Begin
    Result := S;
    For I := 1 to Length (S) do
      if S [I] = Find then
        Result [I] := Replace;
  End;

Function Replace (const Find : CharSet; const Replace : Char; const S : String) : String;
var I : Integer;
  Begin
    Result := S;
    For I := 1 to Length (S) do
      if S [I] in Find then
        Result [I] := Replace;
  End;

Function QuoteText (const S : String; const Quotes : Char) : String;
  Begin
    Result := Quotes + Replace (Quotes, Quotes + Quotes, S) + Quotes;
  End;

Function Remove (const Ch : CharSet; const S : String) : String;
var I : Integer;
  Begin
    Result := S;
    I := 1;
    While I <= Length (Result) do
      if Result [I] in Ch then
        Delete (Result, I, 1) else
        Inc (I);
  End;

Function RemoveDup (const Ch : CharSet; const S : String) : String;
var I : Integer;
  Begin
    Result := S;
    I := 2;
    While I <= Length (Result) do
      if (Result [I] in Ch) and (Result [I] = Result [I - 1]) then
        Delete (Result, I, 1) else
        Inc (I);
  End;



{                                                                              }
{ Count                                                                        }
{                                                                              }
Function Count (const Find, S : String) : Integer;
var I : Integer;
  Begin
    Result := 0;
    I := PosNext (Find, S);
    While I > 0 do
      begin
        Inc (Result);
        I := PosNext (Find, S, I + Length (Find) - 1);
      end;
  End;

Function Count (const Find : CharSetArray; const S : String) : Integer;
var I : Integer;
  Begin
    Result := 0;
    I := PosNext (Find, S);
    While I > 0 do
      begin
        Inc (Result);
        I := PosNext (Find, S, I + Length (Find) - 1);
      end;
  End;

Function Count (const Find : CharSet; const S : String) : Integer;
var I : Integer;
  Begin
    Result := 0;
    For I := 1 to Length (S) do
      if S [I] in Find then
        Inc (Result);
  End;

Function CountWords (const S : String) : Integer;
  Begin
    Result := Count (AsCharSetArray ([cs_Alpha, cs_WordDelim]), S);
    if Result = 0 then
      if PosNext (cs_Alpha, S) > 0 then
        Result := 1;
  End;



{                                                                              }
{ Pad                                                                          }
{                                                                              }
Function PadLeft (const S : String; const PadCh : Char; const Length : Integer;
         const Cut : Boolean = False) : String;
  Begin
    Result := Dup (PadCh, Length - System.Length (S)) + S;
    if Cut then
      SetLength (Result, Length);
  End;

Function PadRight (const S : String; const PadCh : Char; const Length : Integer;
         const Cut : Boolean = False) : String;
  Begin
    Result := S + Dup (PadCh, Length - System.Length (S));
    if Cut then
      SetLength (Result, Length);
  End;

Function Pad (const S : String; const PadCh : Char; const Length : Integer;
         const Cut : Boolean = False) : String;
var I : Integer;
  Begin
    I := Length - System.Length (S);
    Result := Dup (PadCh, I div 2) + S + Dup (PadCh, (I + 1) div 2);
    if Cut then
      SetLength (Result, Length);
  End;

Function PadInside (const S : String; const PadCh : Char; const Length : Integer) : String;
var I, J, K, C, M : Integer;
    PNP : CharSetArray;
  Begin
    if System.Length (S) >= Length then
      begin
        Result := S;
        exit;
      end;

    PNP := AsCharSetArray ([[PadCh], cs_Everything - [PadCh]]);
    I := Count (PNP, S);
    if I = 0 then // nowhere to pad inside
      begin
        Result := S;
        exit;
      end;

    C := (Length - System.Length (S)) div I; // min characters to insert in each "space"
    M := (Length - System.Length (S)) mod I; // extra characters to insert anywhere
    I := PosNext (PNP, S);
    K := 0;
    Result := CopyLeft (S, I);
    Repeat
      Result := Result + Dup (PadCh, C);
      if K < M then
        Result := Result + PadCh;
      Inc (K);
      J := I;
      I := PosNext (PNP, S, J);
      if I > 0 then
        Result := Result + CopyRange (S, J + 1, I);
    Until I = 0;
    Result := Result + CopyFrom (S, J + 1);
  End;



{                                                                              }
{ Type testing                                                                 }
{                                                                              }
{ ('0'..'9')+ }
Function IsNumber (const S : String) : Boolean;
var L : Integer;
  Begin
    L := Length (S);
    Result := (L > 0) and Match (cs_Numeric, S, 1, L);
  End;

{ ('0'..'9','A'..'F')+ }
Function IsHexNumber (const S : String) : Boolean;
var L : Integer;
  Begin
    L := Length (S);
    Result := (L > 0) and Match (cs_HexDigit, S, 1, L);
  End;

{ ['+','-']<number> }
Function IsInteger (const S : String) : Boolean;
  Begin
    if S = '' then
      Result := False else
      if S [1] in cs_Sign then
        Result := (Length (S) > 1) and IsNumber (CopyFrom (S, 2)) else
        Result := IsNumber (S);
  End;

{ <integer>['.'<number>]   or }
{          ['.'<number>]      }
Function IsReal (const S : String) : Boolean;
var F : Integer;
  Begin
    F := Pos (c_DecimalPoint, S);
    if F = 0 then
      Result := IsInteger (S) else
      Result := ((F = 1) or IsInteger (CopyLeft (S, F - 1))) and
                IsNumber (CopyFrom (S, F + 1));
  End;

{ <real>['e'<int>] }
Function IsScientificReal (const S : String) : Boolean;
var F : Integer;
  Begin
    F := PosNext (cs_Exponent, S);
    if F = 0 then
      Result := IsReal (S) else
      Result := IsReal (CopyLeft (S, F - 1)) and IsInteger (CopyFrom (S, F + 1));
  End;

{ <quote><text incl double quoted quotes><quote> }
Function IsQuotedString (const S : String) : Boolean;
var Quote : Char;
    I     : Integer;
  Begin
    if (Length (S) < 2) or (S [1] <> S [Length (S)]) then
      begin
        Result := False;
        exit;
      end;
    Quote := S [1];
    I := 1;
    Repeat
      I := PosNext (Quote, S, I);
      if (I > 0) and (I < Length (S)) and (S [I + 1] <> Quote) then // not double quoted
        begin
          Result := False;
          exit;
        end;
    Until I = Length (S);
    Result := True;
  End;



{                                                                              }
{ Natural language                                                             }
{   US style billion = 1,000 million                                           }
{   UK style billion = 1,000,000 million                                       }
{   US style trillion = 1,000 US billion                                       }
{   UK style trillion = 1,000,000 UK billion?                                  }
{                                                                              }
Function Number (const Num : Int64; const USStyle : Boolean = False) : String;
var I : Int64;
const
  Eng_minus    = 'minus ';
  Eng_Numbers  : Array [0..12] of String =
        ('zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine',
         'ten', 'eleven', 'twelve');
  Eng_Prefixes : Array [2..9] of String =
        ('twen', 'thir', 'four', 'fif', 'six', 'seven', 'eigh', 'nin');
  Eng_teen     = 'teen';
  Eng_ty       = 'ty';
  Eng_hundred  = ' hundred';
  Eng_and      = ' and ';
  Eng_thousand  = ' thousand';
  Eng_million   = ' million';
  Eng_billion   = ' billion';
  Eng_trillion  = ' trillion';
  USBillion  : Int64 = 1000000000;
  USTrillion : Int64 = 1000000000000;
  UKBillion  : Int64 = 1000000000000;
  UKTrillion : Int64 = 1000000000000000000;

  Begin
    if Num < 0 then
      Result := Eng_minus + Number (-Num) else
    if Num <= 12 then
      Result := Eng_Numbers [Num] else
    if Num <= 19 then
      Result := Eng_Prefixes [Num mod 10] + Eng_teen else
    if Num <= 99 then
      begin
        Result := Eng_Prefixes [Num div 10] + Eng_ty;
        if Num mod 10 > 0 then
          Result := Result + ' ' + Eng_Numbers [Num mod 10];
      end else
    if Num <= 999 then
      begin
        Result := Number (Num div 100) + Eng_hundred;
        if Num mod 100 > 0 then
          Result := Result + Eng_and + Number (Num mod 100);
      end else
    if Num <= 999999 then
      begin
        Result := Number (Num div 1000) + Eng_thousand;
        if Num mod 1000 > 0 then
          Result := Result + ' ' + Number (Num mod 1000);
      end else
    if ((Num < USBillion) and USStyle) or
       ((Num < UKBillion) and not USStyle) then
      begin
        Result := Number (Num div 1000000) + Eng_million;
        if Num mod 1000000 > 0 then
          Result := Result + ' ' + Number (Num mod 1000000);
      end else
    if ((Num < USTrillion) and USStyle) or
       ((Num < UKTrillion) and not USStyle) then
      begin
        if USStyle then
          I := USBillion else
          I := UKBillion;
        Result := Number (Num div I) + Eng_billion;
        if Num mod I > 0 then
          Result := Result + ' ' + Number (Num mod I, USStyle);
      end else
      begin
        if USStyle then
          I := USTrillion else
          I := UKTrillion;
        Result := Number (Num div I) + Eng_trillion;
        if Num mod I > 0 then
          Result := Result + ' ' + Number (Num mod I, USStyle);
      end;
  End;

Function Number (const Num : Extended; const USStyle : Boolean = False) : String;
const Eng_point = ' point';
var N, I : Int64;
  Begin
    Result := Number (Trunc (Num), USStyle);
    N := Abs (Round (Frac (Num) * 1000000));
    if N > 0 then
      begin
        Result := Result + Eng_point;
        I := 100000;
        While (I > 1) and (N > 0) do
          begin
            Result := Result + ' ' + Number (N div I);
            N := N mod I;
            I := I div 10;
          end;
      end;
  End;



{ Pack                                                                         }
Function Pack (const D : Int64) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function Pack (const D : Integer) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function Pack (const D : Byte) : String;
  Begin
    Result := Char (D);
  End;

Function Pack (const D : Word) : String;
  Begin
    Result := Char (Lo (D)) + Char (Hi (D));
  End;

Function Pack (const D : String) : String;
  Begin
    Result := Pack (Length (D)) + D;
  End;

Function PackShortString (const D : ShortString) : String;
  Begin
    Result := D [0] + D;
  End;

Function PackSingle (const D : Single) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function PackDouble (const D : Double) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function Pack (const D : Extended) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function PackCurrency (const D : Currency) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function PackDateTime (const D : TDateTime) : String;
  Begin
    SetLength (Result, Sizeof (D));
    Move (D, Result [1], Sizeof (D));
  End;

Function Pack (const D : Boolean) : String;
  Begin
    Result := Char (Ord (D));
  End;



end.

