unit janShapeController;



interface

uses
  Windows,
  TypInfo,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  menus,
  DsgnIntf,
  janshape,
  janShapeEditor,
  printers;

type
  TonShapeSelected=procedure(sender:TObject;AShape:TjanShape) of object;
  TonMouseModeChanged= procedure (sender:TObject; AMouseMode:TShapeMouseMode) of object;
  TjanShapeController = class(TPopupMenu)
  private
    { Private declarations }
    FObjList:TList;
    FSelObjList:TList;
    FShape:TjanShape;
    FImageList:TImageList;
    FMenuCaptions:TStringlist;
    startimgshape:integer;
    startimggradient:integer;
    startimgUser:integer;
    FGradientColor: Tcolor;
    FBrushColor: Tcolor;
    FBackColor: TColor;
    FMouseMode: TShapeMouseMode;
    FMouseModeController: boolean;
    FMouseGrid: boolean;
    FMouseGridSize: integer;
    FonShapeDeleted: TNotifyEvent;
    FonMouseModeChanged: TonMouseModeChanged;
    FonShapeSelected: TonShapeSelected;
    FUserShapes: TStringlist;
    procedure CreateShapeImages;
    procedure initGradientStyleMenu;
    procedure initShapeStyleMenu;
    procedure initUserShapeStyleMenu;
    procedure initBrushStyleMenu;
    procedure initPenStyleMenu;
    procedure initBrushColorMenu;
    procedure initGradientmenu;
    procedure initGradientColorMenu;
    procedure initCaptionMenu;
    procedure initMouseMenu;
    procedure initOtherMenu;
    procedure SetBrushColor(const Value: Tcolor);
    procedure SetGradientColor(const Value: Tcolor);
    procedure SetBackColor(const Value: TColor);
    procedure ShapeStyleClick(sender:TObject);
    procedure GradientStyleClick(sender:TObject);
    procedure BrushStyleClick(sender:TObject);
    procedure PenStyleClick(sender:TObject);
    procedure BrushColorClick(sender:Tobject);
    procedure PenColorClick(sender:Tobject);
    procedure GradientColorClick(sender:Tobject);
    procedure GradientClick(sender:Tobject);
    procedure FontClick(sender:Tobject);
    procedure CaptionClick(sender:Tobject);
    procedure AngleClick(sender:Tobject);
    procedure PenSizeClick(sender:Tobject);
    procedure CenteredClick(sender:Tobject);
    procedure MouseModeClick(sender:TObject);
    procedure PolyLineModeClick(sender:TObject);
    procedure GroupNameClick(sender:TObject);
    procedure ActiveMouseClick(sender:Tobject);
    procedure popping(sender:TObject);
    function ScaleX(I: Integer): Integer;
    function ScaleY(I: Integer): Integer;
    procedure SetMouseMode(const Value: TShapeMouseMode);
    procedure SetMouseModeController(const Value: boolean);
    procedure SetMouseGrid(const Value: boolean);
    procedure SetMouseGridSize(const Value: integer);
    procedure BringToFrontClick(sender: TObject);
    procedure DeleteObjectClick(sender: TObject);
    procedure SendToBackClick(sender: TObject);
    procedure SetonShapeDeleted(const Value: TNotifyEvent);
    procedure SetonMouseModeChanged(const Value: TonMouseModeChanged);
    procedure SetonShapeSelected(const Value: TonShapeSelected);
    function FindMenuCaption(ACaption: string): TmenuItem;
    procedure SetUserShapes(const Value: TStringlist);
    procedure UserShapeStyleClick(sender: TObject);
    procedure ClearMenu;
    procedure initAlignMenu;
    procedure alBottomClick(sender: TObject);
    procedure alLeftClick(sender: TObject);
    procedure alRightClick(sender: TObject);
    procedure alTopClick(sender: TObject);
    procedure eqHeightClick(sender: TObject);
    procedure eqWidthClick(sender: TObject);
    procedure ConnectorNClick(sender: TObject);
    procedure ConnectorEClick(sender: TObject);
    procedure ConnectorSClick(sender: TObject);
    procedure ConnectorWClick(sender: TObject);
  protected
    { Protected declarations }
  public
    { Public declarations }
    {Assign a TjanShapeController as the popup component of TjanShape components and you can provide your users with full control over the shapes.}
    {}
    {Specially when you are using the shapes in a vector drawing program this is a real time saver.}
    constructor Create(AOwner:Tcomponent);override;
    {Creates an instance of TjanShapeController.}
    destructor Destroy;override;
    {Destroys an instance of TjanShapeController.}
    procedure CreateMenu;
    {Creates all the items of the popupmenu and all the images that are displayed with the menu items.}
    procedure ModifyshapeStyle(AShape:TjanShape;AShapeStyle:string);
    {Modifies the shape style using a string value.}
    {}
    {The value must be one of the valid shape styles.}
    procedure ModifyGradientStyle(AShape: TjanShape; AGradientStyle: string);
    {Modifies the gradient style using a string value.}
    procedure PrintShapes(var Alist:Tlist;scale:extended);
    {Prints all the TjanShape objects in Tlist to scale.}
    procedure SelectingShape(Ashape:TjanShape);
    {Select a TjanShape object. A selected shape is displayed with a square red mark in the center.}
    procedure SmoothResize(var Src, Dst: TBitmap);
    {Creates a smooth resized copy of Src bitmap in Dst bitmap.}
    {}
    {This is just a utility function you can use at will.}
    procedure MovingShape(AShape:TjanShape;dx,dy:integer);
    {You will normally not use this procedure directly.}
    {  }
    {TjanShape object that have a TjanShapeController as their popupcomponent will call this procedure when the user moves them. The ShapeController will move then all the shapes with the same groupname.}
    {}
    {The shapecontroller will also take care of positioning connector shapes if any.}
    procedure AddObj(AShape:TjanShape);
    {Add a TjanShape object to the object list.}
    procedure InsertObj(AShape: TjanShape);
    {Inserts a TjanShape object at the top of the object list.}
    procedure DeleteObj(AShape:TJanShape);
    {Deletes a TjanShape object from the list and frees the object.}
    procedure DeSelectAll;
    {Deselects all shapes in the Shapecontroller's shapes collection.}
    procedure ClearAll;
    {Clears all shapes from the shapecontroller's shapes collection and frees the shapes.}
    procedure SaveAll(Afilename:string);
    {Saves all TjanShape objects in the object list to a file with a .obd extension.}
    procedure SelectRect(R:Trect);
    {Selects all objects from the object list that are withitn the rectangle.}
    procedure ZOrderUp;
    {Changes the Zorder from the objects in the shapes collection from the last one up.}
    procedure ZorderDown;
    {Changes the Zorder from the objects in the shapes collection from the first one down.}
  published
    { Published declarations }
    property UserShapes:TStringlist read FUserShapes write SetUserShapes;
    {Hold all user defined shapes. A special property editor allows you to design new user shapes at design time.}
    {}
    {At run time when a TjanShape instance has a TjanShapeController as popupmenu the user can select any of these user shapes.}
    property BrushColor:Tcolor read FBrushColor write SetBrushColor;
    {Determines the BrushColor of the popup menu images.}
    property GradientColor:Tcolor read FGradientColor write SetGradientColor;
    {Determines the GradientColor of the popup menu images.}
    property BackColor:TColor read FBackColor write SetBackColor;
    {Determines the BackColor of the popup menu images.}
    property MouseMode:TShapeMouseMode read FMouseMode write SetMouseMode;
    {Determines the default mouse mode.}
    property MouseModeController:boolean read FMouseModeController write SetMouseModeController;
    {Determines if the shapecontroller will be used as a mousemode controller.}
    property MouseGrid:boolean read FMouseGrid write SetMouseGrid;
    {Determines if a grid will be used when sizing and moving shapes.}
    property MouseGridSize:integer read FMouseGridSize write SetMouseGridSize;
    {Determines the size of the grid in pixels.}
    property onShapeDeleted:TNotifyEvent read FonShapeDeleted write SetonShapeDeleted; //event
    {Occurs when the user deletes an object from the object list via the popup menu.}
    property onMouseModeChanged:TonMouseModeChanged read FonMouseModeChanged write SetonMouseModeChanged;//event
    {Occurs when the user changes the mousemode with the popup menu.}
    property onShapeSelected:TonShapeSelected read FonShapeSelected write SetonShapeSelected; //event
    {Occurs when the user selects an object from the list.}
  end;

  TUserShapeProperty=class (TClassproperty)
  public
    function Getattributes: TPropertyAttributes; override;
    procedure Edit;override;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('jans', [TjanShapeController]);
  RegisterPropertyEditor(TypeInfo(TStringList),TjanShapeController,'UserShapes',TUserShapeProperty);
end;

procedure TjanShapeController.ClearMenu;
var i,ic:integer;
    m:TMenuItem;

  procedure ClearSubmenu(mm:Tmenuitem);
  var ii,iic:integer;
      subm:TmenuItem;
  begin
     iic:=mm.Count ;
     if iic=0 then exit;
     for ii:=iic-1 downto 0 do
     begin
      subm:=mm.items[ii];
      ClearSubMenu(subm);
      mm.Delete (ii);
      subm.free;
     end;
  end;

begin
  ic:=items.Count ;
  if ic=0 then exit;
  for i:=ic-1 downto 0 do
  begin
    m:=items[i];
    ClearSubMenu(m);
    items.Delete (i);
    m.free;
  end;
end;

procedure TjanShapeController.CreateMenu;
begin
  FmenuCaptions.clear;
  FImageList.Clear ;
  ClearMenu;
  FMouseMode:=mmfixed;
  FMouseModeController:=false;
  CreateShapeImages;
  initShapeStyleMenu;
  initUserShapeStyleMenu;
  initGradientStyleMenu;
  initGradientmenu;
  initGradientColorMenu;
  initBrushColorMenu;
  initBrushStyleMenu;
  initPenStyleMenu;
  initCaptionMenu;
  initMouseMenu;
  initAlignmenu;
  initOthermenu;
end;

procedure TjanShapeController.CreateShapeImages;
var
  PropInfo: PPropInfo;
  ptd: PtypeData;
  i,j,imgcount,uc:integer;
  PropValue:integer;
  g:TjanShape;
  bm:tbitmap;
  R:Trect;
  cap:string;
begin
  if not assigned(images) then exit;
  g:=Fshape;
  g.Gradient :=false;
  R:=rect(0,0,16,16);
  PropInfo:=Getpropinfo(TjanShape.classinfo,'shape');
  ptd:=GetTypeData(PropInfo.proptype^);
  bm:=tbitmap.create;
  bm.height:=16;
  bm.width:=16;
  Images.Height :=16;
  Images.width:=16;
  images.AllocBy :=1;
  j:=0;
  startimgshape:=0;
// shapes
  for i:=ptd.MinValue to ptd.MaxValue do
  begin
    setordprop(g,propinfo,i);
//    g.Refresh;
    bm.canvas.brush.color:=FBackColor;
    bm.canvas.fillrect(R);
    FShape.PaintToCanvas(bm.canvas,0,0,1,1);
//    bm.Canvas.CopyRect (R,g.canvas,R);
    images.add(bm,nil);
    inc(j);
  end;
// gradients
  g.shape:=jstsquare;
  PropInfo:=Getpropinfo(TjanShape.classinfo,'FillDirection');
  ptd:=GetTypeData(PropInfo.proptype^);
  startimggradient:=j;
  g.Gradient :=true;
  for i:=ptd.MinValue to ptd.MaxValue do
  begin
    setordprop(g,propinfo,i);
//    g.Refresh;
//    bm.Canvas.CopyRect (R,g.canvas,R);
    bm.canvas.brush.color:=FBackcolor;
    bm.canvas.fillrect(R);
    Fshape.PainttoCanvas(bm.canvas,0,0,1,1);
    Images.add(bm,nil);
    inc(j);
  end;
// usershapes
  if assigned(FUserShapes) then
    if FUsershapes.count>0 then
    begin
      uc:=FUserShapes.count;
      g.shape:=jstuser;
      startimguser:=j;
      g.Gradient :=false;
      for i:=0 to uc-1 do
      begin
        g.UserVectorString :=FuserShapes[i];
        bm.canvas.brush.color:=FBackcolor;
        bm.canvas.fillrect(R);
        Fshape.PainttoCanvas(bm.canvas,0,0,1,1);
        Images.add(bm,nil);
        inc(j);
        end;
    end;
  bm.free;
end;

procedure TjanShapeController.initShapeStyleMenu;
var
 m,item:Tmenuitem;
 cap:string;
   PropInfo: PPropInfo;
   ptd: PtypeData;
   I,j:integer;
   PropValue:integer;
begin
 m:=Tmenuitem.Create (self);
 items.Add(m);
 m.caption:='Shape Style';
 PropInfo:=Getpropinfo(TjanShape.classinfo,'shape');
 ptd:=GetTypeData(PropInfo.proptype^);
 j:=0;
 for i:=ptd.MinValue to ptd.MaxValue do
 begin
   item:=tmenuitem.Create(self);
   item.caption:=getenumName(propinfo.proptype^,i);
   item.ImageIndex :=j;
   item.onclick:=shapestyleclick;
   FMenuCaptions.AddObject (item.caption,item);
   inc(j);
   m.add(item);
 end;
end;

procedure TjanShapeController.initUserShapeStyleMenu;
var
 m,item:Tmenuitem;
 cap:string;
 PropInfo: PPropInfo;
 ptd: PtypeData;
 I,j:integer;
 PropValue:integer;
 sc:integer;
begin
 if not assigned(FUserShapes) then exit;
 sc:=FUserShapes.count;
 if sc=0 then exit;
 m:=Tmenuitem.Create (self);
 items.Add(m);
 m.caption:='UserShape Style';
 j:=startimguser;
 for i:=0 to sc-1 do
 begin
   item:=tmenuitem.Create(self);
   item.caption:=FUserShapes.Names [i];
   item.ImageIndex :=j;
   item.onclick:=usershapestyleclick;
   FMenuCaptions.AddObject (item.caption,item);
   inc(j);
   m.add(item);
 end;
end;


procedure TjanShapeController.initGradientStyleMenu;
var
 m,item:Tmenuitem;
   PropInfo: PPropInfo;
   ptd: PtypeData;
   I,j,k:integer;
   PropValue:integer;
begin
 m:=Tmenuitem.Create (self);
 items.Add(m);
 m.caption:='GradientStyle';
 PropInfo:=Getpropinfo(TjanShape.classinfo,'FillDirection');
 ptd:=GetTypeData(PropInfo.proptype^);
 j:=0;
 k:=startimggradient;
 for i:=ptd.MinValue to ptd.MaxValue  do begin
 item:=tmenuitem.Create(m.owner);
 item.caption:=getenumName(propinfo.proptype^,i);
 item.ImageIndex :=k;
 FMenuCaptions.AddObject (item.caption,item);
 inc(k);
 item.onclick:=gradientstyleclick;
 inc(j);
 m.add(item);
 end;
end;

procedure TjanShapeController.ModifyshapeStyle(AShape:TjanShape;AShapeStyle:string);
begin
  setEnumprop(Ashape,'Shape',AshapeStyle);
end;

procedure TjanShapeController.ModifyGradientStyle(AShape:TjanShape;AGradientStyle:string);
begin
  setEnumprop(Ashape,'FillDirection',AGradientStyle);
  Ashape.gradient:=true;
end;


procedure TjanShapeController.SetBrushColor(const Value: Tcolor);
begin
  FBrushColor := Value;
  FShape.brush.Color :=value;
end;

procedure TjanShapeController.SetGradientColor(const Value: Tcolor);
begin
  FGradientColor := Value;
  FShape.GradientColor :=value;
end;

constructor TjanShapeController.Create(AOwner: Tcomponent);
begin
  inherited create(AOwner);
  FBrushColor:=clblue;
  FGradientColor:=clwhite;
  FBackColor:=clwhite;
  FMouseGrid:=false;
  FMouseGridSize:=10;
  FShape:=TjanShape.Create (self);
  Fshape.width:=16;
  FShape.height:=16;
  Fshape.GradientColor :=clwhite;
  Fshape.Brush.color:=clblue;
  FImageList:=TImageList.Create (self);
  FImageList.Width :=16;
  FImageList.Height :=16;
  FImageList.AllocBy:=1;
  FmenuCaptions:=TStringlist.create;
  FUserShapes:=TStringlist.create;
  FObjList:=Tlist.create;
  FSelObjList:=TList.create;
  Images:=FImageList;
  onpopup:=popping;

end;

procedure TjanShapeController.SetBackColor(const Value: TColor);
begin
  FBackColor := Value;
end;

destructor TjanShapeController.Destroy;
begin
  FShape.free;
  FImageList.free;
  FMenuCaptions.free;
  FuserShapes.free;
  FObjList.free;
  FSelObjList.free;
  inherited Destroy;
end;


procedure TjanShapeController.GradientStyleClick(sender: TObject);
var Ashape:Tjanshape;
    AGradientStyle:string;
begin
  Ashape:=Tjanshape(popupcomponent);
  AGradientStyle:=Tmenuitem(sender).caption;
  ModifyGradientStyle(Ashape,AGradientStyle);
end;

procedure TjanShapeController.ShapeStyleClick(sender: TObject);
var Ashape:Tjanshape;
    AshapeStyle:string;
begin
  Ashape:=Tjanshape(popupcomponent);
  AshapeStyle:=Tmenuitem(sender).caption;
  ModifyShapeStyle(Ashape,AshapeStyle);
end;

procedure TjanShapeController.UserShapeStyleClick(sender: TObject);
var Ashape:Tjanshape;
    AshapeStyle:string;
    index:integer;
begin
  Ashape:=Tjanshape(popupcomponent);
  AshapeStyle:=Tmenuitem(sender).caption;
  index:=FUserShapes.IndexOfName (AShapeStyle);
  if index=-1 then exit;
  Ashape.UserVectorString :=FUserShapes[index];
  Ashape.Shape:=jstUser;
end;


procedure TjanShapeController.BrushStyleClick(sender: TObject);
var Ashape:Tjanshape;
    AStyle:string;
begin
  Ashape:=Tjanshape(popupcomponent);
  AStyle:=Tmenuitem(sender).caption;
  setEnumprop(Ashape.brush,'Style',AStyle);
end;

procedure TjanShapeController.PenStyleClick(sender: TObject);
var Ashape:Tjanshape;
    AStyle:string;
begin
  Ashape:=Tjanshape(popupcomponent);
  AStyle:=Tmenuitem(sender).caption;
  setEnumprop(Ashape.Pen,'Style',AStyle);
end;

procedure TjanShapeController.initBrushStyleMenu;
var
 m,item:Tmenuitem;
 cap:string;
   PropInfo: PPropInfo;
   ptd: PtypeData;
   I,j:integer;
   PropValue:integer;
begin
 m:=Tmenuitem.Create (self);
 items.Add(m);
 m.caption:='Brush Style';
 PropInfo:=Getpropinfo(TBrush.classinfo,'Style');
 ptd:=GetTypeData(PropInfo.proptype^);
 j:=0;
 for i:=ptd.MinValue to ptd.MaxValue do
 begin
   item:=tmenuitem.Create(self);
   item.caption:=getenumName(propinfo.proptype^,i);
   item.onclick:=brushstyleclick;
   item.RadioItem :=true;
   item.groupindex:=4;
   FMenuCaptions.AddObject (item.caption,item);
   inc(j);
   m.add(item);
 end;
end;

procedure TjanShapeController.initPenStyleMenu;
var
 m,item,subm:Tmenuitem;
 cap:string;
   PropInfo: PPropInfo;
   ptd: PtypeData;
   I,j:integer;
   PropValue:integer;
begin
 m:=Tmenuitem.Create (self);
 m.caption:='Pen';
 items.Add(m);
 subm:=Tmenuitem.Create (self);
 subm.caption:='Style';
 m.Add (subm);
 PropInfo:=Getpropinfo(TPen.classinfo,'Style');
 ptd:=GetTypeData(PropInfo.proptype^);
 j:=0;
 for i:=ptd.MinValue to ptd.MaxValue do
 begin
   item:=tmenuitem.Create(self);
   item.caption:=getenumName(propinfo.proptype^,i);
   item.onclick:=penstyleclick;
   item.RadioItem :=true;
   item.groupindex:=5;
   FMenuCaptions.AddObject (item.caption,item);
   inc(j);
   subm.add(item);
 end;
 subm:=Tmenuitem.Create (self);
 subm.caption:='Color...';
 subm.onclick:=pencolorclick;
 m.Add(subm);

 subm:=Tmenuitem.Create (self);
 subm.caption:='Size...';
 subm.onclick:=pensizeclick;
 m.Add(subm);


end;


procedure TjanShapeController.BrushColorClick(sender: Tobject);
var Ashape:Tjanshape;
    dlg:TColorDialog;
begin
  Ashape:=Tjanshape(popupcomponent);
  dlg:=Tcolordialog.create(self);
  dlg.color:=Ashape.brush.color;
  if dlg.Execute then
    Ashape.Brush.color:=dlg.color;
  dlg.free;
end;

procedure TjanShapeController.FontClick(sender: Tobject);
var Ashape:Tjanshape;
    dlg:TFontDialog;
begin
  Ashape:=Tjanshape(popupcomponent);
  dlg:=TFontdialog.create(self);
  dlg.font.assign(ashape.font);
  if dlg.execute then
    Ashape.Font.Assign (dlg.font);
  dlg.free;
end;

procedure TjanShapeController.GradientColorClick(sender: Tobject);
var Ashape:Tjanshape;
    dlg:TColorDialog;
begin
  Ashape:=Tjanshape(popupcomponent);
  dlg:=Tcolordialog.create(self);
  dlg.Color :=ashape.GradientColor ;
  if dlg.Execute then
    Ashape.GradientColor:=dlg.color;
  dlg.free;
end;

procedure TjanShapeController.PenColorClick(sender: Tobject);
var Ashape:Tjanshape;
    dlg:TColorDialog;
begin
  Ashape:=Tjanshape(popupcomponent);
  dlg:=Tcolordialog.create(self);
  dlg.color:=ashape.pen.color;

  if dlg.Execute then
    Ashape.Pen.color:=dlg.color;
  dlg.free;
end;

procedure TjanShapeController.initBrushColorMenu;
var
  m:Tmenuitem;
begin
  m:=Tmenuitem.Create (self);
  m.caption:='Brush Color...';
  m.onclick:=brushcolorclick;
  items.Add(m);
end;

procedure TjanShapeController.initGradientColorMenu;
var
  m:Tmenuitem;
begin
  m:=Tmenuitem.Create (self);
  m.caption:='Gradient Color...';
  m.onclick:=gradientcolorclick;
  items.Add(m);
end;



procedure TjanShapeController.initGradientmenu;
var
  m:Tmenuitem;
begin
  m:=Tmenuitem.Create (self);
  m.caption:='Gradient';
  m.onclick:=gradientclick;
  FMenuCaptions.AddObject (m.caption,m);
  items.Add(m);
end;

procedure TjanShapeController.GradientClick(sender: Tobject);
var Ashape:Tjanshape;
begin
  Ashape:=Tjanshape(popupcomponent);
  TMenuItem(sender).checked:= not TMenuItem(sender).checked;
  ashape.Gradient :=TMenuItem(sender).checked;
end;

procedure TjanShapeController.initCaptionMenu;
var
  m,subm:Tmenuitem;
begin
  m:=Tmenuitem.Create (self);
  m.caption:='Caption';
  items.Add(m);
  subm:=Tmenuitem.Create (self);
  subm.Caption :='Caption...';
  subm.onclick:=captionclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Font...';
  subm.onclick:=fontclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Centered';
  subm.onclick:=centeredclick;
  FMenuCaptions.AddObject (subm.caption,subm);
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Angle';
  subm.onclick:=angleclick;
  m.Add (subm);

end;


procedure TjanShapeController.initAlignMenu;
var
  m,subm:Tmenuitem;
begin
  m:=Tmenuitem.Create (self);
  m.caption:='Align/Size';
  items.Add(m);
  subm:=Tmenuitem.Create (self);
  subm.Caption :='Top';
  subm.onclick:=alTopclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Left';
  subm.onclick:=alLeftclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Right';
  subm.onclick:=alRightclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Bottom';
  subm.onclick:=alBottomclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='-';
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Equal Height';
  subm.onclick:=eqHeightclick;
  m.Add (subm);
  subm:=Tmenuitem.Create (self);
  subm.caption:='Equal Width';
  subm.onclick:=eqWidthclick;
  m.Add (subm);

end;



procedure TjanShapeController.CaptionClick(sender: Tobject);
var Ashape:Tjanshape;
    Acap:string;
begin
  Ashape:=Tjanshape(popupcomponent);
  Acap:=ashape.Caption ;
  if inputquery('Shape Controller','Enter Caption',acap) then
    ashape.caption:=acap;
end;

procedure TjanShapeController.CenteredClick(sender: Tobject);
var Ashape:Tjanshape;
begin
  Ashape:=Tjanshape(popupcomponent);
  ashape.CaptionCentered  :=not ashape.CaptionCentered;
end;

procedure TjanShapeController.AngleClick(sender: Tobject);
var Ashape:Tjanshape;
    Acap:string;
    A:integer;
begin
  Ashape:=Tjanshape(popupcomponent);
  Acap:=inttostr(ashape.CaptionAngle);
  if inputquery('Shape Controller','Enter Caption Angle (-180..0..180)',acap) then
  begin
    a:=strtointdef(acap,0);
    if ((a>=-180) and (a<=180)) then
      ashape.CaptionAngle:=a;
  end;
end;

procedure TjanShapeController.PenSizeClick(sender: Tobject);
var Ashape:Tjanshape;
    Acap:string;
    A:integer;
begin
  Ashape:=Tjanshape(popupcomponent);
  Acap:=inttostr(ashape.pen.Width );
  if inputquery('Shape Controller','Enter Pen Size (1..20)',acap) then
  begin
    a:=strtointdef(acap,1);
    if ((a>=1) and (a<=25)) then
      ashape.pen.width:=a;
  end;
end;

procedure TjanShapeController.initMouseMenu;
var
 m,item:Tmenuitem;
 cap:string;
   PropInfo: PPropInfo;
   ptd: PtypeData;
   I,j:integer;
   PropValue:integer;
begin
 m:=Tmenuitem.Create (self);
 m.caption:='Mouse Mode';
 items.Add(m);
 PropInfo:=Getpropinfo(TjanShape.classinfo,'MouseMode');
 ptd:=GetTypeData(PropInfo.proptype^);
 j:=0;
 for i:=ptd.MinValue to ptd.MaxValue do
 begin
   item:=tmenuitem.Create(self);
   item.caption:=getenumName(propinfo.proptype^,i);
   item.onclick:=mousemodeclick;
   item.RadioItem :=true;
   item.GroupIndex :=3;
   FMenuCaptions.AddObject (item.caption,item);
   inc(j);
   m.add(item);
 end;
 m:=Tmenuitem.Create (self);
 m.caption:='ActiveMouse';
 m.onclick:=ActiveMouseClick;
 FMenuCaptions.AddObject (m.caption,m);
 items.Add(m);

end;


procedure TjanShapeController.MouseModeClick(sender: TObject);
var Ashape:Tjanshape;
    AStyle:string;
    AMouseMode:TShapeMouseMode;
begin
  Ashape:=Tjanshape(popupcomponent);
  AStyle:=Tmenuitem(sender).caption;
  setEnumprop(Ashape,'MouseMode',AStyle);
  if assigned(onmousemodechanged) then
    onMouseModeChanged(self,Ashape.mousemode);

end;

procedure TjanShapeController.ActiveMouseClick(sender: Tobject);
var Ashape:Tjanshape;
begin
  Ashape:=Tjanshape(popupcomponent);
  TmenuItem(sender).checked:= not TmenuItem(sender).checked;
  ashape.ActiveMouse :=TmenuItem(sender).checked;
end;

procedure TjanShapeController.popping(sender: TObject);
var g:TjanShape;
    m:TMenuItem;
    objmm:TShapeMouseMode;
    cap:string;

    function f(Aname:String):TmenuItem;
    var index:integer;
    begin
       result:=nil;
       index:=FMenuCaptions.indexof(Aname);
       if index=-1 then exit;
       result:=TMenuItem(FMenuCaptions.objects[index]);
    end;
begin
  g:=TjanShape(popupcomponent);
  m:=F('PolyLineMode');
  if m<>nil then m.checked:= g.PolylineMode;
  m:=F('Gradient');
  if m<>nil then m.Checked := g.Gradient;
  m:=F('ActiveMouse');
  if m<>nil then m.Checked := g.ActiveMouse ;
  objmm:=g.MouseMode ;
  cap:=getenumprop(g,'MouseMode');
  m:=F(cap);
  if m<>nil then m.checked:=true;
  cap:=getenumprop(g.brush,'Style');
  m:=F(cap);
  if m<>nil then m.checked:=true;
  cap:=getenumprop(g.pen,'Style');
  m:=F(cap);
  if m<>nil then m.checked:=true;
end;

procedure TjanShapeController.SmoothResize(var Src, Dst: TBitmap);
var
x,y,xP,yP,
yP2,xP2:     Integer;
Read,Read2:  PByteArray;
t,z,z2,iz2:  Integer;
pc:PBytearray;
w1,w2,w3,w4: Integer;
Col1r,col1g,col1b,Col2r,col2g,col2b:   byte;
begin
  xP2:=((src.Width-1)shl 15)div Dst.Width;
  yP2:=((src.Height-1)shl 15)div Dst.Height;
  yP:=0;
  for y:=0 to Dst.Height-1 do
  begin
    xP:=0;
    Read:=src.ScanLine[yP shr 15];
    if yP shr 16<src.Height-1 then
      Read2:=src.ScanLine [yP shr 15+1]
    else
      Read2:=src.ScanLine [yP shr 15];
    pc:=Dst.scanline[y];
    z2:=yP and $7FFF;
    iz2:=$8000-z2;
    for x:=0 to Dst.Width-1 do
    begin
      t:=xP shr 15;
      Col1r:=Read[t*3];
      Col1g:=Read[t*3+1];
      Col1b:=Read[t*3+2];
      Col2r:=Read2[t*3];
      Col2g:=Read2[t*3+1];
      Col2b:=Read2[t*3+2];
      z:=xP and $7FFF;
      w2:=(z*iz2)shr 15;
      w1:=iz2-w2;
      w4:=(z*z2)shr 15;
      w3:=z2-w4;
      pc[x*3+2]:=
        (Col1b*w1+Read[(t+1)*3+2]*w2+
         Col2b*w3+Read2[(t+1)*3+2]*w4)shr 15;
      pc[x*3+1]:=
        (Col1g*w1+Read[(t+1)*3+1]*w2+
         Col2g*w3+Read2[(t+1)*3+1]*w4)shr 15;
      pc[x*3]:=
        (Col1r*w1+Read2[(t+1)*3]*w2+
         Col2r*w3+Read2[(t+1)*3]*w4)shr 15;
      Inc(xP,xP2);
    end;
    Inc(yP,yP2);
  end;
end;

function TjanShapeController.ScaleX(I:Integer): Integer;
begin
  Result :=round( I * (GetDeviceCaps(Printer.Handle, LOGPIXELSX) / Screen.PixelsPerInch));
end;

function TjanShapeController.ScaleY(I:Integer): Integer;
begin
  Result := round(I * (GetDeviceCaps(Printer.Handle, LOGPIXELSY) / Screen.PixelsPerInch));
end;

procedure TjanShapeController.PrintShapes(var Alist: Tlist;scale:extended);
var i,c,x,y:integer;
    w,h:integer;
    obj:TjanShape;
    R:Trect;
    Ax0,Ay0:integer;
    AScaleX,ascaleY:extended;
begin
  c:=alist.Count ;
  if c=0 then exit;
  AScaleX:=scale*GetDeviceCaps(Printer.Handle, LOGPIXELSX) / Screen.PixelsPerInch;
  AScaleY:=scale*GetDeviceCaps(Printer.Handle, LOGPIXELSY) / Screen.PixelsPerInch;
  printer.BeginDoc;
  for i:=0 to c-1 do
  begin
    obj:=TjanShape(alist[i]);
    Ax0:=round(AscaleX*obj.Left) ;
    Ay0:=round(AscaleY*obj.top);
    obj.PaintToCanvas (printer.canvas,Ax0,ay0,AScaleX,AscaleY);
  end;
  printer.EndDoc ;
end;

procedure TjanShapeController.SetMouseMode(const Value: TShapeMouseMode);
begin
  FMouseMode := Value;
end;

procedure TjanShapeController.SetMouseModeController(const Value: boolean);
begin
  FMouseModeController := Value;
end;

procedure TjanShapeController.SetMouseGrid(const Value: boolean);
begin
  FMouseGrid := Value;
end;

procedure TjanShapeController.SetMouseGridSize(const Value: integer);
begin
  FMouseGridSize := Value;
end;

procedure TjanShapeController.initOtherMenu;
var
  m:Tmenuitem;
begin
  m:=Tmenuitem.Create (self);
  m.caption:='GroupName';
  m.onclick:=GroupNameClick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='ConnectorN';
  m.onclick:=ConnectorNClick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='ConnectorE';
  m.onclick:=ConnectorEClick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='ConnectorS';
  m.onclick:=ConnectorSclick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='ConnectorW';
  m.onclick:=ConnectorWClick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='PolyLineMode';
  m.onclick:=PolyLineModeClick;
  items.Add(m);
  FMenuCaptions.AddObject (m.caption,m);
  m:=Tmenuitem.Create (self);
  m.caption:='-';
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='BringToFront';
  m.onclick:=BringToFrontClick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='SendToBack';
  m.onclick:=SendToBackClick;
  items.Add(m);
  m:=Tmenuitem.Create (self);
  m.caption:='Delete Object';
  m.onclick:=DeleteObjectClick;
  items.Add(m);
end;
procedure TjanShapeController.SendToBackClick(sender:TObject);
var Ashape:Tjanshape;
    index:integer;
begin
  Ashape:=Tjanshape(popupcomponent);
  ashape.SendToBack ;
  index:=FobjList.IndexOf (ashape);
  if index=-1 then exit;
  FobjList.Move (index,0);
end;

procedure TjanShapeController.BringToFrontClick(sender:TObject);
var Ashape:Tjanshape;
    index:integer;
begin
  Ashape:=Tjanshape(popupcomponent);
  ashape.BringToFront ;
  index:=FobjList.IndexOf (ashape);
  if index=-1 then exit;
  FobjList.Move (index,FobjList.count-1);
end;

procedure TjanShapeController.DeleteObjectClick(sender:TObject);
var
  i,index:integer;
begin
  if FSelObjList.count=0 then exit;
  for i:=0 to FSelObjList.count-1 do
  begin
    index:=FObjList.indexof(FSelObjList[i]);
    if index<>-1 then
    begin
      TjanShape(FObjList[index]).free;
      FObjList.Delete(index);
    end;
  end;
  FselObjList.clear;
  if assigned(onshapedeleted) then
    onshapedeleted(self);
end;


procedure TjanShapeController.SetonShapeDeleted(const Value: TNotifyEvent);
begin
  FonShapeDeleted := Value;
end;


procedure TjanShapeController.SetonMouseModeChanged(
  const Value: TonMouseModeChanged);
begin
  FonMouseModeChanged := Value;
end;


procedure TjanShapeController.SetonShapeSelected(
  const Value: TonShapeSelected);
begin
  FonShapeSelected := Value;
end;

procedure TjanShapeController.SelectingShape(Ashape: TjanShape);
var index,i:integer;
    gns:string;
    obj:TjanShape;
begin
  gns:=Ashape.GroupName ;
  index:=FSelObjList.IndexOf (AShape);
  if index<>-1 then
  begin
    if gns='' then
    begin
      AShape.Selected :=false;
      FSelObjList.Delete (index);
    end
    else
    begin
      for i:=FSelObjList.count-1 downto 0 do
      begin
        obj:=TjanShape(FSelObjList[i]);
        if obj.groupname=gns then
        begin
          obj.Selected :=false;
          FSelObjList.delete(i);
        end;
      end;
    end;
  end
  else
  begin
    if gns='' then
    begin
      Ashape.selected:=true;
      FSelObjList.Add(Ashape);
    end
    else begin
      for i:=0 to FObjList.count-1 do
      begin
        obj:=TjanShape(FobjList[i]);
        if obj.groupname=gns then
          if FSelObjList.indexof(obj)=-1 then
          begin
            obj.Selected :=true;
            FSelObjList.add(obj);
          end;
      end;
    end;
  end;


  if assigned(onshapeselected) then
    onshapeselected(self,Ashape);
end;

Function TjanShapeController.FindMenuCaption(ACaption:string):TmenuItem;
var i,c:integer;
    m:TMenuitem;

  function FindSubMenucaption(Amenuitem:TMenuItem;Acaption:string):TMenuItem;
  var ii,cc:integer;
      mm:TmenuItem;
  begin
    cc:=AMenuitem.count;
    if cc=0 then
    begin
      result:=nil;
      exit;
    end
    else
    for ii:=0 to cc-1 do
    begin
      mm:=AMenuItem.items[ii];
      if mm.caption=Acaption then
      begin
        result:=mm;
        exit;
      end
      else if mm.count>0 then
      begin
        result:=FindSubMenucaption(mm,Acaption); // recursion
        if result<>nil then exit;
      end;
    end;
    result:=nil;
  end;

begin
  c:=items.Count;
  if c=0 then
  begin
    result:=nil;
    exit;
  end
  else
  for i:=0 to c-1 do
  begin
    m:=items[i];
    if m.caption=Acaption then
    begin
      result:=m;
      exit;
    end
    else if m.count>0 then
    begin
      result:=FindSubMenucaption(m,Acaption);
      if result<>nil then exit;
    end;
  end;
  result:=nil;
end;

procedure TjanShapeController.SetUserShapes(const Value: TStringlist);
begin
  FUserShapes.assign(Value);
end;

{ TUserShapeProperty }

procedure TUserShapeProperty.Edit;
var
  ShapeForm: TShapeEditorForm;
  usr:TStringlist;
  obj:TjanShapeController;
begin
  try
  ShapeForm:=TShapeEditorForm.Create (application);
  obj:=getComponent(0) as TjanShapeController;
  ShapeForm.UserList.items.assign(obj.usershapes);
  if ShapeForm.showmodal=mrOK then
  begin
     obj.usershapes.Assign (ShapeForm.userlist.items);
  end;
  finally
    shapeForm.free;
  end;
end;

function TUserShapeProperty.Getattributes: TPropertyAttributes;
begin
  result:= inherited GetAttributes + [paDialog]-[pasubproperties];
end;

procedure TjanShapeController.PolyLineModeClick(sender: TObject);
var Ashape:Tjanshape;
begin
  Ashape:=Tjanshape(popupcomponent);
  TmenuItem(sender).checked:= not TmenuItem(sender).checked;
  ashape.PolyLineMode :=TmenuItem(sender).checked;
end;

procedure TjanShapeController.MovingShape(AShape: TjanShape;dx,dy:integer);
var i,index,x,y,w,h,dw,dh,x2,y2,yh,xw:integer;
    obj:TjanShape;

begin
  if FSelObjList.count<>0 then
  for i:=0 to FSelObjList.count-1 do
    if FSelObjList[i]<>Ashape then
    begin
      obj:=Tjanshape(FSelObjList[i]);
      obj.left:=obj.left+dx;
      obj.top:=obj.top+dy;
    end;
  x:=Ashape.left;
  y:=Ashape.top;
  h:=Ashape.height;
  w:=Ashape.Width;
  yh:=y+h-1;
  xw:=x+w-1;
  x2:=x+ (w div 2);
  y2:=y+ (h div 2);
  for i:=0 to FobjList.count-1 do
  begin
    obj:=TjanShape(FObjList[i]);
    if (obj.GroupName<>'') and(obj<>Ashape) then
    begin
      if AShape.ConnectorN=obj.GroupName then
      begin
        if y>obj.top+5 then
        begin
          obj.height:=y-obj.top;
        end;
      end
      else if AShape.ConnectorE=obj.GroupName then
      begin
        if xw<(obj.left+obj.width-5) then
        begin
          obj.width:=obj.left+obj.width-xw-1;
          obj.left:=xw+1;
        end;
      end
      else if AShape.ConnectorS=obj.GroupName then
      begin
        if yh<(obj.top+obj.height-5) then
        begin
          obj.height:=obj.top+obj.height-yh-1;
          obj.top:=yh+1;
        end;
      end
      else if AShape.ConnectorW=obj.GroupName then
      begin
        if  x>(obj.left+5) then
        begin
          obj.width:=x-obj.left;
        end;
      end;
    end;
  end;
end;


procedure TjanShapeController.InsertObj(AShape: TjanShape);
var index:integer;
begin
  index:=FObjList.IndexOf (Ashape);
  if index=-1 then
   if FobjList.count=0 then
     FObjList.Add (Ashape)
   else
     FObjList.Insert (0,Ashape);
end;


procedure TjanShapeController.AddObj(AShape: TjanShape);
var index:integer;
begin
  index:=FObjList.IndexOf (Ashape);
  if index=-1 then
     FObjList.Add (Ashape);
end;

procedure TjanShapeController.DeleteObj(AShape: TJanShape);
var i,index:integer;
    obj:TjanShape;
begin
  if FselObjList.count=0 then exit;
  for i:=0 to FSelObjList.count-1 do
  begin
    index:=FObjList.indexof(FSelObjList[i]);
    if index<>-1 then
    begin
      obj:=TjanShape(FObjList[index]);
      FobjList.Delete (index);
      obj.free;
    end;
  FSelObjList.Clear ;
  end;
end;

procedure TjanShapeController.DeSelectAll;
var i:integer;
begin
  if FSelObjList.count=0 then exit;
  for i:=0 to FSelObjList.count-1 do
    TjanShape(FSelObjList[i]).selected:=false;
  FSelObjList.clear;
end;

procedure TjanShapeController.SelectRect(R: Trect);
var i,oc:integer;
    obj:TjanShape;

  function objInRect(Aobj:TjanShape;RR:Trect):boolean;
  var xc,yc:integer;
  begin
    xc:=AObj.left+(Aobj.width div 2);
    yc:=Aobj.top+(AObj.height div 2);
    result:= (xc>=RR.left) and (yc>=RR.top) and
      (xc<=RR.right) and (yc<+RR.bottom);
  end;

begin
  oc:=FobjList.count;
  if oc=0 then exit;
  DeselectAll;
  for i:=0 to oc-1 do
  begin
    obj:=FobjList[i];
    if objInRect(obj,R) then
    begin
      obj.selected:=true;
      FSelObjList.Add (obj);
    end;
  end;
end;

procedure TjanShapeController.GroupNameClick(sender: TObject);
var
  i,index:integer;
  gns:string;
  obj:TjanShape;
begin
  if FSelObjList.count=0 then
  begin
    obj:=Tjanshape(popupcomponent);
    gns:=obj.GroupName ;
    if inputquery('ShapeController','Enter Groupname',gns) then
      obj.GroupName :=gns;
  end
  else
  begin
  gns:=TjanShape(FselObjList[0]).GroupName;
  if not inputquery('ShapeController','Enter Groupname',gns) then exit;
  for i:=0 to FSelObjList.count-1 do
     TjanShape(FSelObjList[i]).GroupName:=gns;
  end;
end;

procedure TjanShapeController.ConnectorEClick(sender: TObject);
var
  i,index:integer;
  gns:string;
  obj:TjanShape;
begin
    obj:=Tjanshape(popupcomponent);
    gns:=obj.ConnectorE;
    if inputquery('ShapeController','Enter ConnectorE',gns) then
      obj.ConnectorE :=gns;
end;

procedure TjanShapeController.ConnectorNClick(sender: TObject);
var
  i,index:integer;
  gns:string;
  obj:TjanShape;
begin
    obj:=Tjanshape(popupcomponent);
    gns:=obj.ConnectorN;
    if inputquery('ShapeController','Enter ConnectorN',gns) then
      obj.ConnectorN :=gns;
end;


procedure TjanShapeController.ClearAll;
var i:integer;
begin
  if FobjList.count=0 then exit;
  for i:=FobjList.count-1 downto 0 do
    TjanShape(FObjList[i]).free;
  FobjList.Clear ;
  FSelObjList.clear;
end;

procedure TjanShapeController.SaveAll(Afilename:string);
var
  fs:TFilestream;
  i:integer;
begin
  FS := TFileStream.Create(Afilename,fmCreate or fmShareDenyWrite);
  if FObjList.count=0 then exit;
  try
    for i := 0 to FObjList.count-1 do
      FS.WriteComponent(TjanShape(FobjList[i]));
  finally FS.Free;
  end;
end;

procedure TjanShapeController.ZOrderUp;
var i:integer;
begin
  if FobjList.count=0 then exit;
  for i:=FObjList.Count-1 downto 0 do
    TjanShape(FObjList[i]).bringtofront;
end;

procedure TjanShapeController.ZorderDown;
var i:integer;
begin
  if FobjList.count=0 then exit;
  for i:=0 to FObjList.Count-1 do
    TjanShape(FObjList[i]).bringtofront;
end;

procedure TjanShapeController.alLeftClick(sender: TObject);
var
  i,value:integer;
begin
  if FSelObjList.count<2 then exit;
  value:=TjanShape(FSelObjList[0]).left;
  for i:=1 to FSelObjList.count-1 do
     TjanShape(FSelObjList[i]).left:=value;
end;

procedure TjanShapeController.alRightClick(sender: TObject);
var
  i,value:integer;
  obj:TjanShape;
begin
  if FSelObjList.count<2 then exit;
  obj:=TjanShape(FSelObjList[0]);
  value:=obj.left+obj.width-1;
  for i:=1 to FSelObjList.count-1 do
  begin
     obj:=TjanShape(FSelObjList[i]);
     obj.left:=value-obj.width+1;
  end;
end;


procedure TjanShapeController.alTopClick(sender: TObject);
var
  i,value:integer;
begin
  if FSelObjList.count<2 then exit;
  value:=TjanShape(FSelObjList[0]).top;
  for i:=1 to FSelObjList.count-1 do
     TjanShape(FSelObjList[i]).top:=value;
end;

procedure TjanShapeController.alBottomClick(sender: TObject);
var
  i,value:integer;
  obj:TjanShape;
begin
  if FSelObjList.count<2 then exit;
  obj:=TjanShape(FSelObjList[0]);
  value:=obj.top+obj.height-1;
  for i:=1 to FSelObjList.count-1 do
  begin
     obj:=TjanShape(FSelObjList[i]);
     obj.top:=value-obj.height+1;
  end;
end;


procedure TjanShapeController.eqHeightClick(sender: TObject);
var
  i,value:integer;
begin
  if FSelObjList.count<2 then exit;
  value:=TjanShape(FSelObjList[0]).height;
  for i:=1 to FSelObjList.count-1 do
     TjanShape(FSelObjList[i]).height:=value;
end;

procedure TjanShapeController.eqWidthClick(sender: TObject);
var
  i,value:integer;
begin
  if FSelObjList.count<2 then exit;
  value:=TjanShape(FSelObjList[0]).width;
  for i:=1 to FSelObjList.count-1 do
     TjanShape(FSelObjList[i]).width:=value;
end;


procedure TjanShapeController.ConnectorSClick(sender: TObject);
var
  gns:string;
  obj:TjanShape;
begin
    obj:=Tjanshape(popupcomponent);
    gns:=obj.ConnectorS;
    if inputquery('ShapeController','Enter ConnectorS',gns) then
      obj.ConnectorS :=gns;
end;

procedure TjanShapeController.ConnectorWClick(sender: TObject);
var
  gns:string;
  obj:TjanShape;
begin
    obj:=Tjanshape(popupcomponent);
    gns:=obj.ConnectorW;
    if inputquery('ShapeController','Enter ConnectorW',gns) then
      obj.ConnectorW :=gns;
end;

end.

