unit Yearplan;

{  Year Planner component written by Jonathan Hosking, November 1999.

   Get future component updates from the following address
   Website: http://www.the-hoskings.freeserve.co.uk/

   Send any bugs, suggestions, etc to the following Email
   Email: jonathan@the-hoskings.freeserve.co.uk

   Thanks to Simon Nicholson for helping with the control updating routines
   Email: Simon.Nicholson@helmstone.co.uk

   Thanks to Richard Haven for helping with the heading setup routine
   Email: lanframe-news@scruznet.com

   Thanks to Wolfgang Kleinrath for helping with the data setup routine and
   providing the code for the original routines for loading and saving cell
   data to INI files
   Email: wkleinrath@xpoint.at

   Thanks to Nacho Urenda for helping with the size calculation routine
   Email: NachoUrenda@compuserve.com

   Thanks to Rob Schoenaker for improving the drawing routines
   Email: rschoenaker@kraan.com

   Thanks to Robert Gesswein for adding the NoDayPriority and StartDayOfWeek
   properties and for helping with the SetColorAtDate routine
   Email: rgesswein@matmus.com

   Thanks to Paul Fisher for adding printing support, the new routines for
   loading and saving cell data to streams, and for helping out with the new
   cell selection routines
   Email: PFisher@emis-support.demon.co.uk

   Notes: CellData is not saved, even though it is a property.  This is
          because it is changed at runtime

          Borland's routine for testing for leap years has been used here as
          Delphi 1 had no such routine }

interface

{ If you want to use a blob stream to load and save data, uncomment the next
  line }
{.$DEFINE USEBLOB}

uses
  {$IFDEF WIN32} Windows, {$ELSE} WinTypes, WinProcs, {$ENDIF}
  SysUtils, Messages, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Menus, {$IFDEF USEBLOB} DBTables, {$ENDIF} Printers;

type
  TypDOW = (ypMonday,ypTuesday,ypWednesday,ypThursday,ypFriday,ypSaturday,ypSunday);
  TypSel = (ypNotSelecting,ypSelecting,ypSelected);
  TYearEvent = procedure(StDays,EnDays,StMonth,EnMonth:integer) of object;
  {$IFNDEF WIN32}
  tagSIZE = record
    cx, cy : Longint;
  end;
  {$ENDIF}
  { Compiling under Delphi 1 limits us to a 64KB data limit, so the record
    cannot be too long.  Under later versions there are bigger data limits }
  TCellData = record
    CellHint: String{$IFNDEF WIN32}[150]{$ENDIF};
    CellColor: TColor;
    CustomColor: Boolean;
    CellDate:TDateTime;
  end;
  TCurrentDate = record
    Day,Month: Byte;
  end;
  TYearPlanAbout = (abNone,abAbout);

  TYearPlanner = class(TCustomControl)
  private
    { Private declarations }
    fAbout: TYearPlanAbout;
    fControl: TBitmap;
    fDayColor: TColor;
    fUseBitmap: Boolean;
    fGridLineColor: TColor;
    fHeadingColor: TColor;
    fHintColor: TColor;
    fHintDelay: Integer;
    fMonthColor: TColor;
    fNoDayColor: TColor;
    fNoDayPriority: Boolean;
    fOnClick: TNotifyEvent;
    fOnYearRightClick: TYearEvent;
    fSelectionColor: TColor;
    fStringList: TStringList;
    fUseFreeSpace: Boolean;
    fWeekendColor: TColor;
    fWeekendHeadingColor: TColor;
    fYear: Word;
    fStartDayOfWeek: TypDOW;
    procedure CalculateCalendar;
    procedure CalculateData;
    procedure CalculateSizes;
    procedure SetupHeadings;
    procedure SetDayColor(Val: TColor);
    procedure SetGridLineColor(Val: TColor);
    procedure SetHeadingColor(Val: TColor);
    procedure SetHintColor(Val: TColor);
    procedure SetHintDelay(Val: Integer);
    procedure SetMonthColor(Val: TColor);
    procedure SetNoDayColor(Val: TColor);
    procedure SetNoDayPriority(Val: Boolean);
    procedure SetSelectionColor(Val: TColor);
    procedure SetUseFreeSpace(Val: Boolean);
    procedure SetWeekendColor(Val: TColor);
    procedure SetWeekendHeadingColor(Val: TColor);
    procedure SetYear(Val: Word);
    procedure SetStartDayOfWeek(Val: TypDOW);
    procedure ShowAbout(Val: TYearPlanAbout);
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message wm_EraseBkgnd;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message wm_LButtonDown;
    procedure WMLMouseButtonUp(var Message: TWMLButtonUp); message wm_LButtonUp;
    procedure WMRMouseButtonDown(var Message: TWMRButtonDown); message wm_RButtonDown;
    procedure WMMouseMove(var Message: TWMMouseMove); message wm_MouseMove;
  protected
    { Protected declarations }
    procedure Paint; override;
  public
    { Public declarations }
    CellData: Array[1..12,1..31] of TCellData;
    CurrentDate: TCurrentDate;
    EnDay: Integer;
    EnMonth: Integer;
    StDay: Integer;
    StMonth: Integer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SaveToStream(var Stream:{$IFDEF USEBLOB}TBlobStream{$ELSE}TStream{$ENDIF});
    procedure LoadFromStream(var Stream:{$IFDEF USEBLOB}TBlobStream{$ELSE}TStream{$ENDIF});
    procedure SetColorAtDate(dt: TDateTime; cellColor: TColor; UpdateControl: Boolean);
    procedure Print;
  published
    { Published declarations }
    property About: TYearPlanAbout read fAbout write ShowAbout default abNone;
    property Align;
    property Color;
    property DayColor: TColor read fDayColor write SetDayColor default clWhite;
    property DragCursor;
    property DragMode;
    property DrawOffScreen: Boolean read fUseBitmap write fUseBitmap default True;
    property Enabled;
    property Font;
    property GridLineColor: TColor read fGridLineColor write SetGridLineColor default clSilver;
    property HeadingColor: TColor read fHeadingColor write SetHeadingColor default clGray;
    property HintColor: TColor read fHintColor write SetHintColor default clYellow;
    property HintDelay: Integer read fHintDelay write SetHintDelay default 0;
    property MonthColor: TColor read fMonthColor write SetMonthColor default clGray;
    property NoDayColor: TColor read fNoDayColor write SetNoDayColor default clSilver;
    property NoDayPriority: boolean read fNoDayPriority write SetNoDayPriority default False;
    property SelectionColor: TColor read fSelectionColor write SetSelectionColor default clBlue;
    property StartDayOfWeek: TypDOW read fStartDayOfWeek write SetStartDayOfWeek default ypMonday;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property UseFreeSpace: Boolean read fUseFreeSpace write SetUseFreeSpace default True;
    property Visible;
    property WeekendColor: TColor read fWeekendColor write SetWeekendColor default clGray;
    property WeekendHeadingColor: TColor read fWeekendHeadingColor write SetWeekendHeadingColor default clSilver;
    property Year: Word read fYear write SetYear;
    property OnClick: TNotifyEvent read fOnClick write fOnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseRightClick: TYearEvent read fOnYearRightClick write fOnYearRightClick;
  end;

procedure Register;

implementation

{ TYearPlanner }

const
  CopyRightStr: PChar = 'TYearPlanner Component v1.61 (22/11/1999)'+#13+#13+
    'By Jonathan Hosking'+#13+#13+'Compiled in '+
    {$IFDEF VER80}  'Delphi 1.0' {$ENDIF}
    {$IFDEF VER90}  'Delphi 2.0' {$ENDIF}
    {$IFDEF VER100} 'Delphi 3.0' {$ENDIF}
    {$IFDEF VER120} 'Delphi 4.0' {$ENDIF}
    {$IFDEF VER130} 'Delphi 5.0' {$ENDIF}
    {$IFDEF VER93}  'C++Builder 1.0' {$ENDIF}
    {$IFDEF VER110} 'C++Builder 3.0' {$ENDIF}
    {$IFDEF VER125} 'C++Builder 4.0' {$ENDIF};
const
  MonthDays : array[1..12] of Integer = (31,28,31,30,31,30,31,31,30,31,30,31);
  ChrTab = #9;
var
  CopyRightPtr: Pointer;
  Cells: Array[0..37,0..12] of string[9];
  Heights: Array[0..12] of Integer;
  Widths: Array[0..37] of Integer;
  cX,cY,OldX,OldY: Integer;
  FirstTickCount:{$IFDEF WIN32}Cardinal{$ELSE}LongInt{$ENDIF};
  HintWin: THintWindow;
  hUpdating,hWaiting,hWaitingToDestroy,Printing: Boolean;
  HintDate:TDateTime;
  InDay,InMonth: Integer;
  Selecting: TypSel;

{ Procedure to test for a leap year - This is the routine used in Delphi 5,
  but I have used it here as Delphi 1 did not have such a procedure }
function IsLeapYear(Year: Word): Boolean;
begin
  Result := (Year mod 4 = 0) and ((Year mod 100 <> 0) or (Year mod 400 = 0));
end;

{ Converts mouse coordinates to cell coordinates }
procedure XYToCell(X,Y: Integer;var CellX,CellY: Integer);
begin
  { Work out the column }
  if X < Widths[0] then CellX := 0 else
  begin
    CellX := ((X - Widths[0]) div Widths[1]) + 1;
    if CellX > 37 then CellX := 37;
  end;
  { Work out the row }
  if Y < Heights[0] then CellY := 0 else
  begin
    CellY := ((Y - Heights[0]) div Heights[1]) + 1;
    if CellY > 12 then CellY := 12;
  end;
end;

{ Thanks to Paul Fisher for this routine }
procedure TYearPlanner.LoadFromStream(var Stream:{$IFDEF USEBLOB}TBlobStream{$ELSE}TStream{$ENDIF});

procedure ParseStringList;
var
  x,y,z,ct: integer;
  l,h: String;
begin
  ct := fStringList.Count - 1;
  for z := 0 to ct do
  begin
    { Copy one line of stringlist into string l }
    l := fStringList[z];
    x := StrToIntDef(Copy(l,1,2),1);
    y := StrToIntDef(Copy(l,4,2),1);
    Delete(l,1,5);
    h := l;
    CellData[x,y].CellHint := h;
    CellData[x,y].CellColor := clYellow;
    CellData[x,y].CustomColor := True;
  end;
end;

begin
  try
    fStringList.LoadFromStream(Stream);
    ParseStringList;
    Self.Refresh;
  finally
  end;
end;

{ Thanks to Paul Fisher for this routine }
procedure TYearPlanner.SaveToStream(var Stream:{$IFDEF USEBLOB}TBlobStream{$ELSE}TStream{$ENDIF});
var
  x,y: Integer;
  s,i,v,finished: String;
begin
  s := Format('%4.4d',[fYear]);
  try
    for x := 1 to 12 do
      for y := 1 to 31 do
      begin
        i := Format('%2.2d_%2.2d',[x,y]);
        with CellData[x,y] do
        begin
          if CellHint <> '' then
          begin
            v := CellHint;
            finished := I + V;
            fStringList.Add(finished);
          end;
        end;
      end;
  finally
    fStringList.SaveToStream(Stream);
  end;
end;

{ Thanks to Robert Gesswein for improving this procedure }
procedure TYearPlanner.CalculateCalendar;
var
  I,J: Byte;
  DaysInMonth,StartDay: Integer;
begin
  { Set the Year cell }
  Cells[0, 0] := IntToStr(Self.Year);
  { Clear the cell contents }
  for I := 1 to 37 do
    for J := 1 to 12 do
      Cells[I,J] := '';
  { Setup the cells }
  for I := 1 to 12 do
  begin
    StartDay := DayOfWeek(EncodeDate(Year,I,1));
    StartDay := (StartDay+7-Ord(fStartDayOfWeek)-2) mod 7;
    DaysInMonth := MonthDays[I] + byte(IsLeapYear(Year) and (I = 2));
    for J := 1 to DaysInMonth do Cells[J + StartDay,I] := IntToStr(J);
  end;
end;

{ Thanks to Paul Fisher and Wolfgang Kleinrath for improving this procedure }
procedure TYearPlanner.CalculateData;
var
  I,J: Byte;
  DaysInMonth: Integer;
begin
  { Setup the hints }
  for I := 1 to 12 do
  begin
    DaysInMonth := MonthDays[I] + byte(IsLeapYear(Year) and (I = 2));
    for J := 1 to DaysInMonth do
    begin
      with CellData[I,J] do
      begin
        CellColor := $00000000;
        CustomColor := False;
        CellDate := EncodeDate(Year,I,J);
        CellHint := '';
      end;
    end;
  end;
end;

{ Thanks to Nacho Urenda and Paul Fisher for helping with this procedure }
procedure TYearPlanner.CalculateSizes;
var
  I: Byte;
begin
  { Calculate the cell sizes based on whether or not we are printing or
    using the free space }
  if fUseFreeSpace then
  begin
    if Printing then
    begin
      Heights[0] := Printer.PageHeight - ((printer.PageHeight div 13) * 12);
      Widths[0] := Printer.PageWidth - ((printer.PageWidth div 41) * 37);
    end
    else
    begin
      Heights[0] := Height - ((Height div 13) * 12);
      Widths[0] := Width - ((Width div 41) * 37);
    end;
  end
  else
  begin
    if Printing then
    begin
      Heights[0] := (Printer.PageHeight div 13);
      Widths[0] := (Printer.PageWidth div 41) * 4;
    end
    else
    begin
      Heights[0] := (Height div 13);
      Widths[0] := (Width div 41) * 4;
    end
  end;
  if Printing then
  begin
    for I := 1 to 37 do Widths[I] := (Printer.PageWidth div 41);
    for I := 1 to 12 do Heights[I] := (Printer.PageHeight div 13);
  end
  else
  begin
    for I := 1 to 37 do Widths[I] := (Width div 41);
    for I := 1 to 12 do Heights[I] := (Height div 13);
  end;
end;

{ Thanks to Richard Haven and Robert Gesswein for this improved procedure }
procedure TYearPlanner.SetupHeadings;
var
   I: Byte;
begin
  for I := 1 to 37 do Cells[I,0] := ShortDayNames[((I+(Ord(fStartDayOfWeek)) - 1) mod 7) + 2][1];
  for I := 1 to 12 do Cells[0,I] := LongMonthNames[I];
end;

procedure TYearPlanner.SetDayColor(Val: TColor);
begin
  if fDayColor <> Val then
  begin
    fDayColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetGridLineColor(Val: TColor);
begin
  if fGridLineColor <> Val then
  begin
    fGridLineColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetHeadingColor(Val: TColor);
begin
  if fHeadingColor <> Val then
  begin
    fHeadingColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetHintColor(Val: TColor);
begin
  if fHintColor <> Val then
  begin
    fHintColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetHintDelay(Val: Integer);
begin
  if fHintDelay <> Val then
  begin
    fHintDelay := Val;
    if fHintDelay < 0 then fHintDelay := 0;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetMonthColor(Val: TColor);
begin
  if fMonthColor <> Val then
  begin
    fMonthColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetNoDayColor(Val: TColor);
begin
  if fNoDayColor <> Val then
  begin
    fNoDayColor := Val;
    Invalidate;
  end;
end;

{ Thanks to Robert Gesswein contributing this procedure }
procedure TYearPlanner.SetNoDayPriority(Val: Boolean);
begin
  if fNoDayPriority <> Val then
  begin
    fNoDayPriority := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetSelectionColor(Val: TColor);
begin
  if fSelectionColor <> Val then
  begin
    fSelectionColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetUseFreeSpace(Val: Boolean);
begin
  if fUseFreeSpace <> Val then
  begin
    fUseFreeSpace := Val;
    CalculateSizes;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetWeekendColor(Val: TColor);
begin
  if fWeekendColor <> Val then
  begin
    fWeekendColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetWeekendHeadingColor(Val: TColor);
begin
  if fWeekendHeadingColor <> Val then
  begin
    fWeekendHeadingColor := Val;
    Invalidate;
  end;
end;

procedure TYearPlanner.SetYear(Val: Word);
begin
  if fYear <> Val then
  begin
    fYear := Val;
    CalculateCalendar;
    CalculateData;
    Invalidate;
  end;
end;

{ Thanks to Robert Gesswein for contributing this procedure }
procedure TYearPlanner.SetStartDayOfWeek(Val: TypDOW);
begin
  if fStartDayOfWeek <> Val then
  begin
    fStartDayOfWeek := Val;
    SetupHeadings;
    CalculateCalendar;
    CalculateData;
    Invalidate;
  end;
end;

procedure TYearPlanner.ShowAbout(Val: TYearPlanAbout);
begin
  if fAbout <> Val then
  begin
    if Val = abNone then fAbout := Val else
    begin
      fAbout := abNone;
      MessageDlg(StrPas(CopyRightStr), mtInformation, [mbOk], 0);
    end;
    Invalidate;
  end;
end;

procedure TYearPlanner.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

{ Thanks to Paul Fisher and Rob Schoenaker for improving this routine }
procedure TYearPlanner.Paint;
var
  I,J: Byte;
  T,tH,tW,X,Y: Integer;
  GridCellRect: TRect;
  GridCol,OldColor: TColor;
  CurrWidth, CurrHeight : Integer;
  CellText: string;
  CellTextLen: Integer;
  TheCanvas: TCanvas;
  DrawDC : HDC;
  SizeRec: {$IFDEF WIN32} tagSize; {$ELSE} tSize; {$ENDIF}

{ This function determines if a cell is selected }
function CellSelected: Boolean;
begin
  Result := False;
  if (I >= StDay) and (I <= EnDay) and (J >= StMonth) and (J <= EnMonth)
    then Result := True;
end;

{ This procedure works out the color of a cell - Thanks to Robert Gesswein for
  improving it }
function GridColor: TColor;
var
  Dy,Mn: Byte;
begin
  if I = 0 then
  begin
    { Month heading }
    Result := fMonthColor;
    Exit;
  end;
  if (Cells[I,J] <> '') and (J > 0) and (J < 13) then
  begin
    { It's a calender day, so check for a color }
    Dy := StrToInt(Cells[I,J]);
    Mn := J;
    if CellSelected then
    begin
      { It's a selected cell }
      Result := fSelectionColor;
      Exit;
    end;
    if CellData[Mn,Dy].CustomColor then
    begin
      { Use the custom color }
      Result := CellData[Mn,Dy].CellColor;
      Exit;
    end;
  end;
  if (((I+Ord(fStartDayOfWeek) in [0,6,7,13,14,20,21,27,28,34,35,41,42]) or (J = 0))
    and ((not fNoDayPriority) or (Cells[I,J] <> ''))) then
  begin
    { Weekend day or heading }
    Result := fWeekendColor;
    if J = 0 then
      if (I+Ord(fStartDayOfWeek) in [6,7,13,14,20,21,27,28,34,35,41,42]) then
        Result := fWeekendHeadingColor else
          Result := fHeadingColor;
  end
  else
  begin
    { Normal day }
    if Cells[I,J] = '' then Result := fNoDayColor
      else Result := fDayColor;
  end;
end;

{ Thanks to Rob Schoenaker for this routine }
procedure DrawGridLines;
var
  L: Integer;
  LineHeight: Integer;
begin
  with TheCanvas do
  begin
    { Draw the grid }
    Pen.Color := fGridLineColor;
    X := Widths[0] - 1;
    Y := Heights[0] - 1;
    LineHeight := Heights[1] shl 2 + Heights[1] shl 3 + 1;
    for L := 1 to 38 do
    begin
      {$IFDEF WIN32}
      Windows.MoveToEx(DrawDC, X, Y, nil);
      Windows.LineTo(DrawDC, X, Y + LineHeight);
      {$ELSE}
      WinProcs.MoveToEx(DrawDC, X, Y, nil);
      WinProcs.LineTo(DrawDC, X, Y + LineHeight);
      {$ENDIF}
      if L < 38 then
        Inc(X, Widths[L]);
    end;
    for L := 1 to 13 do
    begin
      {$IFDEF WIN32}
      Windows.MoveToEx(DrawDC, Widths[0], Y, nil);
      Windows.LineTo(DrawDC, X, Y);
      {$ELSE}
      WinProcs.MoveToEx(DrawDC, Widths[0], Y, nil);
      WinProcs.LineTo(DrawDC, X, Y);
      {$ENDIF}
      Inc(Y, Heights[L])
    end;
  end;
end;

begin
  { Setup the offscreen bitmap }
  CalculateSizes;
  if fUseBitmap and not (csDesigning in ComponentState) then
  begin
    fControl.Width := Width;
    fControl.Height := Height;
    TheCanvas := fControl.Canvas;
  end
  else
    TheCanvas := Canvas;
  if Printing then TheCanvas := Printer.Canvas;
  with TheCanvas do
  begin
    { Draw the calender cells and text }
    Brush.Style := bsSolid;
    Font := Self.Font;
    DrawDC := TheCanvas.Handle;
    SetBKMode(DrawDC, TRANSPARENT);
    X := 0;
    for I := 0 to 37 do
    begin
      J := 0;
      Y := 0;
      CurrWidth := Widths[I];
      OldColor := GridColor;
      repeat
        T := Y;
        repeat
          Inc(Y,Heights[J]);
          Inc(J);
          GridCol := GridColor;
        until (GridCol <> OldColor) or (J = 13);
        GridCellRect := Rect(X, T, X + CurrWidth, Y);
        Brush.Color := OldColor;
        OldColor := GridCol;
        {$IFDEF WIN32}
        Windows.FillRect(DrawDC, GridCellRect, Brush.Handle);
        {$ELSE}
        WinProcs.FillRect(DrawDC, GridCellRect, Brush.Handle);
        {$ENDIF}
      until
        J = 13;
      Y := 0;
      for J := 0 to 12 do
      begin
        CurrHeight := Heights[J];
        GridCellRect := Rect(X,Y + 1,X + CurrWidth - 1,Y + CurrHeight - 1);
        if (I = 0) or (J = 0) then
        {$IFDEF WIN32}
          DrawEdge(DrawDC, GridCellRect, BDR_RAISEDINNER, BF_RECT);
        {$ELSE}
          Frame3D(TheCanvas,GridCellRect,clBtnHighlight,clBtnShadow,1);
        {$ENDIF}
        SetBkColor(DrawDC,GridColor);
        CellText := Cells[I,J];
        CellTextLen := Length(CellText);
        if CellTextLen <> 0 then
        begin
          { Draw the text in the center of the cell }
          {$IFNDEF WIN32}
          GetTextExtentPoint(DrawDC, @CellText[1], CellTextLen, SizeRec);
          {$ELSE}
          GetTextExtentPoint32(DrawDC, PChar(CellText), CellTextLen, SizeRec);
          {$ENDIF}
          tW := (CurrWidth - SizeRec.cx) shr 1;
          tH := (CurrHeight - SizeRec.cy) shr 1;
          {$IFNDEF WIN32}
          WinProcs.TextOut(DrawDC, X + tW, Y + tH, @CellText[1], CellTextLen);
          {$ELSE}
          Windows.TextOut(DrawDC, X + tW, Y + tH, PChar(CellText), CellTextLen);
          {$ENDIF}
        end;
        Inc(Y,CurrHeight);
      end;
      Inc(X,CurrWidth);
    end;
    DrawGridLines;
  end;
  { Now copy the bitmap to the screen }
  if fUseBitmap then
    BitBlt(Canvas.Handle, 0, 0, Width, Height, DrawDC, 0, 0, SRCCOPY);
end;

constructor TYearPlanner.Create(AOwner: TComponent);
var
  Dy,Mn,Yr: Word;
begin
  { Setup the control }
  Inherited Create(AOwner);
  HintWin := THintWindow.Create(Self);
  fStringList := TStringList.Create;
  CopyRightPtr := @CopyRightStr;
  Width := 615;
  Height := 300;
  Color := clGray;
  fAbout := abNone;
  fDayColor := clWhite;
  fGridLineColor := clBlack;
  fHeadingColor := clGray;
  fHintColor := clYellow;
  fHintDelay := 0;
  fMonthColor := clGray;
  fNoDayColor := clSilver;
  fNoDayPriority := False;
  fSelectionColor := clBlue;
  fStartDayOfWeek := ypMonday;
  fUseBitmap := True;
  fUseFreeSpace := True;
  fWeekendColor := clGray;
  fWeekendHeadingColor := clSilver;
  DecodeDate(Date,Yr,Mn,Dy);
  fYear := Yr;
  hUpdating := False;
  hWaiting := False;
  hWaitingToDestroy := False;
  OldX := -1;
  OldY := -1;
  Printing := False;
  Selecting := ypNotSelecting;
  CurrentDate.Day := 0;
  CurrentDate.Month := 0;
  SetupHeadings;
  CalculateCalendar;
  CalculateData;
  CalculateSizes;
  { Create the off screen bitmap }
  fControl := TBitmap.Create;
end;

destructor TYearPlanner.Destroy;
begin
  { Kill the control }
  fStringList.Free;
  { Inform the hint window that the control is destroying }
  hWaitingToDestroy := True;
  { Wait for the hint window to disappear before losing it }
  repeat Application.ProcessMessages until not hUpdating;
  HintWin.Free;
  { Free used bitmap }
  fControl.Free;
  Inherited Destroy;
end;

{ Thanks to Paul Fisher for helping with this routine }
procedure TYearPlanner.WMLButtonDown(var Message: TWMLButtonDown);
var
  Temp: TPoint;
  tX,tY: Integer;
begin
  { If the user has assigned an OnClick event, we cannot use selections }
  if Assigned(fOnClick) then
  begin
    fOnClick(Self);
    Exit;
  end;
  { Check to see if the mouse is over a cell }
  Temp := ClientToScreen(Point(Message.XPos,Message.YPos));
  if not (FindDragTarget(Temp, True) = Self) then Exit;
  XYToCell(Message.XPos,Message.YPos,tX,tY);
  { Set the initial and start coordinates }
  Selecting := ypSelecting;
  InDay := tX;
  InMonth := tY;
  StDay := InDay;
  StMonth := InMonth;
  EnDay := InDay;
  EnMonth := InMonth;
  Invalidate;
end;

{ Thanks to Paul Fisher for helping with this routine }
procedure TYearPlanner.WMLMouseButtonUp(var Message: TWMLButtonUp);
var
  CountX,CountY: Integer;
begin
  { We cannot allow the user to select a range of cells which do not
    contain dates }
  Selecting := ypNotSelecting;
  for CountX := StDay to EnDay do
    for CountY := StMonth to EnMonth do
      if Cells[CountX,CountY] <> '' then
      begin
        Selecting := ypSelected;
        Exit;
      end;
end;

{ Thanks to Paul Fisher for helping with this routine }
procedure TYearPlanner.WMRMouseButtonDown(var Message: TWMRButtonDown);
begin
  { If a selection has been made, and a right click procedure has been set,
    execute it }
  if (Selecting = ypSelected) and (Assigned(fOnYearRightClick)) then
    fOnYearRightClick(StDay,EnDay,EnMonth,StMonth);
end;

procedure TYearPlanner.WMMouseMove(var Message: TWMMouseMove);
var
  Temp: TPoint;
  HintText,TmpHint,TmpText: String;
  HintRect: TRect;
  HDelay : {$IFDEF WIN32}Cardinal{$ELSE}LongInt{$ENDIF};
  HintH,HintLines,HintW: Integer;
  Dy,Mn: Byte;
  swapTmp:integer;
begin
  Inherited;
  { If the control is destroying we cannot continue }
  if hWaitingToDestroy then Exit;
  { Check to see if the mouse is over a cell }
  Temp := ClientToScreen(Point(Message.XPos,Message.YPos));
  if not (FindDragTarget(Temp, True) = Self) then Exit;
  XYToCell(Message.XPos,Message.YPos,cX,cY);
  { We do not use hints when selecting cells }
  if Selecting = ypSelecting then
  begin
    { Update the selection coordinates }
    StDay := InDay;
    StMonth := InMonth;
    EnDay := cX;
    EnMonth := cY;
    if StDay > EnDay then
    begin
      swapTmp := StDay;
      StDay := EnDay;
      EnDay := swapTmp;
    end;
    if StMonth > EnMonth then
    begin
      swapTmp := StMonth;
      StMonth := EnMonth;
      EnMonth := swapTmp;
    end;
    Invalidate;
    Exit;
  end;
  if (OldX = cX) and (OldY = cY) then Exit;
  { Do we update the current date ? }
  if (Cells[cX,cY] <> '') and (cX > 0) and (cY > 0) then
  begin
    { Yes - It's a calender day }
    CurrentDate.Day := StrToInt(Cells[cX,cY]);
    CurrentDate.Month := cY;
  end;
  { Now check to see if we can use hints }
  if not (Application.ShowHint and (ShowHint or ParentShowHint)) then Exit;
  { If a hint is being displayed, we mark a hint status flag to say that
    another hint is waiting }
  if hUpdating then
  begin
    hWaiting := True;
    Exit;
  end;
  { Now we setup the hint }
  OldX := cX;
  OldY := cY;
  if (Cells[cX,cY] <> '') and (cX > 0) and (cY > 0) then
  begin
    { It's a calender day }
    Dy := CurrentDate.Day;
    Mn := CurrentDate.Month;
    HintText := CellData[Mn,Dy].CellHint;
    if HintText = '' then HintText := DateTimeToStr(EncodeDate(Year,Mn,Dy));
    HintDate := CellData[Mn,Dy].CellDate;
  end
  else Exit;
  { Set the hint status flags }
  hUpdating := True;
  hWaiting := False;
  { Set the hint height and width }
  TmpHint := HintText;
  if TmpHint[length(TmpHint)] <> #13 then
    TmpHint := TmpHint + #13;
  HintLines := 0;
  HintW := 0;
  repeat
    Inc(HintLines);
    TmpText := Copy(TmpHint,1,Pos(#13,TmpHint)-1);
    if HintWin.Canvas.TextWidth(TmpText) + 5 > HintW then
      HintW := HintWin.Canvas.TextWidth(TmpText) + 5;
    Delete(TmpHint,1,Pos(#13,TmpHint));
  until Pos(#13,TmpHint) = 0;
  HintH := (HintWin.Canvas.TextHeight('0') * HintLines) + 5;
  { Set the delay length }
  if fHintDelay = 0 then HDelay := Application.HintPause else
    HDelay := fHintDelay;
  { Display the hint }
  HintRect := Rect(Temp.X,Temp.Y+HintH,Temp.X+HintW,Temp.Y+(HintH*2));
  HintWin.Color := fHintColor;
  HintWin.ActivateHint(HintRect,HintText);
  { Display the hint window for some time }
  FirstTickCount := GetTickCount;
  repeat
    { If another hint is waiting, get rid of this hint }
    Application.ProcessMessages;
    if (Selecting = ypSelecting) or (hWaiting) or (hWaitingToDestroy) then Break;
  until GetTickCount - FirstTickCount > HDelay;
  { Destroy the hint window }
  HintWin.ReleaseHandle;
  hUpdating := False;
end;

{ Thanks to Robert Gesswein for helping with this procedure }
procedure TYearPlanner.SetColorAtDate(dt: TDateTime; cellColor: TColor; UpdateControl: Boolean);
var
  mm,dd,yy: word;
begin
  DecodeDate(dt, yy, mm, dd);
  CellData[mm, dd].CellColor := cellColor;
  CellData[mm, dd].CustomColor := True;
  if UpdateControl then Invalidate;
end;

{ Thanks to Paul Fisher for this routine }
procedure TYearPlanner.Print;
begin;
  Printing := True;
  Printer.Orientation := poLandscape;
  Printer.BeginDoc;
  self.Paint;
  Printer.EndDoc;
  Printing := False;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TYearPlanner]);
end;

end.
